@@ -1,8 +1,7 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2005
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1996-2009 Oracle.  All rights reserved.
  */
 /*
  * Copyright (c) 1990, 1993, 1994, 1995, 1996
@@ -36,17 +35,11 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- * $Id: db_conv.c,v 12.1 2005/06/16 20:21:09 bostic Exp $
+ * $Id$
  */
 
 #include "db_config.h"
 
-#ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-
-#include <string.h>
-#endif
-
 #include "db_int.h"
 #include "dbinc/crypto.h"
 #include "dbinc/hmac.h"
@@ -71,23 +64,28 @@ __db_pgin(dbenv, pg, pp, cookie)
 	DBT *cookie;
 {
 	DB dummydb, *dbp;
-	DB_PGINFO *pginfo;
 	DB_CIPHER *db_cipher;
 	DB_LSN not_used;
+	DB_PGINFO *pginfo;
+	ENV *env;
 	PAGE *pagep;
-	size_t pg_off, pg_len, sum_len;
+	size_t sum_len;
 	int is_hmac, ret;
-	u_int8_t *chksum, *iv;
+	u_int8_t *chksum;
 
 	pginfo = (DB_PGINFO *)cookie->data;
+	env = dbenv->env;
 	pagep = (PAGE *)pp;
 
 	ret = is_hmac = 0;
-	chksum = iv = NULL;
+	chksum = NULL;
 	memset(&dummydb, 0, sizeof(DB));
 	dbp = &dummydb;
-	dummydb.flags = pginfo->flags;
-	db_cipher = (DB_CIPHER *)dbenv->crypto_handle;
+	dbp->dbenv = dbenv;
+	dbp->env = env;
+	dbp->flags = pginfo->flags;
+	dbp->pgsize = pginfo->db_pagesize;
+	db_cipher = env->crypto_handle;
 	switch (pagep->type) {
 	case P_HASHMETA:
 	case P_BTREEMETA:
@@ -130,7 +128,7 @@ __db_pgin(dbenv, pg, pp, cookie)
 		 * If we are reading in a non-meta page, then if we have
 		 * a db_cipher then we are using hmac.
 		 */
-		is_hmac = CRYPTO_ON(dbenv) ? 1 : 0;
+		is_hmac = CRYPTO_ON(env) ? 1 : 0;
 		break;
 	}
 
@@ -143,67 +141,33 @@ __db_pgin(dbenv, pg, pp, cookie)
 		if (F_ISSET(dbp, DB_AM_SWAP) && is_hmac == 0)
 			P_32_SWAP(chksum);
 		switch (ret = __db_check_chksum(
-		    dbenv, db_cipher, chksum, pp, sum_len, is_hmac)) {
+		    env, NULL, db_cipher, chksum, pp, sum_len, is_hmac)) {
 		case 0:
 			break;
 		case -1:
-			if (DBENV_LOGGING(dbenv))
+			if (DBENV_LOGGING(env))
 				(void)__db_cksum_log(
-				    dbenv, NULL, &not_used, DB_FLUSH);
-			__db_err(dbenv,
+				    env, NULL, &not_used, DB_FLUSH);
+			__db_errx(env,
 	    "checksum error: page %lu: catastrophic recovery required",
 			    (u_long)pg);
-			return (__db_panic(dbenv, DB_RUNRECOVERY));
+			return (__env_panic(env, DB_RUNRECOVERY));
 		default:
 			return (ret);
 		}
 	}
-	if (F_ISSET(dbp, DB_AM_ENCRYPT)) {
-		DB_ASSERT(db_cipher != NULL);
-		DB_ASSERT(F_ISSET(dbp, DB_AM_CHKSUM));
-
-		pg_off = P_OVERHEAD(dbp);
-		DB_ASSERT(db_cipher->adj_size(pg_off) == 0);
-
-		switch (pagep->type) {
-		case P_HASHMETA:
-		case P_BTREEMETA:
-		case P_QAMMETA:
-			/*
-			 * !!!
-			 * For all meta pages it is required that the iv
-			 * be at the same location.  Use BTMETA to get to it
-			 * for any meta type.
-			 */
-			iv = ((BTMETA *)pp)->iv;
-			pg_len = DBMETASIZE;
-			break;
-		case P_INVALID:
-			if (IS_ZERO_LSN(LSN(pagep)) &&
-			    pagep->pgno == PGNO_INVALID) {
-				pg_len = 0;
-				break;
-			}
-			/* FALLTHROUGH */
-		default:
-			iv = P_IV(dbp, pagep);
-			pg_len = pginfo->db_pagesize;
-			break;
-		}
-		if (pg_len != 0 && (ret = db_cipher->decrypt(dbenv,
-		    db_cipher->data, iv, ((u_int8_t *)pagep) + pg_off,
-		    pg_len - pg_off)) != 0)
-			return (ret);
-	}
+	if ((ret = __db_decrypt_pg(env, dbp, pagep)) != 0)
+		return (ret);
 	switch (pagep->type) {
 	case P_INVALID:
 		if (pginfo->type == DB_QUEUE)
-			return (__qam_pgin_out(dbenv, pg, pp, cookie));
+			return (__qam_pgin_out(env, pg, pp, cookie));
 		else
-			return (__ham_pgin(dbenv, dbp, pg, pp, cookie));
+			return (__ham_pgin(dbp, pg, pp, cookie));
+	case P_HASH_UNSORTED:
 	case P_HASH:
 	case P_HASHMETA:
-		return (__ham_pgin(dbenv, dbp, pg, pp, cookie));
+		return (__ham_pgin(dbp, pg, pp, cookie));
 	case P_BTREEMETA:
 	case P_IBTREE:
 	case P_IRECNO:
@@ -211,14 +175,14 @@ __db_pgin(dbenv, pg, pp, cookie)
 	case P_LDUP:
 	case P_LRECNO:
 	case P_OVERFLOW:
-		return (__bam_pgin(dbenv, dbp, pg, pp, cookie));
+		return (__bam_pgin(dbp, pg, pp, cookie));
 	case P_QAMMETA:
 	case P_QAMDATA:
-		return (__qam_pgin_out(dbenv, pg, pp, cookie));
+		return (__qam_pgin_out(env, pg, pp, cookie));
 	default:
 		break;
 	}
-	return (__db_pgfmt(dbenv, pg));
+	return (__db_pgfmt(env, pg));
 }
 
 /*
@@ -235,31 +199,40 @@ __db_pgout(dbenv, pg, pp, cookie)
 	DBT *cookie;
 {
 	DB dummydb, *dbp;
-	DB_CIPHER *db_cipher;
 	DB_PGINFO *pginfo;
+	ENV *env;
 	PAGE *pagep;
-	size_t pg_off, pg_len, sum_len;
 	int ret;
-	u_int8_t *chksum, *iv, *key;
 
 	pginfo = (DB_PGINFO *)cookie->data;
+	env = dbenv->env;
 	pagep = (PAGE *)pp;
 
-	chksum = iv = key = NULL;
 	memset(&dummydb, 0, sizeof(DB));
 	dbp = &dummydb;
-	dummydb.flags = pginfo->flags;
+	dbp->dbenv = dbenv;
+	dbp->env = env;
+	dbp->flags = pginfo->flags;
+	dbp->pgsize = pginfo->db_pagesize;
 	ret = 0;
 	switch (pagep->type) {
 	case P_INVALID:
 		if (pginfo->type == DB_QUEUE)
-			ret = __qam_pgin_out(dbenv, pg, pp, cookie);
+			ret = __qam_pgin_out(env, pg, pp, cookie);
 		else
-			ret = __ham_pgout(dbenv, dbp, pg, pp, cookie);
+			ret = __ham_pgout(dbp, pg, pp, cookie);
 		break;
 	case P_HASH:
+	case P_HASH_UNSORTED:
+		/*
+		 * Support pgout of unsorted hash pages - since online
+		 * replication upgrade can cause pages of this type to be
+		 * written out.
+		 *
+		 * FALLTHROUGH
+		 */
 	case P_HASHMETA:
-		ret = __ham_pgout(dbenv, dbp, pg, pp, cookie);
+		ret = __ham_pgout(dbp, pg, pp, cookie);
 		break;
 	case P_BTREEMETA:
 	case P_IBTREE:
@@ -268,26 +241,108 @@ __db_pgout(dbenv, pg, pp, cookie)
 	case P_LDUP:
 	case P_LRECNO:
 	case P_OVERFLOW:
-		ret = __bam_pgout(dbenv, dbp, pg, pp, cookie);
+		ret = __bam_pgout(dbp, pg, pp, cookie);
 		break;
 	case P_QAMMETA:
 	case P_QAMDATA:
-		ret = __qam_pgin_out(dbenv, pg, pp, cookie);
+		ret = __qam_pgin_out(env, pg, pp, cookie);
 		break;
 	default:
-		return (__db_pgfmt(dbenv, pg));
+		return (__db_pgfmt(env, pg));
 	}
 	if (ret)
 		return (ret);
 
-	db_cipher = (DB_CIPHER *)dbenv->crypto_handle;
-	if (F_ISSET(dbp, DB_AM_ENCRYPT)) {
+	return (__db_encrypt_and_checksum_pg(env, dbp, pagep));
+}
 
-		DB_ASSERT(db_cipher != NULL);
-		DB_ASSERT(F_ISSET(dbp, DB_AM_CHKSUM));
+/*
+ * __db_decrypt_pg --
+ *      Utility function to decrypt a db page.
+ *
+ * PUBLIC: int __db_decrypt_pg __P((ENV *, DB *, PAGE *));
+ */
+int
+__db_decrypt_pg (env, dbp, pagep)
+	ENV *env;
+	DB *dbp;
+	PAGE *pagep;
+{
+	DB_CIPHER *db_cipher;
+	size_t pg_len, pg_off;
+	u_int8_t *iv;
+	int ret;
+
+	db_cipher = env->crypto_handle;
+	ret = 0;
+	iv = NULL;
+	if (F_ISSET(dbp, DB_AM_ENCRYPT)) {
+		DB_ASSERT(env, db_cipher != NULL);
+		DB_ASSERT(env, F_ISSET(dbp, DB_AM_CHKSUM));
 
 		pg_off = P_OVERHEAD(dbp);
-		DB_ASSERT(db_cipher->adj_size(pg_off) == 0);
+		DB_ASSERT(env, db_cipher->adj_size(pg_off) == 0);
+
+		switch (pagep->type) {
+		case P_HASHMETA:
+		case P_BTREEMETA:
+		case P_QAMMETA:
+			/*
+			 * !!!
+			 * For all meta pages it is required that the iv
+			 * be at the same location.  Use BTMETA to get to it
+			 * for any meta type.
+			 */
+			iv = ((BTMETA *)pagep)->iv;
+			pg_len = DBMETASIZE;
+			break;
+		case P_INVALID:
+			if (IS_ZERO_LSN(LSN(pagep)) &&
+			    pagep->pgno == PGNO_INVALID) {
+				pg_len = 0;
+				break;
+			}
+			/* FALLTHROUGH */
+		default:
+			iv = P_IV(dbp, pagep);
+			pg_len = dbp->pgsize;
+			break;
+		}
+		if (pg_len != 0)
+			ret = db_cipher->decrypt(env, db_cipher->data,
+			    iv, ((u_int8_t *)pagep) + pg_off,
+			    pg_len - pg_off);
+	}
+	return (ret);
+}
+
+/*
+ * __db_encrypt_and_checksum_pg --
+ *	Utility function to encrypt and checksum a db page.
+ *
+ * PUBLIC: int __db_encrypt_and_checksum_pg
+ * PUBLIC:     __P((ENV *, DB *, PAGE *));
+ */
+int
+__db_encrypt_and_checksum_pg (env, dbp, pagep)
+	ENV *env;
+	DB *dbp;
+	PAGE *pagep;
+{
+	DB_CIPHER *db_cipher;
+	int ret;
+	size_t pg_off, pg_len, sum_len;
+	u_int8_t *chksum, *iv, *key;
+
+	chksum = iv = key = NULL;
+	db_cipher = env->crypto_handle;
+
+	if (F_ISSET(dbp, DB_AM_ENCRYPT)) {
+		DB_ASSERT(env, db_cipher != NULL);
+		DB_ASSERT(env, F_ISSET(dbp, DB_AM_CHKSUM));
+
+		pg_off = P_OVERHEAD(dbp);
+		DB_ASSERT(env, db_cipher->adj_size(pg_off) == 0);
 
 		key = db_cipher->mac_key;
 
@@ -301,15 +356,15 @@ __db_pgout(dbenv, pg, pp, cookie)
 			 * be at the same location.  Use BTMETA to get to it
 			 * for any meta type.
 			 */
-			iv = ((BTMETA *)pp)->iv;
+			iv = ((BTMETA *)pagep)->iv;
 			pg_len = DBMETASIZE;
 			break;
 		default:
 			iv = P_IV(dbp, pagep);
-			pg_len = pginfo->db_pagesize;
+			pg_len = dbp->pgsize;
 			break;
 		}
-		if ((ret = db_cipher->encrypt(dbenv, db_cipher->data,
+		if ((ret = db_cipher->encrypt(env, db_cipher->data,
 		    iv, ((u_int8_t *)pagep) + pg_off, pg_len - pg_off)) != 0)
 			return (ret);
 	}
@@ -324,15 +379,15 @@ __db_pgout(dbenv, pg, pp, cookie)
 			 * be at the same location.  Use BTMETA to get to it
 			 * for any meta type.
 			 */
-			chksum = ((BTMETA *)pp)->chksum;
+			chksum = ((BTMETA *)pagep)->chksum;
 			sum_len = DBMETASIZE;
 			break;
 		default:
 			chksum = P_CHKSUM(dbp, pagep);
-			sum_len = pginfo->db_pagesize;
+			sum_len = dbp->pgsize;
 			break;
 		}
-		__db_chksum(pp, sum_len, key, chksum);
+		__db_chksum(NULL, (u_int8_t *)pagep, sum_len, key, chksum);
 		if (F_ISSET(dbp, DB_AM_SWAP) && !F_ISSET(dbp, DB_AM_ENCRYPT))
 			 P_32_SWAP(chksum);
 	}
@@ -371,30 +426,32 @@ __db_metaswap(pg)
 
 /*
  * __db_byteswap --
- *	Byteswap a page.
+ *	Byteswap an ordinary database page.
  *
  * PUBLIC: int __db_byteswap
- * PUBLIC:         __P((DB_ENV *, DB *, db_pgno_t, PAGE *, size_t, int));
+ * PUBLIC:         __P((DB *, db_pgno_t, PAGE *, size_t, int));
  */
 int
-__db_byteswap(dbenv, dbp, pg, h, pagesize, pgin)
-	DB_ENV *dbenv;
+__db_byteswap(dbp, pg, h, pagesize, pgin)
 	DB *dbp;
 	db_pgno_t pg;
 	PAGE *h;
 	size_t pagesize;
 	int pgin;
 {
+	ENV *env;
 	BINTERNAL *bi;
 	BKEYDATA *bk;
 	BOVERFLOW *bo;
 	RINTERNAL *ri;
 	db_indx_t i, *inp, len, tmp;
-	u_int8_t *p, *end;
+	u_int8_t *end, *p, *pgend;
 
-	COMPQUIET(pg, 0);
+	if (pagesize == 0)
+		return (0);
+
+	env = dbp->env;
 
-	inp = P_INP(dbp, h);
 	if (pgin) {
 		M_32_SWAP(h->lsn.file);
 		M_32_SWAP(h->lsn.offset);
@@ -405,12 +462,22 @@ __db_byteswap(dbenv, dbp, pg, h, pagesize, pgin)
 		M_16_SWAP(h->hf_offset);
 	}
 
-	switch (h->type) {
+	pgend = (u_int8_t *)h + pagesize;
+
+	inp = P_INP(dbp, h);
+	if ((u_int8_t *)inp >= pgend)
+		goto out;
+
+	switch (TYPE(h)) {
+	case P_HASH_UNSORTED:
 	case P_HASH:
 		for (i = 0; i < NUM_ENT(h); i++) {
 			if (pgin)
 				M_16_SWAP(inp[i]);
 
+			if (P_ENTRY(dbp, h, i) >= pgend)
+				continue;
+
 			switch (HPAGE_TYPE(dbp, h, i)) {
 			case H_KEYDATA:
 				break;
@@ -442,7 +509,7 @@ __db_byteswap(dbenv, dbp, pg, h, pagesize, pgin)
 				SWAP32(p);			/* tlen */
 				break;
 			default:
-				return (__db_pgfmt(dbenv, pg));
+				return (__db_pgfmt(env, pg));
 			}
 
 		}
@@ -481,6 +548,8 @@ __db_byteswap(dbenv, dbp, pg, h, pagesize, pgin)
 			}
 
 			bk = GET_BKEYDATA(dbp, h, i);
+			if ((u_int8_t *)bk >= pgend)
+				continue;
 			switch (B_TYPE(bk->type)) {
 			case B_KEYDATA:
 				M_16_SWAP(bk->len);
@@ -492,7 +561,7 @@ __db_byteswap(dbenv, dbp, pg, h, pagesize, pgin)
 				M_32_SWAP(bo->tlen);
 				break;
 			default:
-				return (__db_pgfmt(dbenv, pg));
+				return (__db_pgfmt(env, pg));
 			}
 
 			if (!pgin)
@@ -505,6 +574,9 @@ __db_byteswap(dbenv, dbp, pg, h, pagesize, pgin)
 				M_16_SWAP(inp[i]);
 
 			bi = GET_BINTERNAL(dbp, h, i);
+			if ((u_int8_t *)bi >= pgend)
+				continue;
+
 			M_16_SWAP(bi->len);
 			M_32_SWAP(bi->pgno);
 			M_32_SWAP(bi->nrecs);
@@ -519,7 +591,7 @@ __db_byteswap(dbenv, dbp, pg, h, pagesize, pgin)
 				M_32_SWAP(bo->tlen);
 				break;
 			default:
-				return (__db_pgfmt(dbenv, pg));
+				return (__db_pgfmt(env, pg));
 			}
 
 			if (!pgin)
@@ -532,6 +604,9 @@ __db_byteswap(dbenv, dbp, pg, h, pagesize, pgin)
 				M_16_SWAP(inp[i]);
 
 			ri = GET_RINTERNAL(dbp, h, i);
+			if ((u_int8_t *)ri >= pgend)
+				continue;
+
 			M_32_SWAP(ri->pgno);
 			M_32_SWAP(ri->nrecs);
 
@@ -539,15 +614,16 @@ __db_byteswap(dbenv, dbp, pg, h, pagesize, pgin)
 				M_16_SWAP(inp[i]);
 		}
 		break;
-	case P_OVERFLOW:
 	case P_INVALID:
+	case P_OVERFLOW:
+	case P_QAMDATA:
 		/* Nothing to do. */
 		break;
 	default:
-		return (__db_pgfmt(dbenv, pg));
+		return (__db_pgfmt(env, pg));
 	}
 
-	if (!pgin) {
+out:	if (!pgin) {
 		/* Swap the header information. */
 		M_32_SWAP(h->lsn.file);
 		M_32_SWAP(h->lsn.offset);
@@ -559,3 +635,99 @@ __db_byteswap(dbenv, dbp, pg, h, pagesize, pgin)
 	}
 	return (0);
 }
+
+/*
+ * __db_pageswap --
+ *	Byteswap any database page.  Normally, the page to be swapped will be
+ *	referenced by the "pp" argument and the pdata argument will be NULL.
+ *	This function is also called by automatically generated log functions,
+ *	where the page may be split into separate header and data parts.  In
+ *	that case, pdata is not NULL we reconsitute
+ *
+ * PUBLIC: int __db_pageswap
+ * PUBLIC:         __P((DB *, void *, size_t, DBT *, int));
+ */
+int
+__db_pageswap(dbp, pp, len, pdata, pgin)
+	DB *dbp;
+	void *pp;
+	size_t len;
+	DBT *pdata;
+	int pgin;
+{
+	ENV *env;
+	db_pgno_t pg;
+	size_t pgsize;
+	void *pgcopy;
+	int ret;
+	u_int16_t hoffset;
+
+	env = dbp->env;
+
+	switch (TYPE(pp)) {
+	case P_BTREEMETA:
+		return (__bam_mswap(env, pp));
+
+	case P_HASHMETA:
+		return (__ham_mswap(env, pp));
+
+	case P_QAMMETA:
+		return (__qam_mswap(env, pp));
+
+	case P_INVALID:
+	case P_OVERFLOW:
+	case P_QAMDATA:
+		/*
+		 * We may have been passed an invalid page, or a queue data
+		 * page, or an overflow page where fields like hoffset have a
+		 * special meaning.  In that case, no swapping of the page data
+		 * is required, just the fields in the page header.
+		 */
+		pdata = NULL;
+		break;
+
+	default:
+		break;
+	}
+
+	if (pgin) {
+		P_32_COPYSWAP(&PGNO(pp), &pg);
+		P_16_COPYSWAP(&HOFFSET(pp), &hoffset);
+	} else {
+		pg = PGNO(pp);
+		hoffset = HOFFSET(pp);
+	}
+
+	if (pdata == NULL)
+		ret = __db_byteswap(dbp, pg, (PAGE *)pp, len, pgin);
+	else {
+		pgsize = hoffset + pdata->size;
+		if ((ret = __os_malloc(env, pgsize, &pgcopy)) != 0)
+			return (ret);
+		memset(pgcopy, 0, pgsize);
+		memcpy(pgcopy, pp, len);
+		memcpy((u_int8_t *)pgcopy + hoffset, pdata->data, pdata->size);
+
+		ret = __db_byteswap(dbp, pg, (PAGE *)pgcopy, pgsize, pgin);
+		memcpy(pp, pgcopy, len);
+
+		/*
+		 * If we are swapping data to be written to the log, we can't
+		 * overwrite the buffer that was passed in: it may be a pointer
+		 * into a page in cache.  We set DB_DBT_APPMALLOC here so that
+		 * the calling code can free the memory we allocate here.
+		 */
+		if (!pgin) {
+			if ((ret =
+			    __os_malloc(env, pdata->size, &pdata->data)) != 0) {
+				__os_free(env, pgcopy);
+				return (ret);
+			}
+			F_SET(pdata, DB_DBT_APPMALLOC);
+		}
+		memcpy(pdata->data, (u_int8_t *)pgcopy + hoffset, pdata->size);
+		__os_free(env, pgcopy);
+	}
+
+	return (ret);
+}
