@@ -1,25 +1,18 @@
 /*
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1998-2005
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1998-2009 Oracle.  All rights reserved.
  *
- * $Id: db_join.c,v 12.6 2005/10/07 20:21:22 ubell Exp $
+ * $Id$
  */
 
 #include "db_config.h"
 
-#ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-
-#include <stdlib.h>
-#include <string.h>
-#endif
-
 #include "db_int.h"
 #include "dbinc/db_page.h"
 #include "dbinc/db_join.h"
 #include "dbinc/btree.h"
+#include "dbinc/lock.h"
 
 static int __db_join_close_pp __P((DBC *));
 static int __db_join_cmp __P((const void *, const void *));
@@ -27,8 +20,8 @@ static int __db_join_del __P((DBC *, u_int32_t));
 static int __db_join_get __P((DBC *, DBT *, DBT *, u_int32_t));
 static int __db_join_get_pp __P((DBC *, DBT *, DBT *, u_int32_t));
 static int __db_join_getnext __P((DBC *, DBT *, DBT *, u_int32_t, u_int32_t));
-static int __db_join_primget __P((DB *,
-    DB_TXN *, u_int32_t, DBT *, DBT *, u_int32_t));
+static int __db_join_primget __P((DB *, DB_THREAD_INFO *,
+    DB_TXN *, DB_LOCKER *, DBT *, DBT *, u_int32_t));
 static int __db_join_put __P((DBC *, DBT *, DBT *, u_int32_t));
 
 /*
@@ -80,24 +73,24 @@ __db_join(primary, curslist, dbcp, flags)
 	DBC **curslist, **dbcp;
 	u_int32_t flags;
 {
-	DB_ENV *dbenv;
 	DBC *dbc;
+	ENV *env;
 	JOIN_CURSOR *jc;
 	size_t ncurs, nslots;
 	u_int32_t i;
 	int ret;
 
-	dbenv = primary->dbenv;
+	env = primary->env;
 	dbc = NULL;
 	jc = NULL;
 
-	if ((ret = __os_calloc(dbenv, 1, sizeof(DBC), &dbc)) != 0)
+	if ((ret = __os_calloc(env, 1, sizeof(DBC), &dbc)) != 0)
 		goto err;
 
-	if ((ret = __os_calloc(dbenv, 1, sizeof(JOIN_CURSOR), &jc)) != 0)
+	if ((ret = __os_calloc(env, 1, sizeof(JOIN_CURSOR), &jc)) != 0)
 		goto err;
 
-	if ((ret = __os_malloc(dbenv, 256, &jc->j_key.data)) != 0)
+	if ((ret = __os_malloc(env, 256, &jc->j_key.data)) != 0)
 		goto err;
 	jc->j_key.ulen = 256;
 	F_SET(&jc->j_key, DB_DBT_USERMEM);
@@ -161,16 +154,16 @@ __db_join(primary, curslist, dbcp, flags)
 	jc->j_fdupcurs = NULL;
 	jc->j_exhausted = NULL;
 
-	if ((ret = __os_calloc(dbenv, nslots, sizeof(DBC *),
+	if ((ret = __os_calloc(env, nslots, sizeof(DBC *),
 	    &jc->j_curslist)) != 0)
 		goto err;
-	if ((ret = __os_calloc(dbenv, nslots, sizeof(DBC *),
+	if ((ret = __os_calloc(env, nslots, sizeof(DBC *),
 	    &jc->j_workcurs)) != 0)
 		goto err;
-	if ((ret = __os_calloc(dbenv, nslots, sizeof(DBC *),
+	if ((ret = __os_calloc(env, nslots, sizeof(DBC *),
 	    &jc->j_fdupcurs)) != 0)
 		goto err;
-	if ((ret = __os_calloc(dbenv, nslots, sizeof(u_int8_t),
+	if ((ret = __os_calloc(env, nslots, sizeof(u_int8_t),
 	    &jc->j_exhausted)) != 0)
 		goto err;
 	for (i = 0; curslist[i] != NULL; i++) {
@@ -200,13 +193,13 @@ __db_join(primary, curslist, dbcp, flags)
 	 * function beware!
 	 */
 	if ((ret =
-	    __db_c_dup(jc->j_curslist[0], jc->j_workcurs, DB_POSITION)) != 0)
+	    __dbc_dup(jc->j_curslist[0], jc->j_workcurs, DB_POSITION)) != 0)
 		goto err;
 
-	dbc->c_close = __db_join_close_pp;
-	dbc->c_del = __db_join_del;
-	dbc->c_get = __db_join_get_pp;
-	dbc->c_put = __db_join_put;
+	dbc->close = dbc->c_close = __db_join_close_pp;
+	dbc->del = dbc->c_del = __db_join_del;
+	dbc->get = dbc->c_get = __db_join_get_pp;
+	dbc->put = dbc->c_put = __db_join_put;
 	dbc->internal = (DBC_INTERNAL *)jc;
 	dbc->dbp = primary;
 	jc->j_primary = primary;
@@ -216,52 +209,50 @@ __db_join(primary, curslist, dbcp, flags)
 
 	*dbcp = dbc;
 
-	MUTEX_LOCK(dbenv, primary->mutex);
+	MUTEX_LOCK(env, primary->mutex);
 	TAILQ_INSERT_TAIL(&primary->join_queue, dbc, links);
-	MUTEX_UNLOCK(dbenv, primary->mutex);
+	MUTEX_UNLOCK(env, primary->mutex);
 
 	return (0);
 
 err:	if (jc != NULL) {
 		if (jc->j_curslist != NULL)
-			__os_free(dbenv, jc->j_curslist);
+			__os_free(env, jc->j_curslist);
 		if (jc->j_workcurs != NULL) {
 			if (jc->j_workcurs[0] != NULL)
-				(void)__db_c_close(jc->j_workcurs[0]);
-			__os_free(dbenv, jc->j_workcurs);
+				(void)__dbc_close(jc->j_workcurs[0]);
+			__os_free(env, jc->j_workcurs);
 		}
 		if (jc->j_fdupcurs != NULL)
-			__os_free(dbenv, jc->j_fdupcurs);
+			__os_free(env, jc->j_fdupcurs);
 		if (jc->j_exhausted != NULL)
-			__os_free(dbenv, jc->j_exhausted);
-		__os_free(dbenv, jc);
+			__os_free(env, jc->j_exhausted);
+		__os_free(env, jc);
 	}
 	if (dbc != NULL)
-		__os_free(dbenv, dbc);
+		__os_free(env, dbc);
 	return (ret);
 }
 
 /*
  * __db_join_close_pp --
- *	DBC->c_close pre/post processing for join cursors.
+ *	DBC->close pre/post processing for join cursors.
  */
 static int
 __db_join_close_pp(dbc)
 	DBC *dbc;
 {
-	DB_ENV *dbenv;
-	DB_THREAD_INFO *ip;
 	DB *dbp;
+	DB_THREAD_INFO *ip;
+	ENV *env;
 	int handle_check, ret, t_ret;
 
 	dbp = dbc->dbp;
-	dbenv = dbp->dbenv;
+	env = dbp->env;
 
-	PANIC_CHECK(dbenv);
+	ENV_ENTER(env, ip);
 
-	ENV_ENTER(dbenv, ip);
-
-	handle_check = IS_ENV_REPLICATED(dbenv);
+	handle_check = IS_ENV_REPLICATED(env);
 	if (handle_check &&
 	    (ret = __db_rep_enter(dbp, 1, 0, dbc->txn != NULL)) != 0) {
 		handle_check = 0;
@@ -270,10 +261,10 @@ __db_join_close_pp(dbc)
 
 	ret = __db_join_close(dbc);
 
-	if (handle_check && (t_ret = __env_db_rep_exit(dbenv)) != 0 && ret == 0)
+	if (handle_check && (t_ret = __env_db_rep_exit(env)) != 0 && ret == 0)
 		ret = t_ret;
 
-err:	ENV_LEAVE(dbenv, ip);
+err:	ENV_LEAVE(env, ip);
 	return (ret);
 }
 
@@ -284,8 +275,7 @@ __db_join_put(dbc, key, data, flags)
 	DBT *data;
 	u_int32_t flags;
 {
-	PANIC_CHECK(dbc->dbp->dbenv);
-
+	COMPQUIET(dbc, NULL);
 	COMPQUIET(key, NULL);
 	COMPQUIET(data, NULL);
 	COMPQUIET(flags, 0);
@@ -297,8 +287,7 @@ __db_join_del(dbc, flags)
 	DBC *dbc;
 	u_int32_t flags;
 {
-	PANIC_CHECK(dbc->dbp->dbenv);
-
+	COMPQUIET(dbc, NULL);
 	COMPQUIET(flags, 0);
 	return (EINVAL);
 }
@@ -314,22 +303,20 @@ __db_join_get_pp(dbc, key, data, flags)
 	u_int32_t flags;
 {
 	DB *dbp;
-	DB_ENV *dbenv;
 	DB_THREAD_INFO *ip;
+	ENV *env;
 	u_int32_t handle_check, save_flags;
 	int ret, t_ret;
 
 	dbp = dbc->dbp;
-	dbenv = dbp->dbenv;
+	env = dbp->env;
 
 	/* Save the original flags value. */
 	save_flags = flags;
 
-	PANIC_CHECK(dbenv);
-
 	if (LF_ISSET(DB_READ_COMMITTED | DB_READ_UNCOMMITTED | DB_RMW)) {
-		if (!LOCKING_ON(dbp->dbenv))
-			return (__db_fnl(dbp->dbenv, "DBcursor->c_get"));
+		if (!LOCKING_ON(env))
+			return (__db_fnl(env, "DBC->get"));
 		LF_CLR(DB_READ_COMMITTED | DB_READ_UNCOMMITTED | DB_RMW);
 	}
 
@@ -338,7 +325,7 @@ __db_join_get_pp(dbc, key, data, flags)
 	case DB_JOIN_ITEM:
 		break;
 	default:
-		return (__db_ferr(dbp->dbenv, "DBcursor->c_get", 0));
+		return (__db_ferr(env, "DBC->get", 0));
 	}
 
 	/*
@@ -353,14 +340,14 @@ __db_join_get_pp(dbc, key, data, flags)
 	 * and causes us no headaches, so we permit it.
 	 */
 	if (F_ISSET(key, DB_DBT_PARTIAL)) {
-		__db_err(dbp->dbenv,
+		__db_errx(env,
 		    "DB_DBT_PARTIAL may not be set on key during join_get");
 		return (EINVAL);
 	}
 
-	ENV_ENTER(dbenv, ip);
+	ENV_ENTER(env, ip);
 
-	handle_check = IS_ENV_REPLICATED(dbp->dbenv);
+	handle_check = IS_ENV_REPLICATED(env);
 	if (handle_check &&
 	    (ret = __db_rep_enter(dbp, 1, 0, dbc->txn != NULL)) != 0) {
 		handle_check = 0;
@@ -372,10 +359,11 @@ __db_join_get_pp(dbc, key, data, flags)
 
 	ret = __db_join_get(dbc, key, data, flags);
 
-	if (handle_check && (t_ret = __env_db_rep_exit(dbenv)) != 0 && ret == 0)
+	if (handle_check && (t_ret = __env_db_rep_exit(env)) != 0 && ret == 0)
 		ret = t_ret;
 
-err:	ENV_LEAVE(dbenv, ip);
+err:	ENV_LEAVE(env, ip);
+	__dbt_userfree(env, key, NULL, NULL);
 	return (ret);
 }
 
@@ -385,14 +373,16 @@ __db_join_get(dbc, key_arg, data_arg, flags)
 	DBT *key_arg, *data_arg;
 	u_int32_t flags;
 {
-	DBT *key_n, key_n_mem;
 	DB *dbp;
 	DBC *cp;
+	DBT *key_n, key_n_mem;
+	ENV *env;
 	JOIN_CURSOR *jc;
 	int db_manage_data, ret;
 	u_int32_t i, j, operation, opmods;
 
 	dbp = dbc->dbp;
+	env = dbp->env;
 	jc = (JOIN_CURSOR *)dbc->internal;
 
 	operation = LF_ISSET(DB_OPFLAGS_MASK);
@@ -409,8 +399,8 @@ __db_join_get(dbc, key_arg, data_arg, flags)
 	 * management flags.  If necessary, use a stack-allocated DBT;
 	 * we'll appropriately copy and/or allocate the data later.
 	 */
-	if (F_ISSET(key_arg, DB_DBT_USERMEM) ||
-	    F_ISSET(key_arg, DB_DBT_MALLOC)) {
+	if (F_ISSET(key_arg,
+	    DB_DBT_MALLOC | DB_DBT_USERCOPY | DB_DBT_USERMEM)) {
 		/* We just use the default buffer;  no need to go malloc. */
 		key_n = &key_n_mem;
 		memset(key_n, 0, sizeof(DBT));
@@ -421,6 +411,8 @@ __db_join_get(dbc, key_arg, data_arg, flags)
 		 */
 		key_n = key_arg;
 	}
+	if (F_ISSET(key_arg, DB_DBT_USERCOPY))
+		key_arg->data = NULL;
 
 	/*
 	 * If our last attempt to do a get on the primary key failed,
@@ -430,12 +422,12 @@ __db_join_get(dbc, key_arg, data_arg, flags)
 		goto samekey;
 	F_CLR(jc, JOIN_RETRY);
 
-retry:	ret = __db_c_get(jc->j_workcurs[0], &jc->j_key, key_n,
+retry:	ret = __dbc_get(jc->j_workcurs[0], &jc->j_key, key_n,
 	    opmods | (jc->j_exhausted[0] ? DB_NEXT_DUP : DB_CURRENT));
 
 	if (ret == DB_BUFFER_SMALL) {
 		jc->j_key.ulen <<= 1;
-		if ((ret = __os_realloc(dbp->dbenv,
+		if ((ret = __os_realloc(env,
 		    jc->j_key.ulen, &jc->j_key.data)) != 0)
 			goto mem_err;
 		goto retry;
@@ -458,7 +450,7 @@ retry:	ret = __db_c_get(jc->j_workcurs[0], &jc->j_key, key_n,
 	 */
 	for (i = 1; i < jc->j_ncurs; i++) {
 		if (jc->j_fdupcurs[i] != NULL &&
-		    (ret = __db_c_close(jc->j_fdupcurs[i])) != 0)
+		    (ret = __dbc_close(jc->j_fdupcurs[i])) != 0)
 			goto err;
 		jc->j_fdupcurs[i] = NULL;
 	}
@@ -479,10 +471,10 @@ retry:	ret = __db_c_get(jc->j_workcurs[0], &jc->j_key, key_n,
 
 	/* We have the first element; now look for it in the other cursors. */
 	for (i = 1; i < jc->j_ncurs; i++) {
-		DB_ASSERT(jc->j_curslist[i] != NULL);
+		DB_ASSERT(env, jc->j_curslist[i] != NULL);
 		if (jc->j_workcurs[i] == NULL)
 			/* If this is NULL, we need to dup curslist into it. */
-			if ((ret = __db_c_dup(jc->j_curslist[i],
+			if ((ret = __dbc_dup(jc->j_curslist[i],
 			    &jc->j_workcurs[i], DB_POSITION)) != 0)
 				goto err;
 
@@ -534,7 +526,7 @@ retry2:		cp = jc->j_workcurs[i];
 					 * and let strange things happen--we
 					 * can't make rope childproof.
 					 */
-					if ((ret = __db_c_close(
+					if ((ret = __dbc_close(
 					    jc->j_workcurs[j])) != 0)
 						goto err;
 					if (!SORTED_SET(jc, 0) ||
@@ -547,7 +539,7 @@ retry2:		cp = jc->j_workcurs[i];
 						jc->j_workcurs[j] = NULL;
 					else
 						/* Partial reset suffices. */
-						if ((__db_c_dup(
+						if ((__dbc_dup(
 						    jc->j_fdupcurs[j],
 						    &jc->j_workcurs[j],
 						    DB_POSITION)) != 0)
@@ -567,12 +559,12 @@ retry2:		cp = jc->j_workcurs[i];
 			    jc->j_workcurs[j] != NULL;
 			    j++) {
 				if ((ret =
-				    __db_c_close(jc->j_workcurs[j])) != 0)
+				    __dbc_close(jc->j_workcurs[j])) != 0)
 					goto err;
 				jc->j_exhausted[j] = 0;
 				if (jc->j_fdupcurs[j] == NULL)
 					jc->j_workcurs[j] = NULL;
-				else if ((ret = __db_c_dup(jc->j_fdupcurs[j],
+				else if ((ret = __dbc_dup(jc->j_fdupcurs[j],
 				    &jc->j_workcurs[j], DB_POSITION)) != 0)
 					goto err;
 			}
@@ -582,9 +574,9 @@ retry2:		cp = jc->j_workcurs[i];
 
 		if (ret == DB_BUFFER_SMALL) {
 			jc->j_key.ulen <<= 1;
-			if ((ret = __os_realloc(dbp->dbenv, jc->j_key.ulen,
+			if ((ret = __os_realloc(env, jc->j_key.ulen,
 			    &jc->j_key.data)) != 0) {
-mem_err:			__db_err(dbp->dbenv,
+mem_err:			__db_errx(env,
 				    "Allocation failed for join key, len = %lu",
 				    (u_long)jc->j_key.ulen);
 				goto err;
@@ -618,7 +610,7 @@ mem_err:			__db_err(dbp->dbenv,
 		 * duplicate duplicates;  store this into jc->j_fdupcurs[i].
 		 */
 		if (SORTED_SET(jc, i) && jc->j_fdupcurs[i] == NULL && (ret =
-		    __db_c_dup(cp, &jc->j_fdupcurs[i], DB_POSITION)) != 0)
+		    __dbc_dup(cp, &jc->j_fdupcurs[i], DB_POSITION)) != 0)
 			goto err;
 	}
 
@@ -630,7 +622,7 @@ samekey:	/*
 		 * Get the key we tried and failed to return last time;
 		 * it should be the current datum of all the secondary cursors.
 		 */
-		if ((ret = __db_c_get(jc->j_workcurs[0],
+		if ((ret = __dbc_get(jc->j_workcurs[0],
 		    &jc->j_key, key_n, DB_CURRENT | opmods)) != 0)
 			return (ret);
 		F_CLR(jc, JOIN_RETRY);
@@ -643,11 +635,12 @@ samekey:	/*
 	 * back into the dbt we were given for the key; call __db_retcopy.
 	 * Otherwise, assert that we do not need to copy anything and proceed.
 	 */
-	DB_ASSERT(F_ISSET(
-	    key_arg, DB_DBT_USERMEM | DB_DBT_MALLOC) || key_n == key_arg);
+	DB_ASSERT(env, F_ISSET(key_arg, DB_DBT_USERMEM | DB_DBT_MALLOC |
+	    DB_DBT_USERCOPY) || key_n == key_arg);
 
-	if (F_ISSET(key_arg, DB_DBT_USERMEM | DB_DBT_MALLOC) &&
-	    (ret = __db_retcopy(dbp->dbenv,
+	if ((F_ISSET(key_arg, DB_DBT_USERMEM | DB_DBT_MALLOC |
+	    DB_DBT_USERCOPY)) &&
+	    (ret = __db_retcopy(env,
 	    key_arg, key_n->data, key_n->size, NULL, NULL)) != 0) {
 		/*
 		 * The retcopy failed, most commonly because we have a user
@@ -661,9 +654,6 @@ samekey:	/*
 	/*
 	 * If DB_JOIN_ITEM is set, we return it; otherwise we do the lookup
 	 * in the primary and then return.
-	 *
-	 * Note that we use key_arg here;  it is safe (and appropriate)
-	 * to do so.
 	 */
 	if (operation == DB_JOIN_ITEM)
 		return (0);
@@ -678,14 +668,19 @@ samekey:	/*
 	 * Instead, use memory that is managed by the join cursor, in
 	 * jc->j_rdata.
 	 */
-	if (!F_ISSET(data_arg, DB_DBT_MALLOC | DB_DBT_REALLOC | DB_DBT_USERMEM))
+	if (!F_ISSET(data_arg, DB_DBT_MALLOC | DB_DBT_REALLOC |
+	    DB_DBT_USERMEM | DB_DBT_USERCOPY))
 		db_manage_data = 1;
 	else
 		db_manage_data = 0;
-	if ((ret = __db_join_primget(jc->j_primary,
-	    jc->j_curslist[0]->txn, jc->j_curslist[0]->locker, key_arg,
+	if ((ret = __db_join_primget(jc->j_primary, dbc->thread_info,
+	    jc->j_curslist[0]->txn, jc->j_curslist[0]->locker, key_n,
 	    db_manage_data ? &jc->j_rdata : data_arg, opmods)) != 0) {
-		if (ret == DB_NOTFOUND)
+		if (ret == DB_NOTFOUND) {
+			if (LF_ISSET(DB_READ_UNCOMMITTED) ||
+			    (jc->j_curslist[0]->txn != NULL && F_ISSET(
+			    jc->j_curslist[0]->txn, TXN_READ_UNCOMMITTED)))
+				goto retry;
 			/*
 			 * If ret == DB_NOTFOUND, the primary and secondary
 			 * are out of sync;  every item in each secondary
@@ -694,7 +689,7 @@ samekey:	/*
 			 * Wail.
 			 */
 			ret = __db_secondary_corrupt(jc->j_primary);
-		else
+		} else
 			/*
 			 * The get on the primary failed for some other
 			 * reason, most commonly because we're using a user
@@ -713,7 +708,7 @@ samekey:	/*
 
 /*
  * __db_join_close --
- *	DBC->c_close for join cursors.
+ *	DBC->close for join cursors.
  *
  * PUBLIC: int __db_join_close __P((DBC *));
  */
@@ -722,14 +717,15 @@ __db_join_close(dbc)
 	DBC *dbc;
 {
 	DB *dbp;
-	DB_ENV *dbenv;
+	DB_THREAD_INFO *ip;
+	ENV *env;
 	JOIN_CURSOR *jc;
 	int ret, t_ret;
 	u_int32_t i;
 
 	jc = (JOIN_CURSOR *)dbc->internal;
 	dbp = dbc->dbp;
-	dbenv = dbp->dbenv;
+	env = dbp->env;
 	ret = t_ret = 0;
 
 	/*
@@ -737,12 +733,11 @@ __db_join_close(dbc)
 	 * must happen before any action that can fail and return, or else
 	 * __db_close may loop indefinitely.
 	 */
-	MUTEX_LOCK(dbenv, dbp->mutex);
+	MUTEX_LOCK(env, dbp->mutex);
 	TAILQ_REMOVE(&dbp->join_queue, dbc, links);
-	MUTEX_UNLOCK(dbenv, dbp->mutex);
-
-	PANIC_CHECK(dbenv);
+	MUTEX_UNLOCK(env, dbp->mutex);
 
+	ENV_ENTER(env, ip);
 	/*
 	 * Close any open scratch cursors.  In each case, there may
 	 * not be as many outstanding as there are cursors in
@@ -756,22 +751,23 @@ __db_join_close(dbc)
 	 */
 	for (i = 0; i < jc->j_ncurs; i++) {
 		if (jc->j_workcurs[i] != NULL &&
-		    (t_ret = __db_c_close(jc->j_workcurs[i])) != 0)
+		    (t_ret = __dbc_close(jc->j_workcurs[i])) != 0)
 			ret = t_ret;
 		if (jc->j_fdupcurs[i] != NULL &&
-		    (t_ret = __db_c_close(jc->j_fdupcurs[i])) != 0)
+		    (t_ret = __dbc_close(jc->j_fdupcurs[i])) != 0)
 			ret = t_ret;
 	}
+	ENV_LEAVE(env, ip);
 
-	__os_free(dbenv, jc->j_exhausted);
-	__os_free(dbenv, jc->j_curslist);
-	__os_free(dbenv, jc->j_workcurs);
-	__os_free(dbenv, jc->j_fdupcurs);
-	__os_free(dbenv, jc->j_key.data);
+	__os_free(env, jc->j_exhausted);
+	__os_free(env, jc->j_curslist);
+	__os_free(env, jc->j_workcurs);
+	__os_free(env, jc->j_fdupcurs);
+	__os_free(env, jc->j_key.data);
 	if (jc->j_rdata.data != NULL)
-		__os_ufree(dbenv, jc->j_rdata.data);
-	__os_free(dbenv, jc);
-	__os_free(dbenv, dbc);
+		__os_ufree(env, jc->j_rdata.data);
+	__os_free(env, jc);
+	__os_free(env, dbc);
 
 	return (ret);
 }
@@ -813,7 +809,7 @@ __db_join_getnext(dbc, key, data, exhausted, opmods)
 		 */
 		memset(&ldata, 0, sizeof(DBT));
 		F_SET(&ldata, DB_DBT_MALLOC);
-		if ((ret = __db_c_get(dbc,
+		if ((ret = __dbc_get(dbc,
 		    key, &ldata, opmods | DB_CURRENT)) != 0)
 			break;
 		cmp = func(dbp, data, &ldata);
@@ -823,10 +819,10 @@ __db_join_getnext(dbc, key, data, exhausted, opmods)
 			 * it into data, then free the buffer we malloc'ed
 			 * above.
 			 */
-			if ((ret = __db_retcopy(dbp->dbenv, data, ldata.data,
+			if ((ret = __db_retcopy(dbp->env, data, ldata.data,
 			    ldata.size, &data->data, &data->size)) != 0)
 				return (ret);
-			__os_ufree(dbp->dbenv, ldata.data);
+			__os_ufree(dbp->env, ldata.data);
 			return (0);
 		}
 
@@ -835,10 +831,10 @@ __db_join_getnext(dbc, key, data, exhausted, opmods)
 		 * dups.  We just forget about ldata and free
 		 * its buffer--data contains the value we're searching for.
 		 */
-		__os_ufree(dbp->dbenv, ldata.data);
+		__os_ufree(dbp->env, ldata.data);
 		/* FALLTHROUGH */
 	case 1:
-		ret = __db_c_get(dbc, key, data, opmods | DB_GET_BOTHC);
+		ret = __dbc_get(dbc, key, data, opmods | DB_GET_BOTHC);
 		break;
 	default:
 		ret = EINVAL;
@@ -862,8 +858,8 @@ __db_join_cmp(a, b)
 	dbca = *((DBC * const *)a);
 	dbcb = *((DBC * const *)b);
 
-	if (__db_c_count(dbca, &counta) != 0 ||
-	    __db_c_count(dbcb, &countb) != 0)
+	if (__dbc_count(dbca, &counta) != 0 ||
+	    __dbc_count(dbcb, &countb) != 0)
 		return (0);
 
 	return ((long)counta - (long)countb);
@@ -875,10 +871,11 @@ __db_join_cmp(a, b)
  * locker ID if we're doing CDB locking.
  */
 static int
-__db_join_primget(dbp, txn, lockerid, key, data, flags)
+__db_join_primget(dbp, ip, txn, locker, key, data, flags)
 	DB *dbp;
+	DB_THREAD_INFO *ip;
 	DB_TXN *txn;
-	u_int32_t lockerid;
+	DB_LOCKER  *locker;
 	DBT *key, *data;
 	u_int32_t flags;
 {
@@ -886,8 +883,8 @@ __db_join_primget(dbp, txn, lockerid, key, data, flags)
 	u_int32_t rmw;
 	int ret, t_ret;
 
-	if ((ret = __db_cursor_int(dbp,
-	    txn, dbp->type, PGNO_INVALID, 0, lockerid, &dbc)) != 0)
+	if ((ret = __db_cursor_int(dbp, ip,
+	    txn, dbp->type, PGNO_INVALID, 0, locker, &dbc)) != 0)
 		return (ret);
 
 	/*
@@ -906,7 +903,7 @@ __db_join_primget(dbp, txn, lockerid, key, data, flags)
 		F_SET(dbc, DBC_READ_COMMITTED);
 
 	LF_CLR(DB_READ_COMMITTED | DB_READ_UNCOMMITTED | DB_RMW);
-	DB_ASSERT(flags == 0);
+	DB_ASSERT(dbp->env, flags == 0);
 
 	F_SET(dbc, DBC_TRANSIENT);
 
@@ -917,9 +914,9 @@ __db_join_primget(dbp, txn, lockerid, key, data, flags)
 	 */
 	SET_RET_MEM(dbc, dbp);
 
-	ret = __db_c_get(dbc, key, data, DB_SET | rmw);
+	ret = __dbc_get(dbc, key, data, DB_SET | rmw);
 
-	if ((t_ret = __db_c_close(dbc)) != 0 && ret == 0)
+	if ((t_ret = __dbc_close(dbc)) != 0 && ret == 0)
 		ret = t_ret;
 
 	return (ret);
@@ -927,8 +924,7 @@ __db_join_primget(dbp, txn, lockerid, key, data, flags)
 
 /*
  * __db_secondary_corrupt --
- *	Report that a secondary index appears corrupt, as it has a record
- * that does not correspond to a record in the primary or vice versa.
+ *	Report primary/secondary inconsistencies.
  *
  * PUBLIC: int __db_secondary_corrupt __P((DB *));
  */
@@ -936,7 +932,9 @@ int
 __db_secondary_corrupt(dbp)
 	DB *dbp;
 {
-	__db_err(dbp->dbenv,
-	    "Secondary index corrupt: not consistent with primary");
+	__db_err(dbp->env, DB_SECONDARY_BAD, "%s%s%s",
+	    dbp->fname == NULL ? "unnamed" : dbp->fname,
+	    dbp->dname == NULL ? "" : "/",
+	    dbp->dname == NULL ? "" : dbp->dname);
 	return (DB_SECONDARY_BAD);
 }
