@@ -1,8 +1,7 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2005
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1996-2009 Oracle.  All rights reserved.
  */
 /*
  * Copyright (c) 1990, 1993, 1994, 1995, 1996
@@ -39,28 +38,22 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- * $Id: db_meta.c,v 12.22 2005/10/27 01:46:34 bostic Exp $
+ * $Id$
  */
 
 #include "db_config.h"
 
-#ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-
-#include <stdlib.h>
-#include <string.h>
-#endif
-
 #include "db_int.h"
 #include "dbinc/db_page.h"
-#include "dbinc/db_shash.h"
 #include "dbinc/lock.h"
+#include "dbinc/log.h"
 #include "dbinc/mp.h"
+#include "dbinc/txn.h"
 #include "dbinc/db_am.h"
+#include "dbinc/hash.h"
 
 static void __db_init_meta __P((DB *, void *, db_pgno_t, u_int32_t));
 #ifdef HAVE_FTRUNCATE
-static void __db_freelist_sort __P((struct pglist *, u_int32_t));
 static int  __db_pglistcmp __P((const void *, const void *));
 static int  __db_truncate_freelist __P((DBC *, DBMETA *,
       PAGE *, db_pgno_t *, u_int32_t, u_int32_t));
@@ -79,8 +72,8 @@ __db_init_meta(dbp, p, pgno, pgtype)
 	db_pgno_t pgno;
 	u_int32_t pgtype;
 {
-	DB_LSN save_lsn;
 	DBMETA *meta;
+	DB_LSN save_lsn;
 
 	meta = (DBMETA *)p;
 	save_lsn = meta->lsn;
@@ -97,37 +90,56 @@ __db_init_meta(dbp, p, pgno, pgtype)
  * __db_new --
  *	Get a new page, preferably from the freelist.
  *
- * PUBLIC: int __db_new __P((DBC *, u_int32_t, PAGE **));
+ * PUBLIC: int __db_new __P((DBC *, u_int32_t, DB_LOCK *, PAGE **));
  */
 int
-__db_new(dbc, type, pagepp)
+__db_new(dbc, type, lockp, pagepp)
 	DBC *dbc;
 	u_int32_t type;
+	DB_LOCK *lockp;
 	PAGE **pagepp;
 {
-	DBMETA *meta;
 	DB *dbp;
+	DBMETA *meta;
 	DB_LOCK metalock;
 	DB_LSN lsn;
 	DB_MPOOLFILE *mpf;
+	ENV *env;
 	PAGE *h;
 	db_pgno_t last, *list, pgno, newnext;
-	u_int32_t meta_flags;
-	int extend, ret, t_ret;
+	int extend, hash, ret, t_ret;
 
 	meta = NULL;
-	meta_flags = 0;
 	dbp = dbc->dbp;
+	env = dbp->env;
 	mpf = dbp->mpf;
 	h = NULL;
 	newnext = PGNO_INVALID;
+	if (lockp != NULL)
+		LOCK_INIT(*lockp);
+
+	hash = 0;
+	ret = 0;
+	LOCK_INIT(metalock);
+
+#ifdef HAVE_HASH
+	if (dbp->type == DB_HASH) {
+		if ((ret = __ham_return_meta(dbc, DB_MPOOL_DIRTY, &meta)) != 0)
+			goto err;
+		if (meta != NULL)
+			hash = 1;
+	}
+#endif
+	if (meta == NULL) {
+		pgno = PGNO_BASE_MD;
+		if ((ret = __db_lget(dbc,
+		    LCK_ALWAYS, pgno, DB_LOCK_WRITE, 0, &metalock)) != 0)
+			goto err;
+		if ((ret = __memp_fget(mpf, &pgno, dbc->thread_info, dbc->txn,
+		    DB_MPOOL_DIRTY, &meta)) != 0)
+			goto err;
+	}
 
-	pgno = PGNO_BASE_MD;
-	if ((ret = __db_lget(dbc,
-	    LCK_ALWAYS, pgno, DB_LOCK_WRITE, 0, &metalock)) != 0)
-		goto err;
-	if ((ret = __memp_fget(mpf, &pgno, 0, &meta)) != 0)
-		goto err;
 	last = meta->last_pgno;
 	if (meta->free == PGNO_INVALID) {
 		if (FLD_ISSET(type, P_DONTEXTEND)) {
@@ -139,7 +151,18 @@ __db_new(dbc, type, pagepp)
 		extend = 1;
 	} else {
 		pgno = meta->free;
-		if ((ret = __memp_fget(mpf, &pgno, 0, &h)) != 0)
+		/*
+		 * Lock the new page.  Do this here because we must do it
+		 * before getting the page and the caller may need the lock
+		 * to keep readers from seeing the page before the transaction
+		 * commits.  We can do this because no one will hold a free
+		 * page locked.
+		 */
+		if (lockp != NULL && (ret =
+		     __db_lget(dbc, 0, pgno, DB_LOCK_WRITE, 0, lockp)) != 0)
+			goto err;
+		if ((ret = __memp_fget(mpf, &pgno, dbc->thread_info, dbc->txn,
+		    DB_MPOOL_DIRTY, &h)) != 0)
 			goto err;
 
 		/*
@@ -150,6 +173,15 @@ __db_new(dbc, type, pagepp)
 		newnext = h->next_pgno;
 		lsn = h->lsn;
 		extend = 0;
+		DB_ASSERT(env, TYPE(h) == P_INVALID);
+
+		if (TYPE(h) != P_INVALID) {
+			__db_errx(env,
+			    "%s page %lu is on free list with type %lu",
+				dbp->fname, (u_long)PGNO(h), (u_long)TYPE(h));
+			return (__env_panic(env, EINVAL));
+		}
+
 	}
 
 	FLD_CLR(type, P_DONTEXTEND);
@@ -167,25 +199,24 @@ __db_new(dbc, type, pagepp)
 	} else
 		LSN_NOT_LOGGED(LSN(meta));
 
-	meta_flags = DB_MPOOL_DIRTY;
 	meta->free = newnext;
 
 	if (extend == 1) {
-		if ((ret = __memp_fget(mpf, &pgno, DB_MPOOL_NEW, &h)) != 0)
+		if (lockp != NULL && (ret =
+		     __db_lget(dbc, 0, pgno, DB_LOCK_WRITE, 0, lockp)) != 0)
 			goto err;
-		DB_ASSERT(last == pgno);
+		if ((ret = __memp_fget(mpf, &pgno, dbc->thread_info, dbc->txn,
+		    DB_MPOOL_NEW, &h)) != 0)
+			goto err;
+		DB_ASSERT(env, last == pgno);
 		meta->last_pgno = pgno;
 		ZERO_LSN(h->lsn);
 		h->pgno = pgno;
 	}
 	LSN(h) = LSN(meta);
 
-	DB_ASSERT(TYPE(h) == P_INVALID);
-
-	if (TYPE(h) != P_INVALID)
-		return (__db_panic(dbp->dbenv, EINVAL));
-
-	ret = __memp_fput(mpf, (PAGE *)meta, DB_MPOOL_DIRTY);
+	if (hash == 0)
+		ret = __memp_fput(mpf, dbc->thread_info, meta, dbc->priority);
 	meta = NULL;
 	if ((t_ret = __TLPUT(dbc, metalock)) != 0 && ret == 0)
 		ret = t_ret;
@@ -212,7 +243,7 @@ __db_new(dbc, type, pagepp)
 		if ((ret = __memp_get_freelist(dbp->mpf, &nelems, &list)) != 0)
 			goto err;
 		if (nelems != 0) {
-			DB_ASSERT(h->pgno == list[0]);
+			DB_ASSERT(env, h->pgno == list[0]);
 			memmove(list, &list[1], (nelems - 1) * sizeof(*list));
 			if ((ret = __memp_extend_freelist(
 			    dbp->mpf, nelems - 1, &list)) != 0)
@@ -223,29 +254,16 @@ __db_new(dbc, type, pagepp)
 	COMPQUIET(list, NULL);
 #endif
 
-	/*
-	 * If dirty reads are enabled and we are in a transaction, we could
-	 * abort this allocation after the page(s) pointing to this
-	 * one have their locks downgraded.  This would permit dirty readers
-	 * to access this page which is ok, but they must be off the
-	 * page when we abort.  We never lock overflow pages or off page
-	 * duplicate trees.
-	 */
-	if (type != P_OVERFLOW && !F_ISSET(dbc, DBC_OPD) &&
-	     F_ISSET(dbc->dbp, DB_AM_READ_UNCOMMITTED) && dbc->txn != NULL) {
-		if ((ret = __db_lget(dbc, 0,
-		    h->pgno, DB_LOCK_WWRITE, 0, &metalock)) != 0)
-			goto err;
-	}
-
 	*pagepp = h;
 	return (0);
 
 err:	if (h != NULL)
-		(void)__memp_fput(mpf, h, 0);
-	if (meta != NULL)
-		(void)__memp_fput(mpf, meta, meta_flags);
+		(void)__memp_fput(mpf, dbc->thread_info, h, dbc->priority);
+	if (meta != NULL && hash == 0)
+		(void)__memp_fput(mpf, dbc->thread_info, meta, dbc->priority);
 	(void)__TLPUT(dbc, metalock);
+	if (lockp != NULL)
+		(void)__LPUT(dbc, *lockp);
 	return (ret);
 }
 
@@ -260,26 +278,33 @@ __db_free(dbc, h)
 	DBC *dbc;
 	PAGE *h;
 {
-	DBMETA *meta;
 	DB *dbp;
+	DBMETA *meta;
 	DBT ddbt, ldbt;
 	DB_LOCK metalock;
+	DB_LSN *lsnp;
 	DB_MPOOLFILE *mpf;
-	db_pgno_t last_pgno, *lp, next_pgno, pgno, prev_pgno;
-	u_int32_t dirty_flag, lflag, nelem;
-	int do_truncate, ret, t_ret;
+	PAGE *prev;
+	db_pgno_t last_pgno, next_pgno, pgno, prev_pgno;
+	u_int32_t lflag;
+	int hash, ret, t_ret;
 #ifdef HAVE_FTRUNCATE
-	db_pgno_t *list;
-	u_int32_t position, start;
+	db_pgno_t *list, *lp;
+	u_int32_t nelem, position, start;
+	int do_truncate;
 #endif
 
 	dbp = dbc->dbp;
 	mpf = dbp->mpf;
 	prev_pgno = PGNO_INVALID;
-	nelem = 0;
 	meta = NULL;
-	do_truncate = 0;
+	prev = NULL;
+	LOCK_INIT(metalock);
+#ifdef HAVE_FTRUNCATE
 	lp = NULL;
+	nelem = 0;
+	do_truncate = 0;
+#endif
 
 	/*
 	 * Retrieve the metadata page.  If we are not keeping a sorted
@@ -291,18 +316,48 @@ __db_free(dbc, h)
 	 * fail, then we need to put the page with which we were called
 	 * back because our caller assumes we take care of it.
 	 */
-	dirty_flag = 0;
+	hash = 0;
+
 	pgno = PGNO_BASE_MD;
-	if ((ret = __db_lget(dbc,
-	    LCK_ALWAYS, pgno, DB_LOCK_WRITE, 0, &metalock)) != 0)
-		goto err;
-	if ((ret = __memp_fget(mpf, &pgno, 0, &meta)) != 0)
-		goto err1;
+#ifdef HAVE_HASH
+	if (dbp->type == DB_HASH) {
+		if ((ret = __ham_return_meta(dbc,
+#ifdef HAVE_FTRUNCATE
+		    0,
+#else
+		    DB_MPOOL_DIRTY,
+#endif
+		&meta)) != 0)
+			goto err;
+		if (meta != NULL)
+			hash = 1;
+	}
+#endif
+	if (meta == NULL) {
+		if ((ret = __db_lget(dbc,
+		    LCK_ALWAYS, pgno, DB_LOCK_WRITE, 0, &metalock)) != 0)
+			goto err;
+
+		/* If we support truncate, we might not dirty the meta page. */
+		if ((ret = __memp_fget(mpf, &pgno, dbc->thread_info, dbc->txn,
+#ifdef HAVE_FTRUNCATE
+		    0,
+#else
+		    DB_MPOOL_DIRTY,
+#endif
+		    &meta)) != 0)
+			goto err1;
+	}
 
 	last_pgno = meta->last_pgno;
 	next_pgno = meta->free;
+	/*
+	 * Assign lsnp here so it always initialized when
+	 * HAVE_FTRUNCATE is not defined.
+	 */
+	lsnp = &LSN(meta);
 
-	DB_ASSERT(h->pgno != next_pgno);
+	DB_ASSERT(dbp->env, h->pgno != next_pgno);
 
 #ifdef HAVE_FTRUNCATE
 	/*
@@ -312,7 +367,7 @@ __db_free(dbc, h)
 	 * meta page and get the previous page.
 	 */
 	if ((ret = __memp_get_freelist(mpf, &nelem, &list)) != 0)
-		goto err;
+		goto err1;
 	if (list == NULL)
 		goto no_sort;
 
@@ -326,24 +381,13 @@ __db_free(dbc, h)
 		if (nelem != 0) {
 			__db_freelist_pos(h->pgno, list, nelem, &position);
 
-			DB_ASSERT(h->pgno != list[position]);
+			DB_ASSERT(dbp->env, h->pgno != list[position]);
 
 			/* Get the previous page if this is not the smallest. */
 			if (position != 0 || h->pgno > list[0])
 				prev_pgno = list[position];
 		}
 
-		/* Put the page number into the list. */
-		if ((ret = __memp_extend_freelist(mpf, nelem + 1, &list)) != 0)
-			return (ret);
-		if (prev_pgno != PGNO_INVALID)
-			lp = &list[position + 1];
-		else
-			lp = list;
-		if (nelem != 0 && position != nelem)
-			memmove(lp + 1, lp,
-			    (size_t)((u_int8_t*)&list[nelem] - (u_int8_t*)lp));
-		*lp = h->pgno;
 	} else if (nelem != 0) {
 		/* Find the truncation point. */
 		for (lp = &list[nelem - 1]; lp >= list; lp--)
@@ -355,22 +399,53 @@ __db_free(dbc, h)
 	}
 
 no_sort:
-	if (prev_pgno != PGNO_INVALID) {
-		if ((ret = __memp_fput(mpf, meta, 0)) != 0)
+	if (prev_pgno == PGNO_INVALID) {
+#ifdef HAVE_HASH
+		if (hash) {
+			if ((ret =
+			    __ham_return_meta(dbc, DB_MPOOL_DIRTY, &meta)) != 0)
+				goto err1;
+		} else
+#endif
+		if ((ret = __memp_dirty(mpf,
+		    &meta, dbc->thread_info, dbc->txn, dbc->priority, 0)) != 0)
 			goto err1;
-		meta = NULL;
+		lsnp = &LSN(meta);
+	} else {
 		pgno = prev_pgno;
-		if ((ret = __memp_fget(mpf, &pgno, 0, &meta)) != 0)
+		if ((ret = __memp_fget(mpf, &pgno,
+		    dbc->thread_info, dbc->txn, DB_MPOOL_DIRTY, &prev)) != 0)
 			goto err1;
-		next_pgno = NEXT_PGNO(meta);
+		next_pgno = NEXT_PGNO(prev);
+		lsnp = &LSN(prev);
 	}
 #endif
 
-	/* Log the change. */
+	/*
+	 * Log the change.
+	 *	We are either logging an update to the metapage or to the
+	 * previous page in the sorted list.
+	 */
 	if (DBC_LOGGING(dbc)) {
 		memset(&ldbt, 0, sizeof(ldbt));
 		ldbt.data = h;
 		ldbt.size = P_OVERHEAD(dbp);
+		/*
+		 * If we are truncating the file, we need to make sure
+		 * the logging happens before the truncation.  If we
+		 * are truncating multiple pages we don't need to flush the
+		 * log here as it will be flushed by __db_truncate_freelist.
+		 * If we are zeroing pages rather than truncating we still
+		 * need to flush since they will not have valid LSNs.
+		 */
+		lflag = 0;
+
+		if (h->pgno == last_pgno
+#ifdef HAVE_FTRUNCATE
+		    && do_truncate == 0
+#endif
+		)
+			lflag = DB_FLUSH;
 		switch (h->type) {
 		case P_HASH:
 		case P_IBTREE:
@@ -383,7 +458,8 @@ no_sort:
 				ddbt.data = (u_int8_t *)h + HOFFSET(h);
 				ddbt.size = dbp->pgsize - HOFFSET(h);
 				if ((ret = __db_pg_freedata_log(dbp, dbc->txn,
-				     &LSN(meta), 0, h->pgno, &LSN(meta), pgno,
+				     lsnp, lflag,
+				     h->pgno, lsnp, pgno,
 				     &ldbt, next_pgno, last_pgno, &ddbt)) != 0)
 					goto err1;
 				goto logged;
@@ -399,28 +475,17 @@ no_sort:
 			ldbt.size += OV_LEN(h);
 			break;
 		default:
-			DB_ASSERT(h->type != P_QAMDATA);
+			DB_ASSERT(dbp->env, h->type != P_QAMDATA);
 		}
 
-		/*
-		 * If we are truncating the file, we need to make sure
-		 * the logging happens before the truncation.  If we
-		 * are truncating multiple pages we don't need to flush the
-		 * log here as it will be flushed by __db_truncate_freelist.
-		 */
-		lflag = 0;
-#ifdef HAVE_FTRUNCATE
-		if (do_truncate == 0 && h->pgno == last_pgno)
-			lflag = DB_FLUSH;
-#endif
 		if ((ret = __db_pg_free_log(dbp,
-		      dbc->txn, &LSN(meta), lflag, h->pgno,
-		      &LSN(meta), pgno, &ldbt, next_pgno, last_pgno)) != 0)
+		      dbc->txn, lsnp, lflag, h->pgno,
+		      lsnp, pgno, &ldbt, next_pgno, last_pgno)) != 0)
 			goto err1;
 	} else
-		LSN_NOT_LOGGED(LSN(meta));
-logged:	LSN(h) = LSN(meta);
+		LSN_NOT_LOGGED(*lsnp);
 
+logged:
 #ifdef HAVE_FTRUNCATE
 	if (do_truncate) {
 		start = (u_int32_t) (lp - list) + 1;
@@ -428,24 +493,57 @@ logged:	LSN(h) = LSN(meta);
 		ret = __db_truncate_freelist(
 		      dbc, meta, h, list, start, nelem);
 		h = NULL;
-	} else if (h->pgno == last_pgno) {
-		if ((ret = __memp_fput(mpf, h, DB_MPOOL_DISCARD)) != 0)
-			goto err;
-		/* Give the page back to the OS. */
-		if ((ret = __memp_ftruncate(mpf, last_pgno, 0)) != 0)
-			goto err;
-		DB_ASSERT(meta->pgno == PGNO_BASE_MD);
-		meta->last_pgno--;
-		h = NULL;
 	} else
 #endif
-
-	{
+	if (h->pgno == last_pgno) {
+		/*
+		 * We are going to throw this page away, but if we are
+		 * using MVCC then this version may stick around and we
+		 * might have to make a copy.
+		 */
+		if (mpf->mfp->multiversion && (ret = __memp_dirty(mpf,
+		    &h, dbc->thread_info, dbc->txn, dbc->priority, 0)) != 0)
+			goto err1;
+		LSN(h) = *lsnp;
+		P_INIT(h, dbp->pgsize,
+		    h->pgno, PGNO_INVALID, next_pgno, 0, P_INVALID);
+		if ((ret = __memp_fput(mpf,
+		    dbc->thread_info, h, DB_PRIORITY_VERY_LOW)) != 0)
+			goto err1;
+		h = NULL;
+		/* Give the page back to the OS. */
+		if ((ret = __memp_ftruncate(mpf, dbc->txn, dbc->thread_info,
+		    last_pgno, 0)) != 0)
+			goto err1;
+		DB_ASSERT(dbp->env, meta->pgno == PGNO_BASE_MD);
+		meta->last_pgno--;
+		h = NULL;
+	} else {
+#ifdef HAVE_FTRUNCATE
+		if (list != NULL) {
+			/* Put the page number into the list. */
+			if ((ret =
+			    __memp_extend_freelist(mpf, nelem + 1, &list)) != 0)
+				goto err1;
+			if (prev_pgno != PGNO_INVALID)
+				lp = &list[position + 1];
+			else
+				lp = list;
+			if (nelem != 0 && position != nelem)
+				memmove(lp + 1, lp, (size_t)
+				    ((u_int8_t*)&list[nelem] - (u_int8_t*)lp));
+			*lp = h->pgno;
+		}
+#endif
 		/*
 		 * If we are not truncating the page then we
 		 * reinitialize it and put it at the head of
 		 * the free list.
 		 */
+		if ((ret = __memp_dirty(mpf,
+		    &h, dbc->thread_info, dbc->txn, dbc->priority, 0)) != 0)
+			goto err1;
+		LSN(h) = *lsnp;
 		P_INIT(h, dbp->pgsize,
 		    h->pgno, PGNO_INVALID, next_pgno, 0, P_INVALID);
 #ifdef DIAGNOSTIC
@@ -455,20 +553,22 @@ logged:	LSN(h) = LSN(meta);
 		if (prev_pgno == PGNO_INVALID)
 			meta->free = h->pgno;
 		else
-			NEXT_PGNO(meta) = h->pgno;
+			NEXT_PGNO(prev) = h->pgno;
 	}
 
 	/* Discard the metadata or previous page. */
-err1:	if (meta != NULL && (t_ret =
-	    __memp_fput(mpf, (PAGE *)meta, DB_MPOOL_DIRTY)) != 0 && ret == 0)
+err1:	if (hash == 0 && meta != NULL && (t_ret = __memp_fput(mpf,
+	    dbc->thread_info, (PAGE *)meta, dbc->priority)) != 0 && ret == 0)
 		ret = t_ret;
 	if ((t_ret = __TLPUT(dbc, metalock)) != 0 && ret == 0)
 		ret = t_ret;
+	if (prev != (PAGE*) meta && prev != NULL && (t_ret = __memp_fput(mpf,
+	    dbc->thread_info, prev, dbc->priority)) != 0 && ret == 0)
+		ret = t_ret;
 
 	/* Discard the caller's page reference. */
-	dirty_flag = DB_MPOOL_DIRTY;
-err:	if (h != NULL &&
-	    (t_ret = __memp_fput(mpf, h, dirty_flag)) != 0 && ret == 0)
+err:	if (h != NULL && (t_ret = __memp_fput(mpf,
+	    dbc->thread_info, h, dbc->priority)) != 0 && ret == 0)
 		ret = t_ret;
 
 	/*
@@ -519,65 +619,63 @@ static int
 __db_pglistcmp(a, b)
 	const void *a, *b;
 {
-	struct pglist *ap, *bp;
+	db_pglist_t *ap, *bp;
 
-	ap = (struct pglist *)a;
-	bp = (struct pglist *)b;
+	ap = (db_pglist_t *)a;
+	bp = (db_pglist_t *)b;
 
 	return ((ap->pgno > bp->pgno) ? 1 : (ap->pgno < bp->pgno) ? -1: 0);
 }
 
 /*
  * __db_freelist_sort -- sort a list of free pages.
+ * PUBLIC: void __db_freelist_sort __P((db_pglist_t *, u_int32_t));
  */
-static void
+void
 __db_freelist_sort(list, nelems)
-	struct pglist *list;
+	db_pglist_t *list;
 	u_int32_t nelems;
 {
-	qsort(list, (size_t)nelems, sizeof(struct pglist), __db_pglistcmp);
+	qsort(list, (size_t)nelems, sizeof(db_pglist_t), __db_pglistcmp);
 }
 
 /*
- * __db_pg_truncate -- sort the freelist and find the truncation point.
+ * __db_pg_truncate -- find the truncation point in a sorted freelist.
  *
  * PUBLIC: #ifdef HAVE_FTRUNCATE
- * PUBLIC: int __db_pg_truncate __P((DB_MPOOLFILE *, struct pglist *list,
- * PUBLIC:    DB_COMPACT *, u_int32_t *, db_pgno_t *, DB_LSN *, int));
+ * PUBLIC: int __db_pg_truncate __P((DBC *, DB_TXN *,
+ * PUBLIC:    db_pglist_t *, DB_COMPACT *, u_int32_t *, 
+ * PUBLIC:    db_pgno_t , db_pgno_t *, DB_LSN *, int));
  * PUBLIC: #endif
  */
 int
-__db_pg_truncate(mpf, list, c_data, nelemp, last_pgno, lsnp, in_recovery)
-	DB_MPOOLFILE *mpf;
-	struct pglist *list;
+__db_pg_truncate(dbc, txn,
+    list, c_data, nelemp, free_pgno, last_pgno, lsnp, in_recovery)
+	DBC *dbc;
+	DB_TXN *txn;
+	db_pglist_t *list;
 	DB_COMPACT *c_data;
 	u_int32_t *nelemp;
-	db_pgno_t *last_pgno;
+	db_pgno_t free_pgno, *last_pgno;
 	DB_LSN *lsnp;
 	int in_recovery;
 {
+	DB *dbp;
+	DBT ddbt;
+	DB_LSN null_lsn;
+	DB_MPOOLFILE *mpf;
 	PAGE *h;
-	struct pglist *lp;
-	db_pgno_t pgno;
-	u_int32_t nelems;
-	int modified, ret;
+	db_pglist_t *lp, *slp;
+	db_pgno_t lpgno, pgno;
+	u_int32_t elems, log_size, tpoint;
+	int last, ret;
 
 	ret = 0;
+	h = NULL;
 
-	nelems = *nelemp;
-	/* Sort the list */
-	__db_freelist_sort(list, nelems);
-
-	/* Find the truncation point. */
-	pgno = *last_pgno;
-	lp = &list[nelems - 1];
-	while (nelems != 0) {
-		if (lp->pgno != pgno)
-			break;
-		pgno--;
-		nelems--;
-		lp--;
-	}
+	dbp = dbc->dbp;
+	mpf = dbp->mpf;
+	elems = tpoint = *nelemp;
 
 	/*
 	 * Figure out what (if any) pages can be truncated immediately and
@@ -585,8 +683,74 @@ __db_pg_truncate(mpf, list, c_data, nelemp, last_pgno, lsnp, in_recovery)
 	 * memp_ftruncate below.  We also use this to avoid ever putting
 	 * these pages on the freelist, which we are about to relink.
 	 */
-	for (lp = list; lp < &list[nelems]; lp++) {
-		if ((ret = __memp_fget(mpf, &lp->pgno, 0, &h)) != 0) {
+	pgno = *last_pgno;
+	lp = &list[elems - 1];
+	last = 1;
+	while (tpoint != 0) {
+		if (lp->pgno != pgno)
+			break;
+		pgno--;
+		tpoint--;
+		lp--;
+	}
+
+	lp = list;
+	slp = &list[elems];
+	/*
+	 * Log the sorted list. We log the whole list so it can be rebuilt.
+	 * Don't overflow the log file.  
+	 */
+again:	if (DBC_LOGGING(dbc)) {
+		last = 1;
+		lpgno = *last_pgno;
+		ddbt.size = elems * sizeof(*lp);
+		ddbt.data = lp;
+		log_size = ((LOG *)dbc->env->
+		    lg_handle->reginfo.primary)->log_size;
+		if (ddbt.size > log_size / 2) {
+			elems = (log_size / 2) / sizeof(*lp);
+			ddbt.size = elems * sizeof(*lp);
+			last = 0;
+			/*
+			 * If we stopped after the truncation point
+			 * then we need to truncate from here.
+			 */
+			if (lp + elems >= &list[tpoint])
+				lpgno = lp[elems - 1].pgno;
+		}
+		/*
+		 * If this is not the begining of the list fetch the end
+		 * of the previous segment.  This page becomes the last_free
+		 * page and will link to this segment if it is not truncated.
+		 */
+		if (lp != list) {
+			if ((ret = __memp_fget(mpf, &lp[-1].pgno,
+			    dbc->thread_info, txn, 0, &h)) != 0) 
+			    	goto err;
+		}
+			
+		slp = &lp[elems];
+
+		ZERO_LSN(null_lsn);
+		if ((ret = __db_pg_trunc_log(dbp, dbc->txn,
+		     lsnp, last == 1 ? DB_FLUSH : 0, PGNO_BASE_MD,
+		     lsnp, h != NULL ? PGNO(h) : PGNO_INVALID,
+		     h != NULL ? &LSN(h) : &null_lsn,
+		     free_pgno, lpgno, &ddbt)) != 0)
+			goto err;
+		if (h != NULL) {
+			LSN(h) = *lsnp;
+			if ((ret = __memp_fput(mpf,
+			    dbc->thread_info, h, dbc->priority)) != 0)
+				goto err;
+		}
+		h = NULL;
+	} else if (!in_recovery)
+		LSN_NOT_LOGGED(*lsnp);
+
+	for (; lp < slp && lp < &list[tpoint]; lp++) {
+		if ((ret = __memp_fget(mpf, &lp->pgno, dbc->thread_info,
+		    txn, !in_recovery ? DB_MPOOL_DIRTY : 0, &h)) != 0) {
 			/* Page may have been truncated later. */
 			if (in_recovery && ret == DB_PAGE_NOTFOUND) {
 				ret = 0;
@@ -594,89 +758,121 @@ __db_pg_truncate(mpf, list, c_data, nelemp, last_pgno, lsnp, in_recovery)
 			}
 			goto err;
 		}
-		modified = 0;
-		if (!in_recovery || log_compare(&LSN(h), &lp->lsn) == 0) {
-			if (lp == &list[nelems - 1])
-				NEXT_PGNO(h) = PGNO_INVALID;
-			else
-				NEXT_PGNO(h) = lp[1].pgno;
-			DB_ASSERT(NEXT_PGNO(h) < *last_pgno);
-
-			LSN(h) = *lsnp;
-			modified = 1;
+		if (in_recovery) {
+			if (LOG_COMPARE(&LSN(h), &lp->lsn) == 0) {
+				if ((ret = __memp_dirty(mpf, &h,
+				    dbc->thread_info,
+				    txn, dbp->priority, 0)) != 0) {
+					(void)__memp_fput(mpf,
+					    dbc->thread_info, h, dbp->priority);
+					goto err;
+				}
+			} else 
+				goto skip;
 		}
-		if ((ret = __memp_fput(mpf, h,
-		    modified ? DB_MPOOL_DIRTY : 0)) != 0)
+
+		if (lp == &list[tpoint - 1])
+			NEXT_PGNO(h) = PGNO_INVALID;
+		else
+			NEXT_PGNO(h) = lp[1].pgno;
+		DB_ASSERT(mpf->env, NEXT_PGNO(h) < *last_pgno);
+
+		LSN(h) = *lsnp;
+skip:		if ((ret = __memp_fput(mpf,
+		    dbc->thread_info, h, dbp->priority)) != 0)
 			goto err;
+		h = NULL;
 	}
 
+	/*
+	 * If we did not log everything try again.  We start from slp and
+	 * try to go to the end of the list.
+	 */
+	if (last == 0) {
+		elems = (u_int32_t)(&list[*nelemp] - slp);
+		lp = slp;
+		goto again;
+	}
+
+	/*
+	 * Truncate the file.  Its possible that the last page is the
+	 * only one that got truncated and that's done in the caller.
+	 */
 	if (pgno != *last_pgno) {
-		if ((ret = __memp_ftruncate(mpf,
+		if (tpoint != *nelemp &&
+		    (ret = __memp_ftruncate(mpf, dbc->txn, dbc->thread_info,
 		    pgno + 1, in_recovery ? MP_TRUNC_RECOVER : 0)) != 0)
 			goto err;
 		if (c_data)
 			c_data->compact_pages_truncated += *last_pgno - pgno;
 		*last_pgno = pgno;
 	}
-	*nelemp = nelems;
+	*nelemp = tpoint;
 
-err:	return (ret);
+	if (0) {
+err:		if (h != NULL)
+			(void)__memp_fput(mpf,
+			    dbc->thread_info, h, dbc->priority);
+	}
+	return (ret);
 }
 
 /*
  * __db_free_truncate --
- *	Truncate free pages at the end of the file.
+ * 	  Build a sorted free list and truncate free pages at the end
+ *	  of the file.
  *
  * PUBLIC: #ifdef HAVE_FTRUNCATE
- * PUBLIC: int __db_free_truncate __P((DB *, DB_TXN *, u_int32_t,
- * PUBLIC:    DB_COMPACT *, struct pglist **, u_int32_t *, db_pgno_t *));
+ * PUBLIC: int __db_free_truncate __P((DB *, DB_THREAD_INFO *, DB_TXN *,
+ * PUBLIC:    u_int32_t, DB_COMPACT *, db_pglist_t **, u_int32_t *,
+ * PUBLIC:    db_pgno_t *));
  * PUBLIC: #endif
  */
 int
-__db_free_truncate(dbp, txn, flags, c_data, listp, nelemp, last_pgnop)
+__db_free_truncate(dbp, ip, txn, flags, c_data, listp, nelemp, last_pgnop)
 	DB *dbp;
+	DB_THREAD_INFO *ip;
 	DB_TXN *txn;
 	u_int32_t flags;
 	DB_COMPACT *c_data;
-	struct pglist **listp;
+	db_pglist_t **listp;
 	u_int32_t *nelemp;
 	db_pgno_t *last_pgnop;
 {
 	DBC *dbc;
-	DB_ENV *dbenv;
 	DBMETA *meta;
-	DBT ddbt;
 	DB_LOCK metalock;
-	DB_LSN null_lsn;
 	DB_MPOOLFILE *mpf;
+	ENV *env;
 	PAGE *h;
+	db_pglist_t *list, *lp;
 	db_pgno_t pgno;
 	u_int32_t nelems;
-	struct pglist *list, *lp;
 	int ret, t_ret;
 	size_t size;
 
 	COMPQUIET(flags, 0);
 	list = NULL;
 	meta = NULL;
-	dbenv = dbp->dbenv;
+	env = dbp->env;
 	mpf = dbp->mpf;
 	h = NULL;
 	nelems = 0;
 	if (listp != NULL) {
 		*listp = NULL;
-		DB_ASSERT(nelemp != NULL);
+		DB_ASSERT(env, nelemp != NULL);
 		*nelemp = 0;
 	}
 
-	if ((ret = __db_cursor(dbp, txn, &dbc, DB_WRITELOCK)) != 0)
+	if ((ret = __db_cursor(dbp, ip, txn, &dbc, DB_WRITELOCK)) != 0)
 		return (ret);
 
 	pgno = PGNO_BASE_MD;
 	if ((ret = __db_lget(dbc,
 	    LCK_ALWAYS, pgno, DB_LOCK_WRITE, 0, &metalock)) != 0)
 		goto err;
-	if ((ret = __memp_fget(mpf, &pgno, 0, &meta)) != 0)
+	if ((ret = __memp_fget(mpf, &pgno, dbc->thread_info, dbc->txn, 0,
+	    &meta)) != 0)
 		goto err;
 
 	if (last_pgnop != NULL)
@@ -685,44 +881,42 @@ __db_free_truncate(dbp, txn, flags, c_data, listp, nelemp, last_pgnop)
 		goto done;
 
 	size = 128;
-	if ((ret = __os_malloc(dbenv, size * sizeof(*list), &list)) != 0)
+	if ((ret = __os_malloc(env, size * sizeof(*list), &list)) != 0)
 		goto err;
 	lp = list;
 
 	do {
 		if (lp == &list[size]) {
 			size *= 2;
-			if ((ret = __os_realloc(dbenv,
+			if ((ret = __os_realloc(env,
 			    size * sizeof(*list), &list)) != 0)
 				goto err;
 			lp = &list[size / 2];
 		}
-		if ((ret = __memp_fget(mpf, &pgno, 0, &h)) != 0)
+		if ((ret = __memp_fget(mpf, &pgno,
+		     dbc->thread_info, dbc->txn, 0, &h)) != 0)
 			goto err;
 
 		lp->pgno = pgno;
+		lp->next_pgno = NEXT_PGNO(h);
 		lp->lsn = LSN(h);
 		pgno = NEXT_PGNO(h);
-		if ((ret = __memp_fput(mpf, h, 0)) != 0)
+		if ((ret = __memp_fput(mpf,
+		    dbc->thread_info, h, dbc->priority)) != 0)
 			goto err;
 		lp++;
 	} while (pgno != PGNO_INVALID);
 	nelems = (u_int32_t)(lp - list);
 
-	/* Log the current state of the free list */
-	if (DBC_LOGGING(dbc)) {
-		ddbt.data = list;
-		ddbt.size = nelems * sizeof(*lp);
-		ZERO_LSN(null_lsn);
-		if ((ret = __db_pg_sort_log(dbp,
-		     dbc->txn, &LSN(meta), DB_FLUSH, PGNO_BASE_MD, &LSN(meta),
-		     PGNO_INVALID, &null_lsn, meta->last_pgno, &ddbt)) != 0)
-			goto err;
-	} else
-		LSN_NOT_LOGGED(LSN(meta));
+	if ((ret = __memp_dirty(mpf,
+	    &meta, dbc->thread_info, dbc->txn, dbc->priority, 0)) != 0)
+		goto err;
 
-	if ((ret = __db_pg_truncate(mpf, list, c_data,
-	    &nelems, &meta->last_pgno, &LSN(meta), 0)) != 0)
+	/* Sort the list */
+	__db_freelist_sort(list, nelems);
+
+	if ((ret = __db_pg_truncate(dbc, txn, list, c_data,
+	    &nelems, meta->free, &meta->last_pgno, &LSN(meta), 0)) != 0)
 		goto err;
 
 	if (nelems == 0)
@@ -748,13 +942,13 @@ done:	if (last_pgnop != NULL)
 	}
 
 err:	if (list != NULL)
-		__os_free(dbenv, list);
-	if (meta != NULL && (t_ret =
-	     __memp_fput(mpf, (PAGE *)meta, DB_MPOOL_DIRTY)) != 0 && ret == 0)
+		__os_free(env, list);
+	if (meta != NULL && (t_ret = __memp_fput(mpf,
+	     dbc->thread_info, (PAGE *)meta, dbc->priority)) != 0 && ret == 0)
 		ret = t_ret;
 	if ((t_ret = __TLPUT(dbc, metalock)) != 0 && ret == 0)
 		ret = t_ret;
-	if ((t_ret = __db_c_close(dbc)) != 0 && ret == 0)
+	if ((t_ret = __dbc_close(dbc)) != 0 && ret == 0)
 		ret = t_ret;
 	return (ret);
 }
@@ -768,59 +962,98 @@ __db_truncate_freelist(dbc, meta, h, list, start, nelem)
 	u_int32_t start, nelem;
 {
 	DB *dbp;
+	DBT ddbt;
 	DB_LSN null_lsn;
 	DB_MPOOLFILE *mpf;
-	DBT ddbt;
 	PAGE *last_free, *pg;
-	db_pgno_t *lp;
-	struct pglist *plist, *pp;
-	int ret;
+	db_pgno_t *lp, free_pgno, lpgno;
+	db_pglist_t *plist, *pp, *spp;
+	u_int32_t elem, log_size;
+	int last, ret;
 
 	dbp = dbc->dbp;
 	mpf = dbp->mpf;
 	plist = NULL;
 	last_free = NULL;
+	pg = NULL;
 
 	if (start != 0 &&
-	    (ret = __memp_fget(mpf, &list[start - 1], 0, &last_free)) != 0)
+	    (ret = __memp_fget(mpf, &list[start - 1],
+	    dbc->thread_info, dbc->txn, DB_MPOOL_DIRTY, &last_free)) != 0)
 		goto err;
 
 	if (DBC_LOGGING(dbc)) {
-		if ((ret = __os_malloc(dbp->dbenv,
+		if ((ret = __os_malloc(dbp->env,
 		     (nelem - start) * sizeof(*pp), &plist)) != 0)
 			goto err;
 
 		pp = plist;
 		for (lp = &list[start]; lp < &list[nelem]; lp++) {
 			pp->pgno = *lp;
-			if ((ret = __memp_fget(mpf, lp, 0, &pg)) != 0)
+			if ((ret = __memp_fget(mpf, lp,
+			     dbc->thread_info, dbc->txn, 0, &pg)) != 0)
 				goto err;
 			pp->lsn = LSN(pg);
-			if ((ret = __memp_fput(mpf, pg, DB_MPOOL_DISCARD)) != 0)
+			pp->next_pgno = NEXT_PGNO(pg);
+			if ((ret = __memp_fput(mpf,
+			    dbc->thread_info, pg, DB_PRIORITY_VERY_LOW)) != 0)
 				goto err;
+			pg = NULL;
 			pp++;
 		}
-		ddbt.data = plist;
-		ddbt.size = (nelem - start) * sizeof(*pp);
 		ZERO_LSN(null_lsn);
-		if (last_free != NULL) {
-			if ((ret = __db_pg_sort_log(dbp, dbc->txn, &LSN(meta),
-			     DB_FLUSH, PGNO(meta), &LSN(meta), PGNO(last_free),
-			     &LSN(last_free), meta->last_pgno, &ddbt)) != 0)
-				goto err;
-		} else if ((ret = __db_pg_sort_log(dbp, dbc->txn,
-		     &LSN(meta), DB_FLUSH, PGNO(meta), &LSN(meta),
-		     PGNO_INVALID, &null_lsn, meta->last_pgno, &ddbt)) != 0)
+		pp = plist;
+		elem = nelem - start;
+		log_size = ((LOG *)dbc->env->
+		    lg_handle->reginfo.primary)->log_size;
+again:		ddbt.data = spp = pp;
+		free_pgno = pp->pgno;
+		lpgno = meta->last_pgno;
+		ddbt.size = elem * sizeof(*pp);
+		if (ddbt.size > log_size / 2) {
+			elem = (log_size / 2) / (u_int32_t)sizeof(*pp);
+			ddbt.size = elem * sizeof(*pp);
+			pp += elem;
+			elem = (nelem - start) - (u_int32_t)(pp - plist);
+			lpgno = pp[-1].pgno;
+			last = 0;
+		} else
+			last = 1;
+		/*
+		 * Get the page which will link to this section if we abort.
+		 * If this is the first segment then its last_free.
+		 */
+		if (spp == plist) 
+			pg = last_free;
+		else if ((ret = __memp_fget(mpf, &spp[-1].pgno,
+		     dbc->thread_info, dbc->txn, DB_MPOOL_DIRTY, &pg)) != 0)
 			goto err;
+			
+		if ((ret = __db_pg_trunc_log(dbp, dbc->txn,
+		     &LSN(meta), last == 1 ? DB_FLUSH : 0,
+		     PGNO(meta), &LSN(meta),
+		     pg != NULL ? PGNO(pg) : PGNO_INVALID,
+		     pg != NULL ? &LSN(pg) : &null_lsn,
+		     free_pgno, lpgno, &ddbt)) != 0)
+			goto err;
+		if (pg != NULL) {
+			LSN(pg) = LSN(meta);
+			if (pg != last_free && (ret = __memp_fput(mpf, 
+			    dbc->thread_info, pg, DB_PRIORITY_VERY_LOW)) != 0)
+				goto err;
+			pg = NULL;
+		}
+		if (last == 0)
+			goto again;
 	} else
 		LSN_NOT_LOGGED(LSN(meta));
-	if (last_free != NULL)
-		LSN(last_free) = LSN(meta);
 
-	if ((ret = __memp_fput(mpf, h, DB_MPOOL_DISCARD)) != 0)
+	if ((ret = __memp_fput(mpf,
+	    dbc->thread_info, h, DB_PRIORITY_VERY_LOW)) != 0)
 		goto err;
 	h = NULL;
-	if ((ret = __memp_ftruncate(mpf, list[start], 0)) != 0)
+	if ((ret = __memp_ftruncate(mpf, dbc->txn, dbc->thread_info,
+	    list[start], 0)) != 0)
 		goto err;
 	meta->last_pgno = list[start] - 1;
 
@@ -828,7 +1061,8 @@ __db_truncate_freelist(dbc, meta, h, list, start, nelem)
 		meta->free = PGNO_INVALID;
 	else {
 		NEXT_PGNO(last_free) = PGNO_INVALID;
-		if ((ret = __memp_fput(mpf, last_free, DB_MPOOL_DIRTY)) != 0)
+		if ((ret = __memp_fput(mpf,
+		    dbc->thread_info, last_free, dbc->priority)) != 0)
 			goto err;
 		last_free = NULL;
 	}
@@ -837,13 +1071,16 @@ __db_truncate_freelist(dbc, meta, h, list, start, nelem)
 	ret = __memp_extend_freelist(mpf, start, &list);
 
 err:	if (plist != NULL)
-		__os_free(dbp->dbenv, plist);
+		__os_free(dbp->env, plist);
 
 	/* We need to put the page on error. */
 	if (h != NULL)
-		(void)__memp_fput(mpf, h, 0);
+		(void)__memp_fput(mpf, dbc->thread_info, h, dbc->priority);
+	if (pg != NULL && pg != last_free)
+		(void)__memp_fput(mpf, dbc->thread_info, pg, dbc->priority);
 	if (last_free != NULL)
-		(void)__memp_fput(mpf, last_free, 0);
+		(void)__memp_fput(mpf,
+		    dbc->thread_info, last_free, dbc->priority);
 
 	return (ret);
 }
@@ -860,16 +1097,16 @@ int
 __db_lprint(dbc)
 	DBC *dbc;
 {
-	DB_ENV *dbenv;
 	DB *dbp;
 	DB_LOCKREQ req;
+	ENV *env;
 
 	dbp = dbc->dbp;
-	dbenv = dbp->dbenv;
+	env = dbp->env;
 
-	if (LOCKING_ON(dbenv)) {
+	if (LOCKING_ON(env)) {
 		req.op = DB_LOCK_DUMP;
-		(void)__lock_vec(dbenv, dbc->locker, 0, &req, 1, NULL);
+		(void)__lock_vec(env, dbc->locker, 0, &req, 1, NULL);
 	}
 	return (0);
 }
@@ -892,23 +1129,24 @@ __db_lget(dbc, action, pgno, mode, lkflags, lockp)
 	DB_LOCK *lockp;
 {
 	DB *dbp;
-	DB_ENV *dbenv;
 	DB_LOCKREQ couple[3], *reqp;
 	DB_TXN *txn;
+	ENV *env;
 	int has_timeout, i, ret;
 
 	dbp = dbc->dbp;
-	dbenv = dbp->dbenv;
+	env = dbp->env;
 	txn = dbc->txn;
 
 	/*
 	 * We do not always check if we're configured for locking before
 	 * calling __db_lget to acquire the lock.
 	 */
-	if (CDB_LOCKING(dbenv) ||
-	    !LOCKING_ON(dbenv) || F_ISSET(dbc, DBC_COMPENSATE) ||
-	    (F_ISSET(dbc, DBC_RECOVER) &&
-	    (action != LCK_ROLLBACK || IS_REP_CLIENT(dbenv))) ||
+	if (CDB_LOCKING(env) || !LOCKING_ON(env) ||
+	    (MULTIVERSION(dbp) && mode == DB_LOCK_READ &&
+	    dbc->txn != NULL && F_ISSET(dbc->txn, TXN_SNAPSHOT)) ||
+	    F_ISSET(dbc, DBC_DONTLOCK) || (F_ISSET(dbc, DBC_RECOVER) &&
+	    (action != LCK_ROLLBACK || IS_REP_CLIENT(env))) ||
 	    (action != LCK_ALWAYS && F_ISSET(dbc, DBC_OPD))) {
 		LOCK_INIT(*lockp);
 		return (0);
@@ -920,8 +1158,6 @@ __db_lget(dbc, action, pgno, mode, lkflags, lockp)
 	else
 		dbc->lock.type = DB_PAGE_LOCK;
 	lkflags &= ~DB_LOCK_RECORD;
-	if (action == LCK_ROLLBACK)
-		lkflags |= DB_LOCK_ABORT;
 
 	/*
 	 * If the transaction enclosing this cursor has DB_LOCK_NOWAIT set,
@@ -948,11 +1184,10 @@ __db_lget(dbc, action, pgno, mode, lkflags, lockp)
 		action = 0;
 	else if (dbc->txn == NULL || action == LCK_COUPLE_ALWAYS)
 		action = LCK_COUPLE;
-	else if (F_ISSET(dbc,
-	    DBC_READ_COMMITTED) && lockp->mode == DB_LOCK_READ)
+	else if (F_ISSET(dbc, DBC_READ_COMMITTED | DBC_WAS_READ_COMMITTED) &&
+	    lockp->mode == DB_LOCK_READ)
 		action = LCK_COUPLE;
-	else if (F_ISSET(dbc,
-	    DBC_READ_UNCOMMITTED) && lockp->mode == DB_LOCK_READ_UNCOMMITTED)
+	else if (lockp->mode == DB_LOCK_READ_UNCOMMITTED)
 		action = LCK_COUPLE;
 	else if (F_ISSET(dbc->dbp,
 	    DB_AM_READ_UNCOMMITTED) && lockp->mode == DB_LOCK_WRITE)
@@ -964,8 +1199,8 @@ __db_lget(dbc, action, pgno, mode, lkflags, lockp)
 	switch (action) {
 	default:
 		if (has_timeout)
-			goto couple;
-		ret = __lock_get(dbenv,
+			goto do_couple;
+		ret = __lock_get(env,
 		    dbc->locker, lkflags, &dbc->lock_dbt, mode, lockp);
 		break;
 
@@ -978,7 +1213,7 @@ __db_lget(dbc, action, pgno, mode, lkflags, lockp)
 		i++;
 		/* FALLTHROUGH */
 	case LCK_COUPLE:
-couple:		couple[i].op = has_timeout? DB_LOCK_GET_TIMEOUT : DB_LOCK_GET;
+do_couple:	couple[i].op = has_timeout? DB_LOCK_GET_TIMEOUT : DB_LOCK_GET;
 		couple[i].obj = &dbc->lock_dbt;
 		couple[i].mode = mode;
 		UMRW_SET(couple[i].timeout);
@@ -992,7 +1227,7 @@ couple:		couple[i].op = has_timeout? DB_LOCK_GET_TIMEOUT : DB_LOCK_GET;
 			i++;
 		}
 
-		ret = __lock_vec(dbenv,
+		ret = __lock_vec(env,
 		    dbc->locker, lkflags, couple, i, &reqp);
 		if (ret == 0 || reqp == &couple[i - 1])
 			*lockp = i == 1 ? couple[0].lock : couple[i - 2].lock;
@@ -1001,8 +1236,8 @@ couple:		couple[i].op = has_timeout? DB_LOCK_GET_TIMEOUT : DB_LOCK_GET;
 
 	if (txn != NULL && ret == DB_LOCK_DEADLOCK)
 		F_SET(txn, TXN_DEADLOCK);
-	return ((ret == DB_LOCK_NOTGRANTED &&
-	     !F_ISSET(dbenv, DB_ENV_TIME_NOTGRANTED)) ? DB_LOCK_DEADLOCK : ret);
+	return ((ret == DB_LOCK_NOTGRANTED && !F_ISSET(env->dbenv,
+		 DB_ENV_TIME_NOTGRANTED)) ? DB_LOCK_DEADLOCK : ret);
 }
 
 /*
@@ -1016,8 +1251,8 @@ __db_lput(dbc, lockp)
 	DBC *dbc;
 	DB_LOCK *lockp;
 {
-	DB_ENV *dbenv;
 	DB_LOCKREQ couple[2], *reqp;
+	ENV *env;
 	int action, ret;
 
 	/*
@@ -1030,19 +1265,18 @@ __db_lput(dbc, lockp)
 		action = LCK_DOWNGRADE;
 	else if (dbc->txn == NULL)
 		action = LCK_COUPLE;
-	else if (F_ISSET(dbc,
-	    DBC_READ_COMMITTED) && lockp->mode == DB_LOCK_READ)
+	else if (F_ISSET(dbc, DBC_READ_COMMITTED | DBC_WAS_READ_COMMITTED) &&
+	    lockp->mode == DB_LOCK_READ)
 		action = LCK_COUPLE;
-	else if (F_ISSET(dbc,
-	    DBC_READ_UNCOMMITTED) && lockp->mode == DB_LOCK_READ_UNCOMMITTED)
+	else if (lockp->mode == DB_LOCK_READ_UNCOMMITTED)
 		action = LCK_COUPLE;
 	else
 		action = 0;
 
-	dbenv = dbc->dbp->dbenv;
+	env = dbc->env;
 	switch (action) {
 	case LCK_COUPLE:
-		ret = __lock_put(dbenv, lockp);
+		ret = __lock_put(env, lockp);
 		break;
 	case LCK_DOWNGRADE:
 		couple[0].op = DB_LOCK_GET;
@@ -1052,7 +1286,7 @@ __db_lput(dbc, lockp)
 		UMRW_SET(couple[0].timeout);
 		couple[1].op = DB_LOCK_PUT;
 		couple[1].lock = *lockp;
-		ret = __lock_vec(dbenv, dbc->locker, 0, couple, 2, &reqp);
+		ret = __lock_vec(env, dbc->locker, 0, couple, 2, &reqp);
 		if (ret == 0 || reqp == &couple[1])
 			*lockp = couple[0].lock;
 		break;
