@@ -1,24 +1,92 @@
-/*-
+/*
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2005
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1996-2009 Oracle.  All rights reserved.
  *
- * $Id: mutex_int.h,v 12.17 2005/11/08 22:26:49 mjc Exp $
+ * $Id$
  */
 
 #ifndef _DB_MUTEX_INT_H_
 #define	_DB_MUTEX_INT_H_
 
+#include "dbinc/atomic.h"
+
+#if defined(__cplusplus)
+extern "C" {
+#endif
+
+/*
+ * Mutexes and Shared Latches
+ *
+ * Mutexes may be test-and-set (spinning & yielding when busy),
+ * native versions (pthreads, WaitForSingleObject)
+ * or a hybrid which has the lower no-contention overhead of test-and-set
+ * mutexes, using operating system calls only to block and wakeup.
+ *
+ * Hybrid exclusive-only mutexes include a 'tas' field.
+ * Hybrid DB_MUTEX_SHARED latches also include a 'shared' field.
+ */
+
 /*********************************************************************
  * POSIX.1 pthreads interface.
  *********************************************************************/
-#ifdef HAVE_MUTEX_PTHREADS
-#include <pthread.h>
-
+#if defined(HAVE_MUTEX_PTHREADS)
+/*
+ * Pthreads-based mutexes (exclusive-only) and latches (possibly shared)
+ * have the same MUTEX_FIELDS union. Different parts of the union are used
+ * depending on:
+ *    -	whether HAVE_SHARED_LATCHES is defined, and
+ *    - if HAVE_SHARED_LATCHES, whether this particular instance of a mutex
+ *	is a shared mutexDB_MUTEX_SHARED.
+ *
+ * The rwlock part of the union is used *only* for non-hybrid shared latches;
+ * in all other cases the mutex and cond fields are the only ones used.
+ *
+ *  configuration &	Who uses the field
+ *  mutex
+ *			mutex	cond	rwlock	tas
+ * Native mutex		y	y
+ * Hybrid mutexes	y	y		y
+ * Native sharedlatches			y
+ * Hybrid sharedlatches	y	y		y
+ *
+ * They all have a condition variable which is used only for
+ * DB_MUTEX_SELF_BLOCK waits.
+ *
+ * There can be no self-blocking shared latches: the pthread_cond_wait() would
+ * require getting a pthread_mutex_t, also it would not make sense.
+ */
 #define	MUTEX_FIELDS							\
-	pthread_mutex_t mutex;		/* Mutex. */			\
-	pthread_cond_t  cond;		/* Condition variable. */
+	union {								\
+		struct {						\
+		    pthread_mutex_t mutex;	/* Mutex */		\
+		    pthread_cond_t  cond;	/* Condition variable */ \
+		} m;							\
+		pthread_rwlock_t rwlock;	/* Read/write lock */	\
+	} u;
+
+#if defined(HAVE_SHARED_LATCHES) && !defined(HAVE_MUTEX_HYBRID)
+#define	RET_SET_PTHREAD_LOCK(mutexp, ret) do {				\
+	if (F_ISSET(mutexp, DB_MUTEX_SHARED))				\
+		RET_SET((pthread_rwlock_wrlock(&(mutexp)->u.rwlock)),	\
+		    ret);						\
+	else								\
+		RET_SET((pthread_mutex_lock(&(mutexp)->u.m.mutex)), ret); \
+} while (0)
+#define	RET_SET_PTHREAD_TRYLOCK(mutexp, ret) do {			\
+	if (F_ISSET(mutexp, DB_MUTEX_SHARED))				\
+		RET_SET((pthread_rwlock_trywrlock(&(mutexp)->u.rwlock)), \
+		    ret);						\
+	else								\
+		RET_SET((pthread_mutex_trylock(&(mutexp)->u.m.mutex)),	\
+		    ret);						\
+} while (0)
+#else
+#define	RET_SET_PTHREAD_LOCK(mutexp, ret)				\
+		RET_SET(pthread_mutex_lock(&(mutexp)->u.m.mutex), ret);
+#define	RET_SET_PTHREAD_TRYLOCK(mutexp, ret)				\
+		RET_SET(pthread_mutex_trylock(&(mutexp)->u.m.mutex), ret);
+#endif
 #endif
 
 #ifdef HAVE_MUTEX_UI_THREADS
@@ -192,9 +260,20 @@ typedef abilock_t tsl_t;
  * correctly as far as we know.
  *********************************************************************/
 #ifdef HAVE_MUTEX_SOLARIS_LOCK_TRY
+#include <sys/atomic.h>
+#define	MUTEX_MEMBAR(x)	membar_enter()
+#define	MEMBAR_ENTER()	membar_enter()
+#define	MEMBAR_EXIT()	membar_exit()
 #include <sys/machlock.h>
 typedef lock_t tsl_t;
 
+/*
+ * The functions are declared in <sys/machlock.h>, but under #ifdef KERNEL.
+ * Re-declare them here to avoid warnings.
+ */
+extern  int _lock_try(lock_t *);
+extern void _lock_clear(lock_t *);
+
 #ifdef LOAD_ACTUAL_MUTEX_CODE
 #define	MUTEX_INIT(x)	0
 #define	MUTEX_SET(x)	_lock_try(x)
@@ -206,7 +285,7 @@ typedef lock_t tsl_t;
  * VMS.
  *********************************************************************/
 #ifdef HAVE_MUTEX_VMS
-#include <sys/mman.h>;
+#include <sys/mman.h>
 #include <builtins.h>
 typedef volatile unsigned char tsl_t;
 
@@ -217,7 +296,7 @@ typedef volatile unsigned char tsl_t;
 #define	MUTEX_SET(tsl)		(!(int)_BBSSI(0, tsl))
 #endif
 #define	MUTEX_UNSET(tsl)	(*(tsl) = 0)
-#define	MUTEX_INIT(tsl)		MUTEX_UNSET(tsl)
+#define	MUTEX_INIT(tsl)         (MUTEX_UNSET(tsl), 0)
 #endif
 #endif
 
@@ -234,7 +313,13 @@ typedef volatile unsigned char tsl_t;
 typedef SEM_ID tsl_t;
 
 #ifdef LOAD_ACTUAL_MUTEX_CODE
-#define	MUTEX_SET(tsl)		(semTake((*tsl), WAIT_FOREVER) == OK)
+/*
+ * Uses of this MUTEX_SET() need to have a local 'nowait' variable,
+ * which determines whether to return right away when the semaphore
+ * is busy or to wait until it is available.
+ */
+#define	MUTEX_SET(tsl)							\
+	(semTake((*(tsl)), nowait ? NO_WAIT : WAIT_FOREVER) == OK)
 #define	MUTEX_UNSET(tsl)	(semGive((*tsl)))
 #define	MUTEX_INIT(tsl)							\
 	((*(tsl) = semBCreate(SEM_Q_FIFO, SEM_FULL)) == NULL)
@@ -262,11 +347,11 @@ typedef unsigned int tsl_t;
 #endif
 
 /*********************************************************************
- * Win32
+ * Win32 - always a hybrid mutex
  *********************************************************************/
 #if defined(HAVE_MUTEX_WIN32) || defined(HAVE_MUTEX_WIN32_GCC)
+typedef LONG volatile tsl_t;
 #define	MUTEX_FIELDS							\
-	LONG volatile tas;						\
 	LONG nwaiters;							\
 	u_int32_t id;	/* ID used for creating events */		\
 
@@ -285,12 +370,12 @@ typedef unsigned int tsl_t;
  * platforms, and it improves performance on Pentium 4 processor platforms."
  */
 #ifdef HAVE_MUTEX_WIN32
-#ifndef _WIN64
+#if !defined(_WIN64) && !defined(DB_WINCE)
 #define	MUTEX_PAUSE		{__asm{_emit 0xf3}; __asm{_emit 0x90}}
 #endif
 #endif
 #ifdef HAVE_MUTEX_WIN32_GCC
-#define	MUTEX_PAUSE		asm volatile ("rep; nop" : : );
+#define	MUTEX_PAUSE		__asm__ volatile ("rep; nop" : : );
 #endif
 #endif
 #endif
@@ -306,7 +391,7 @@ typedef unsigned char tsl_t;
 #define	MUTEX_SET(tsl) ({						\
 	register tsl_t *__l = (tsl);					\
 	int __r;							\
-	    asm volatile("tas  %1; \n					\
+	    __asm__ volatile("tas  %1; \n				\
 			  seq  %0"					\
 		: "=dm" (__r), "=m" (*__l)				\
 		: "1" (*__l)						\
@@ -315,7 +400,7 @@ typedef unsigned char tsl_t;
 })
 
 #define	MUTEX_UNSET(tsl)	(*(tsl) = 0)
-#define	MUTEX_INIT(tsl)		MUTEX_UNSET(tsl)
+#define	MUTEX_INIT(tsl)         (MUTEX_UNSET(tsl), 0)
 #endif
 #endif
 
@@ -336,7 +421,7 @@ static inline int
 MUTEX_SET(tsl_t *tsl) {
 	register tsl_t *__l = tsl;
 	register tsl_t __r;
-	asm volatile(
+	__asm__ volatile(
 		"1:	ldl_l	%0,%2\n"
 		"	blbs	%0,2f\n"
 		"	or	$31,1,%0\n"
@@ -356,7 +441,7 @@ MUTEX_SET(tsl_t *tsl) {
  */
 static inline int
 MUTEX_UNSET(tsl_t *tsl) {
-	asm volatile("	mb\n");
+	__asm__ volatile("	mb\n");
 	return *tsl = 0;
 }
 
@@ -391,7 +476,7 @@ typedef unsigned char tsl_t;
 /* gcc/arm: 0 is clear, 1 is set. */
 #define	MUTEX_SET(tsl) ({						\
 	int __r;							\
-	asm volatile(							\
+	__asm__ volatile(						\
 		"swpb	%0, %1, [%2]\n\t"				\
 		"eor	%0, %0, #1\n\t"					\
 	    : "=&r" (__r)						\
@@ -401,7 +486,7 @@ typedef unsigned char tsl_t;
 })
 
 #define	MUTEX_UNSET(tsl)	(*(volatile tsl_t *)(tsl) = 0)
-#define	MUTEX_INIT(tsl)		MUTEX_UNSET(tsl)
+#define	MUTEX_INIT(tsl)         (MUTEX_UNSET(tsl), 0)
 #endif
 #endif
 
@@ -422,7 +507,7 @@ typedef u_int32_t tsl_t;
 #define	MUTEX_SET(tsl) ({						\
 	register tsl_t *__l = (tsl);					\
 	int __r;							\
-	asm volatile("ldcws 0(%1),%0" : "=r" (__r) : "r" (__l));	\
+	__asm__ volatile("ldcws 0(%1),%0" : "=r" (__r) : "r" (__l));	\
 	__r & 1;							\
 })
 
@@ -442,7 +527,7 @@ typedef volatile unsigned char tsl_t;
 #define	MUTEX_SET(tsl) ({						\
 	register tsl_t *__l = (tsl);					\
 	long __r;							\
-	asm volatile("xchg1 %0=%1,%2" :					\
+	__asm__ volatile("xchg1 %0=%1,%2" :				\
 		     "=r"(__r), "+m"(*__l) : "r"(1));			\
 	__r ^ 1;							\
 })
@@ -452,7 +537,7 @@ typedef volatile unsigned char tsl_t;
  * semantics.
  */
 #define	MUTEX_UNSET(tsl)	(*(tsl_t *)(tsl) = 0)
-#define	MUTEX_INIT(tsl)		MUTEX_UNSET(tsl)
+#define	MUTEX_INIT(tsl)         (MUTEX_UNSET(tsl), 0)
 #endif
 #endif
 
@@ -495,8 +580,8 @@ typedef u_int32_t tsl_t;
  */
 static inline int
 MUTEX_SET(int *tsl)  {
-        int __r;
-        asm volatile (
+	int __r;
+	__asm__ volatile (
 "0:                             \n\t"
 "       lwarx   %0,0,%1         \n\t"
 "       cmpwi   %0,0            \n\t"
@@ -508,23 +593,23 @@ MUTEX_SET(int *tsl)  {
 "1:                             \n\t"
 "       li      %1,0            \n\t"
 "2:                             \n\t"
-         : "=&r" (__r), "+r" (tsl)
-         :
-         : "cr0", "memory");
-         return (int)tsl;
+	 : "=&r" (__r), "+r" (tsl)
+	 :
+	 : "cr0", "memory");
+	 return (int)tsl;
 }
 
 static inline int
 MUTEX_UNSET(tsl_t *tsl) {
-         asm volatile("sync" : : : "memory");
-         return *tsl = 0;
+	 __asm__ volatile("sync" : : : "memory");
+	 return *tsl = 0;
 }
 #define	MUTEX_INIT(tsl)		MUTEX_UNSET(tsl)
 #endif
 #endif
 
 /*********************************************************************
- * OS/390 C
+ * OS/390 C.
  *********************************************************************/
 #ifdef HAVE_MUTEX_S390_CC_ASSEMBLY
 typedef int tsl_t;
@@ -536,7 +621,7 @@ typedef int tsl_t;
  */
 #define	MUTEX_SET(tsl)		(!cs(&zero, (tsl), 1))
 #define	MUTEX_UNSET(tsl)	(*(tsl) = 0)
-#define	MUTEX_INIT(tsl)		MUTEX_UNSET(tsl)
+#define	MUTEX_INIT(tsl)         (MUTEX_UNSET(tsl), 0)
 #endif
 #endif
 
@@ -552,7 +637,7 @@ static inline int
 MUTEX_SET(tsl_t *tsl) {							\
 	register tsl_t *__l = (tsl);					\
 	int __r;							\
-  asm volatile(								\
+  __asm__ volatile(							\
        "    la    1,%1\n"						\
        "    lhi   0,1\n"						\
        "    l     %0,%1\n"						\
@@ -564,7 +649,7 @@ MUTEX_SET(tsl_t *tsl) {							\
 }
 
 #define	MUTEX_UNSET(tsl)	(*(tsl) = 0)
-#define	MUTEX_INIT(tsl)		MUTEX_UNSET(tsl)
+#define	MUTEX_INIT(tsl)         (MUTEX_UNSET(tsl), 0)
 #endif
 #endif
 
@@ -595,7 +680,7 @@ _tsl_set(void *tsl)
 
 #define	MUTEX_SET(tsl)		_tsl_set(tsl)
 #define	MUTEX_UNSET(tsl)	(*(tsl) = 0)
-#define	MUTEX_INIT(tsl)		MUTEX_UNSET(tsl)
+#define	MUTEX_INIT(tsl)         (MUTEX_UNSET(tsl), 0)
 #endif
 #endif
 
@@ -605,19 +690,18 @@ _tsl_set(void *tsl)
 #ifdef HAVE_MUTEX_SPARC_GCC_ASSEMBLY
 typedef unsigned char tsl_t;
 
+#define	MUTEX_ALIGN	8
+
 #ifdef LOAD_ACTUAL_MUTEX_CODE
 /*
- *
  * The ldstub instruction takes the location specified by its first argument
  * (a register containing a memory address) and loads its contents into its
  * second argument (a register) and atomically sets the contents the location
  * specified by its first argument to a byte of 1s.  (The value in the second
  * argument is never read, but only overwritten.)
  *
- * The stbar is needed for v8, and is implemented as membar #sync on v9,
- * so is functional there as well.  For v7, stbar may generate an illegal
- * instruction and we have no way to tell what we're running on.  Some
- * operating systems notice and skip this instruction in the fault handler.
+ * Hybrid mutexes require membar #StoreLoad and #LoadStore ordering on multi-
+ * processor v9 systems.
  *
  * gcc/sparc: 0 is clear, 1 is set.
  */
@@ -630,8 +714,14 @@ typedef unsigned char tsl_t;
 	!__r;								\
 })
 
-#define	MUTEX_UNSET(tsl)	(*(tsl) = 0)
-#define	MUTEX_INIT(tsl)		MUTEX_UNSET(tsl)
+#define	MUTEX_UNSET(tsl)	(*(tsl) = 0, MUTEX_MEMBAR(tsl))
+#define	MUTEX_INIT(tsl)         (MUTEX_UNSET(tsl), 0)
+#define	MUTEX_MEMBAR(x)	\
+	({ __asm__ volatile ("membar #StoreStore|#StoreLoad|#LoadStore"); })
+#define	MEMBAR_ENTER() \
+	({ __asm__ volatile ("membar #StoreStore|#StoreLoad"); })
+#define	MEMBAR_EXIT() \
+	({ __asm__ volatile ("membar #StoreStore|#LoadStore"); })
 #endif
 #endif
 
@@ -654,7 +744,7 @@ typedef int tsl_t;
 #ifdef HAVE_MUTEX_MIPS_GCC_ASSEMBLY
 typedef u_int32_t tsl_t;
 
-#define MUTEX_ALIGN	4
+#define	MUTEX_ALIGN	4
 
 #ifdef LOAD_ACTUAL_MUTEX_CODE
 /*
@@ -664,51 +754,76 @@ typedef u_int32_t tsl_t;
 static inline int
 MUTEX_SET(tsl_t *tsl) {
        register tsl_t *__l = tsl;
-       register tsl_t __r;
-       __asm__ __volatile__(
-               "       .set push           \n"
-               "       .set mips2          \n"
-               "       .set noreorder      \n"
-               "       .set nomacro        \n"
-               "1:     ll      %0,%1       \n"
-               "       bne     %0,$0,1f    \n"
-               "       xori    %0,%0,1     \n"
-               "       sc      %0,%1       \n"
-               "       beql    %0,$0,1b    \n"
-               "       xori    %0,1        \n"
-               "1:     .set pop              "
-               : "=&r" (__r), "+R" (*__l));
-       return __r;
+       register tsl_t __r, __t;
+       __asm__ volatile(
+	       "       .set push           \n"
+	       "       .set mips2          \n"
+	       "       .set noreorder      \n"
+	       "       .set nomacro        \n"
+	       "1:     ll      %0, %3      \n"
+	       "       ori     %2, %0, 1   \n"
+	       "       sc      %2, %1      \n"
+	       "       beqzl   %2, 1b      \n"
+	       "       nop                 \n"
+	       "       andi    %2, %0, 1   \n"
+	       "       sync                \n"
+	       "       .set reorder        \n"
+	       "       .set pop            \n"
+	       : "=&r" (__t), "=m" (*tsl), "=&r" (__r)
+	       : "m" (*tsl)
+	       : "memory");
+       return (!__r);
 }
 
-#define	MUTEX_UNSET(tsl)        (*(volatile tsl_t *)(tsl) = 0)
-#define	MUTEX_INIT(tsl)         MUTEX_UNSET(tsl)
+static inline void
+MUTEX_UNSET(tsl_t *tsl) {
+	__asm__ volatile(
+	       "       .set noreorder      \n"
+	       "       sync                \n"
+	       "       sw      $0, %0      \n"
+	       "       .set reorder        \n"
+	       : "=m" (*tsl)
+	       : "m" (*tsl)
+	       : "memory");
+}
+
+#define	       MUTEX_INIT(tsl)         (*(tsl) = 0)
 #endif
 #endif
 
 /*********************************************************************
- * x86/gcc assembly.
+ * x86/gcc (32- and 64-bit) assembly.
  *********************************************************************/
-#ifdef HAVE_MUTEX_X86_GCC_ASSEMBLY
-typedef unsigned char tsl_t;
+#if defined(HAVE_MUTEX_X86_GCC_ASSEMBLY) || \
+    defined(HAVE_MUTEX_X86_64_GCC_ASSEMBLY)
+typedef volatile unsigned char tsl_t;
 
 #ifdef LOAD_ACTUAL_MUTEX_CODE
 /* gcc/x86: 0 is clear, 1 is set. */
 #define	MUTEX_SET(tsl) ({						\
-	register tsl_t *__l = (tsl);					\
-	int __r;							\
-	asm volatile("movl $1,%%eax\n"					\
-		     "lock\n"						\
-		     "xchgb %1,%%al\n"					\
-		     "xorl $1,%%eax"					\
-	    : "=&a" (__r), "=m" (*__l)					\
-	    : "m1" (*__l)						\
-	    );								\
-	__r & 1;							\
+	tsl_t __r;							\
+	__asm__ volatile("movb $1, %b0\n\t"				\
+		"xchgb %b0,%1"						\
+	    : "=&q" (__r)						\
+	    : "m" (*(tsl_t *)(tsl))					\
+	    : "memory", "cc");						\
+	!__r;	/* return 1 on success, 0 on failure */			\
 })
 
-#define	MUTEX_UNSET(tsl)        (*(volatile tsl_t *)(tsl) = 0)
-#define	MUTEX_INIT(tsl)		MUTEX_UNSET(tsl)
+#define	MUTEX_UNSET(tsl)        (*(tsl_t *)(tsl) = 0)
+#define	MUTEX_INIT(tsl)		(MUTEX_UNSET(tsl), 0)
+/*
+ * We need to pass a valid address to generate the memory barrier
+ * otherwise PURIFY will complain.  Use something referenced recently
+ * and initialized.
+ */
+#if defined(HAVE_MUTEX_X86_GCC_ASSEMBLY)
+#define	MUTEX_MEMBAR(addr)						\
+    ({ __asm__ volatile ("lock; addl $0, %0" ::"m" (addr): "memory"); 1; })
+#else
+#define	MUTEX_MEMBAR(addr)						\
+    ({ __asm__ volatile ("mfence" ::: "memory"); 1; })
+#endif
 
 /*
  * From Intel's performance tuning documentation (and see SR #6975):
@@ -719,35 +834,11 @@ typedef unsigned char tsl_t;
  * instruction does not affect the correctness of programs on existing
  * platforms, and it improves performance on Pentium 4 processor platforms."
  */
-#define	MUTEX_PAUSE		asm volatile ("rep; nop" : : );
+#define	MUTEX_PAUSE		__asm__ volatile ("rep; nop" : : );
 #endif
 #endif
 
-/*********************************************************************
- * x86_64/gcc assembly.
- *********************************************************************/
-#ifdef HAVE_MUTEX_X86_64_GCC_ASSEMBLY
-typedef unsigned char tsl_t;
-
-#ifdef LOAD_ACTUAL_MUTEX_CODE
-/* gcc/x86_64: 0 is clear, 1 is set. */
-#define  MUTEX_SET(tsl) ({           					\
-	register tsl_t *__l = (tsl);          				\
-	int __r;              						\
-	asm volatile("mov $1,%%rax\n"					\
-		     "lock\n"						\
-		     "xchgb %1,%%al\n"					\
-		     "xor $1,%%rax"					\
-		: "=&a" (__r), "=m" (*__l)				\
-		: "1m" (*__l)						\
-		);							\
-	__r & 1;							\
-})
-
-#define	MUTEX_UNSET(tsl)	(*(tsl) = 0)
-#define	MUTEX_INIT(tsl)		MUTEX_UNSET(tsl)
-#endif
-#endif
+/* End of operating system & hardware architecture-specific definitions */
 
 /*
  * Mutex alignment defaults to sizeof(unsigned int).
@@ -769,25 +860,47 @@ typedef unsigned char tsl_t;
 #define	MUTEX_DESTROY(x)
 #endif
 
+/*
+ * Mutex pause defaults to a no-op.
+ */
+#ifndef	MUTEX_PAUSE
+#define	MUTEX_PAUSE
+#endif
+
+/*
+ * If no native atomic support is available then use mutexes to
+ * emulate atomic increment, decrement, and compare-and-exchange.
+ * The address of the atomic value selects which of a small number
+ * of mutexes to use to protect the updates.
+ * The number of mutexes should be somewhat larger than the number of
+ * processors in the system in order to minimize unnecessary contention.
+ * It defaults to 8 to handle most small (1-4) cpu systems, if it hasn't
+ * already been defined (e.g. in db_config.h)
+ */
+#if !defined(HAVE_ATOMIC_SUPPORT) && defined(HAVE_MUTEX_SUPPORT) && \
+    !defined(MAX_ATOMIC_MUTEXES)
+#define	MAX_ATOMIC_MUTEXES	1
+#endif
+
 /*
  * DB_MUTEXMGR --
  *	The mutex manager encapsulates the mutex system.
  */
-typedef struct __db_mutexmgr {
+struct __db_mutexmgr {
 	/* These fields are never updated after creation, so not protected. */
 	DB_ENV	*dbenv;			/* Environment */
 	REGINFO	 reginfo;		/* Region information */
 
 	void	*mutex_array;		/* Base of the mutex array */
-} DB_MUTEXMGR;
+};
 
 /* Macros to lock/unlock the mutex region as a whole. */
 #define	MUTEX_SYSTEM_LOCK(dbenv)					\
-	MUTEX_LOCK(dbenv, ((DB_MUTEXREGION *)((DB_MUTEXMGR *)		\
-	    (dbenv)->mutex_handle)->reginfo.primary)->mtx_region)
+	MUTEX_LOCK(dbenv, ((DB_MUTEXREGION *)				\
+	    (dbenv)->mutex_handle->reginfo.primary)->mtx_region)
 #define	MUTEX_SYSTEM_UNLOCK(dbenv)					\
-	MUTEX_UNLOCK(dbenv, ((DB_MUTEXREGION *)((DB_MUTEXMGR *)		\
-	    (dbenv)->mutex_handle)->reginfo.primary)->mtx_region)
+	MUTEX_UNLOCK(dbenv, ((DB_MUTEXREGION *)				\
+	    (dbenv)->mutex_handle->reginfo.primary)->mtx_region)
 
 /*
  * DB_MUTEXREGION --
@@ -795,7 +908,8 @@ typedef struct __db_mutexmgr {
  */
 typedef struct __db_mutexregion {
 	/* These fields are initialized at create time and never modified. */
-	roff_t		mutex_offset;	/* Offset of mutex array */
+	roff_t		mutex_off_alloc;/* Offset of mutex array */
+	roff_t		mutex_off;	/* Adjusted offset of mutex array */
 	size_t		mutex_size;	/* Size of the aligned mutex */
 	roff_t		thread_off;	/* Offset of the thread area. */
 
@@ -804,26 +918,60 @@ typedef struct __db_mutexregion {
 	/* Protected using the region mutex. */
 	u_int32_t	mutex_next;	/* Next free mutex */
 
+#if !defined(HAVE_ATOMIC_SUPPORT) && defined(HAVE_MUTEX_SUPPORT)
+	/* Mutexes for emulating atomic operations. */
+	db_mutex_t	mtx_atomic[MAX_ATOMIC_MUTEXES];
+#endif
+
 	DB_MUTEX_STAT	stat;		/* Mutex statistics */
 } DB_MUTEXREGION;
 
-typedef struct __mutex_t {		/* Mutex. */
+#ifdef HAVE_MUTEX_SUPPORT
+struct __db_mutex_t {			/* Mutex. */
 #ifdef MUTEX_FIELDS
-	MUTEX_FIELDS
+	MUTEX_FIELDS			/* Opaque thread mutex structures. */
 #endif
-#if !defined(MUTEX_FIELDS) && !defined(HAVE_MUTEX_FCNTL)
-	tsl_t		tas;		/* Test and set. */
+#ifndef HAVE_MUTEX_FCNTL
+#if defined(HAVE_MUTEX_HYBRID) || \
+    (defined(HAVE_SHARED_LATCHES) && !defined(HAVE_MUTEX_PTHREADS))
+	/*
+	 * For hybrid and test-and-set shared latches it is a counter:
+	 * 0 means it is free,
+	 * -1 is exclusively locked,
+	 * > 0 is the number of shared readers.
+	 * Pthreads shared latches use pthread_rwlock instead.
+	 */
+	db_atomic_t	sharecount;
+	tsl_t		tas;
+#elif !defined(MUTEX_FIELDS)
+	/*
+	 * This is the Test and Set flag for exclusive latches (mutexes):
+	 * there is a free value (often 0, 1, or -1) and a set value.
+	 */
+	tsl_t		tas;
 #endif
-	pid_t 		pid;		/* Process owning mutex */
-	db_threadid_t 	tid;		/* Thread owning mutex */
+#endif
+#ifdef HAVE_MUTEX_HYBRID
+	volatile u_int32_t wait;	/* Count of waiters. */
+#endif
+	pid_t		pid;		/* Process owning mutex */
+	db_threadid_t	tid;		/* Thread owning mutex */
 
-	u_int32_t mutex_next_link;	/* Linked list of free mutexes. */
+	db_mutex_t mutex_next_link;	/* Linked list of free mutexes. */
 
 #ifdef HAVE_STATISTICS
 	int	  alloc_id;		/* Allocation ID. */
 
 	u_int32_t mutex_set_wait;	/* Granted after wait. */
 	u_int32_t mutex_set_nowait;	/* Granted without waiting. */
+#ifdef HAVE_SHARED_LATCHES
+	u_int32_t mutex_set_rd_wait;	/* Granted shared lock after wait. */
+	u_int32_t mutex_set_rd_nowait;	/* Granted shared lock w/out waiting. */
+#endif
+#ifdef HAVE_MUTEX_HYBRID
+	u_int32_t hybrid_wait;
+	u_int32_t hybrid_wakeup;	/* for counting spurious wakeups */
+#endif
 #endif
 
 	/*
@@ -833,12 +981,93 @@ typedef struct __mutex_t {		/* Mutex. */
 	 * the possible flags values, getting a single byte on some machines
 	 * is expensive, and the mutex structure is a MP hot spot.
 	 */
-	u_int32_t flags;		/* MUTEX_XXX */
-} DB_MUTEX;
+	volatile u_int32_t flags;		/* MUTEX_XXX */
+};
+#endif
 
 /* Macro to get a reference to a specific mutex. */
-#define	MUTEXP_SET(indx)						\
-	(DB_MUTEX *)							\
-	    ((u_int8_t *)mtxmgr->mutex_array + (indx) * mtxregion->mutex_size);
+#define	MUTEXP_SET(mtxmgr, indx)					\
+	((DB_MUTEX *)((u_int8_t *)mtxmgr->mutex_array +			\
+	    (indx) * ((DB_MUTEXREGION *)mtxmgr->reginfo.primary)->mutex_size))
 
+/* Inverse of the above: get the mutex index from a mutex pointer */
+#define	MUTEXP_GET(mtxmgr, mutexp)					\
+	(((u_int8_t *) (mutexp) - (u_int8_t *)mtxmgr->mutex_array) /	\
+	 ((DB_MUTEXREGION *)mtxmgr->reginfo.primary)->mutex_size)
+
+/*
+ * Check that a particular mutex is exclusively held at least by someone, not
+ * necessarily the current thread.
+ */
+#ifdef HAVE_MUTEX_SUPPORT
+#define	MUTEX_IS_OWNED(env, mutex)					\
+	(mutex == MUTEX_INVALID || !MUTEX_ON(env) ||			\
+	F_ISSET(env->dbenv, DB_ENV_NOLOCKING) ||			\
+	F_ISSET(MUTEXP_SET(env->mutex_handle, mutex), DB_MUTEX_LOCKED))
+#else
+#define	MUTEX_IS_OWNED(env, mutex)	0
+#endif
+
+#if defined(HAVE_MUTEX_HYBRID) ||  defined(DB_WIN32) ||		\
+	(defined(HAVE_SHARED_LATCHES) && !defined(HAVE_MUTEX_PTHREADS))
+#define	MUTEXP_IS_BUSY(mutexp)					\
+	(F_ISSET(mutexp, DB_MUTEX_SHARED) ?			\
+	(atomic_read(&(mutexp)->sharecount) != 0) :		\
+	F_ISSET(mutexp, DB_MUTEX_LOCKED))
+#define	MUTEXP_BUSY_FIELD(mutexp)		\
+	(F_ISSET(mutexp, DB_MUTEX_SHARED) ?	\
+	(atomic_read(&(mutexp)->sharecount)) : (mutexp)->flags)
+#else
+/* Pthread_rwlocks don't have an low-cost 'is it being shared?' predicate. */
+#define	MUTEXP_IS_BUSY(mutexp)	(F_ISSET((mutexp), DB_MUTEX_LOCKED))
+#define	MUTEXP_BUSY_FIELD(mutexp)	((mutexp)->flags)
+#endif
+
+#define	MUTEX_IS_BUSY(env, mutex)					\
+	(mutex == MUTEX_INVALID || !MUTEX_ON(env) ||			\
+	F_ISSET(env->dbenv, DB_ENV_NOLOCKING) ||			\
+	MUTEXP_IS_BUSY(MUTEXP_SET(env->mutex_handle, mutex)))
+
+#define	MUTEX_REQUIRED(env, mutex)					\
+	DB_ASSERT(env, MUTEX_IS_OWNED(env, mutex))
+
+#define	MUTEX_REQUIRED_READ(env, mutex)					\
+	DB_ASSERT(env, MUTEX_IS_OWNED(env, mutex) || MUTEX_IS_BUSY(env, mutex))
+
+/*
+ * Test and set (and thus hybrid) shared latches use compare & exchange
+ * to acquire; the others the mutex-setting primitive defined above.
+ */
+#ifdef LOAD_ACTUAL_MUTEX_CODE
+
+#if defined(HAVE_SHARED_LATCHES)
+/* This is the value of the 'sharecount' of an exclusively held tas latch.
+ * The particular value is not special; it is just unlikely to be caused
+ * by releasing or acquiring a shared latch too many times.
+ */
+#define	MUTEX_SHARE_ISEXCLUSIVE	(-1024)
+
+/*
+ * Get an exclusive lock on a possibly sharable latch. We use the native
+ * MUTEX_SET() operation for non-sharable latches; it usually is faster.
+ */
+#define	MUTEXP_ACQUIRE(mutexp)	\
+	(F_ISSET(mutexp, DB_MUTEX_SHARED) ?			\
+	atomic_compare_exchange(env,				\
+	    &(mutexp)->sharecount, 0, MUTEX_SHARE_ISEXCLUSIVE) :	\
+	MUTEX_SET(&(mutexp)->tas))
+#else
+#define	MUTEXP_ACQUIRE(mutexp)		MUTEX_SET(&(mutexp)->tas)
+#endif
+
+#ifndef MEMBAR_ENTER
+#define	MEMBAR_ENTER()
+#define	MEMBAR_EXIT()
+#endif
+
+#endif
+
+#if defined(__cplusplus)
+}
+#endif
 #endif /* !_DB_MUTEX_INT_H_ */
