@@ -1,16 +1,17 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2005
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1996-2009 Oracle.  All rights reserved.
  *
- * $Id: txn.h,v 12.7 2005/10/13 00:53:00 bostic Exp $
+ * $Id$
  */
 
-#ifndef	_TXN_H_
-#define	_TXN_H_
+#ifndef	_DB_TXN_H_
+#define	_DB_TXN_H_
 
-#include "dbinc/xa.h"
+#if defined(__cplusplus)
+extern "C" {
+#endif
 
 /* Operation parameters to the delayed commit processing code. */
 typedef enum {
@@ -31,7 +32,8 @@ struct __txn_logrec;	typedef struct __txn_logrec DB_TXNLOGREC;
 #define	TXN_MAXIMUM	0xffffffff	/* Maximum number of txn ids. */
 #define	TXN_INVALID	0		/* Invalid transaction ID. */
 
-#define	DEF_MAX_TXNS	20		/* Default max transactions. */
+#define	DEF_MAX_TXNS	100		/* Default max transactions. */
+#define	TXN_NSLOTS	4		/* Initial slots to hold DB refs */
 
 /*
  * Internal data maintained in shared memory for each transaction.
@@ -40,44 +42,40 @@ typedef struct __txn_detail {
 	u_int32_t txnid;		/* current transaction id
 					   used to link free list also */
 	pid_t pid;			/* Process owning txn */
-	db_threadid_t tid;	/* Thread owning txn */
+	db_threadid_t tid;		/* Thread owning txn */
 
-	DB_LSN	last_lsn;		/* last lsn written for this txn */
-	DB_LSN	begin_lsn;		/* lsn of begin record */
+	DB_LSN	last_lsn;		/* Last LSN written for this txn. */
+	DB_LSN	begin_lsn;		/* LSN of begin record. */
 	roff_t	parent;			/* Offset of transaction's parent. */
 	roff_t	name;			/* Offset of txn name. */
 
+	u_int32_t	nlog_dbs;	/* Number of databases used. */
+	u_int32_t	nlog_slots;	/* Number of allocated slots. */
+	roff_t		log_dbs;	/* Databases used. */
+
+	DB_LSN	read_lsn;		/* Read LSN for MVCC. */
+	DB_LSN	visible_lsn;		/* LSN at which this transaction's
+					   changes are visible. */
+	db_mutex_t	mvcc_mtx;	/* Version mutex. */
+	u_int32_t	mvcc_ref;	/* Number of buffers created by this
+					   transaction still in cache.  */
+
 	SH_TAILQ_HEAD(__tdkids)	kids;	/* Linked list of child txn detail. */
 	SH_TAILQ_ENTRY		klinks;
 
-#define	TXN_RUNNING		1
-#define	TXN_ABORTED		2
-#define	TXN_PREPARED		3
-#define	TXN_COMMITTED		4
+	/* TXN_{ABORTED, COMMITTED PREPARED, RUNNING} */
 	u_int32_t status;		/* status of the transaction */
+
 #define	TXN_DTL_COLLECTED	0x1	/* collected during txn_recover */
 #define	TXN_DTL_RESTORED	0x2	/* prepared txn restored */
 #define	TXN_DTL_INMEMORY	0x4	/* uses in memory logs */
+#define	TXN_DTL_SNAPSHOT	0x8	/* On the list of snapshot txns. */
 	u_int32_t flags;
 
-	SH_TAILQ_ENTRY	links;		/* free/active list */
+	SH_TAILQ_ENTRY	links;		/* active/free/snapshot list */
 
-#define	TXN_XA_ABORTED		1
-#define	TXN_XA_DEADLOCKED	2
-#define	TXN_XA_ENDED		3
-#define	TXN_XA_PREPARED		4
-#define	TXN_XA_STARTED		5
-#define	TXN_XA_SUSPENDED	6
-	u_int32_t xa_status;		/* XA status */
-
-	/*
-	 * XID (xid_t) structure: because these fields are logged, the
-	 * sizes have to be explicit.
-	 */
-	u_int8_t xid[XIDDATASIZE];	/* XA global transaction id */
-	u_int32_t bqual;		/* bqual_length from XID */
-	u_int32_t gtrid;		/* gtrid_length from XID */
-	int32_t format;			/* XA format */
+	u_int8_t gid[DB_GID_SIZE];	/* global transaction id */
+	roff_t slots[TXN_NSLOTS];	/* Initial DB slot allocation. */
 } TXN_DETAIL;
 
 /*
@@ -98,17 +96,17 @@ struct __db_txnmgr {
 	u_int32_t n_discards;		/* Number of txns discarded. */
 
 	/* These fields are never updated after creation, so not protected. */
-	DB_ENV	*dbenv;			/* Environment. */
+	ENV	*env;			/* Environment. */
 	REGINFO	 reginfo;		/* Region information. */
 };
 
 /* Macros to lock/unlock the transaction region as a whole. */
-#define	TXN_SYSTEM_LOCK(dbenv)						\
-	MUTEX_LOCK(dbenv, ((DB_TXNREGION *)((DB_TXNMGR *)		\
-	    (dbenv)->tx_handle)->reginfo.primary)->mtx_region)
-#define	TXN_SYSTEM_UNLOCK(dbenv)					\
-	MUTEX_UNLOCK(dbenv, ((DB_TXNREGION *)((DB_TXNMGR *)		\
-	    (dbenv)->tx_handle)->reginfo.primary)->mtx_region)
+#define	TXN_SYSTEM_LOCK(env)						\
+	MUTEX_LOCK(env, ((DB_TXNREGION *)				\
+	    (env)->tx_handle->reginfo.primary)->mtx_region)
+#define	TXN_SYSTEM_UNLOCK(env)						\
+	MUTEX_UNLOCK(env, ((DB_TXNREGION *)				\
+	    (env)->tx_handle->reginfo.primary)->mtx_region)
 
 /*
  * DB_TXNREGION --
@@ -131,6 +129,7 @@ struct __db_txnregion {
 	u_int32_t	flags;
 					/* active TXN list */
 	SH_TAILQ_HEAD(__active) active_txn;
+	SH_TAILQ_HEAD(__mvcc) mvcc_txn;
 };
 
 /*
@@ -219,7 +218,10 @@ struct __txn_logrec {
 #define	TXN_EXPECTED	5
 #define	TXN_UNEXPECTED	6
 
+#if defined(__cplusplus)
+}
+#endif
+
 #include "dbinc_auto/txn_auto.h"
 #include "dbinc_auto/txn_ext.h"
-#include "dbinc_auto/xa_ext.h"
-#endif /* !_TXN_H_ */
+#endif /* !_DB_TXN_H_ */
