@@ -1,10 +1,9 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 2000-2005
- *      Sleepycat Software.  All rights reserved.
+ * Copyright (c) 2000-2009 Oracle.  All rights reserved.
  *
- * $Id: EntryBinding.java,v 12.1 2005/01/31 19:27:29 mark Exp $
+ * $Id$
  */
 
 package com.sleepycat.bind;
@@ -14,9 +13,15 @@ import com.sleepycat.db.DatabaseEntry;
 /**
  * A binding between a key or data entry and a key or data object.
  *
+ * <p><em>WARNING:</em> Binding instances are typically shared by multiple
+ * threads and binding methods are called without any special synchronization.
+ * Therefore, bindings must be thread safe.  In general no shared state should
+ * be used and any caching of computed values must be done with proper
+ * synchronization.</p>
+ *
  * @author Mark Hayes
  */
-public interface EntryBinding {
+public interface EntryBinding<E> {
 
     /**
      * Converts a entry buffer into an Object.
@@ -25,7 +30,7 @@ public interface EntryBinding {
      *
      * @return the resulting Object.
      */
-    Object entryToObject(DatabaseEntry entry);
+    E entryToObject(DatabaseEntry entry);
 
     /**
      * Converts an Object into a entry buffer.
@@ -34,5 +39,5 @@ public interface EntryBinding {
      *
      * @param entry is the destination entry buffer.
      */
-    void objectToEntry(Object object, DatabaseEntry entry);
+    void objectToEntry(E object, DatabaseEntry entry);
 }
