@@ -1,10 +1,9 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 2000-2005
- *      Sleepycat Software.  All rights reserved.
+ * Copyright (c) 2000-2009 Oracle.  All rights reserved.
  *
- * $Id: StoredMap.java,v 12.3 2005/10/05 20:40:10 mark Exp $
+ * $Id$
  */
 
 package com.sleepycat.collections;
@@ -14,40 +13,34 @@ import java.util.Collections;
 import java.util.Iterator;
 import java.util.Map;
 import java.util.Set;
+import java.util.concurrent.ConcurrentMap;
 
 import com.sleepycat.bind.EntityBinding;
 import com.sleepycat.bind.EntryBinding;
 import com.sleepycat.db.Database;
+import com.sleepycat.db.OperationStatus;
+import com.sleepycat.util.keyrange.KeyRangeException;
 
 /**
  * A Map view of a {@link Database}.
  *
- * <p><em>Note that this class does not conform to the standard Java
- * collections interface in the following ways:</em></p>
- * <ul>
- * <li>The {@link #size} method always throws
- * <code>UnsupportedOperationException</code> because, for performance reasons,
- * databases do not maintain their total record count.</li>
- * <li>All iterators must be explicitly closed using {@link
- * StoredIterator#close()} or {@link StoredIterator#close(java.util.Iterator)}
- * to release the underlying database cursor resources.</li>
- * </ul>
- *
  * <p>In addition to the standard Map methods, this class provides the
  * following methods for stored maps only.  Note that the use of these methods
  * is not compatible with the standard Java collections interface.</p>
  * <ul>
- * <li>{@link #duplicates(Object)}</li>
- * <li>{@link #append(Object)}</li>
+ * <li>{@link #duplicates}</li>
+ * <li>{@link #duplicatesMap}</li>
+ * <li>{@link #append}</li>
  * </ul>
  *
  * @author Mark Hayes
  */
-public class StoredMap extends StoredContainer implements Map {
+public class StoredMap<K,V> extends StoredContainer
+    implements ConcurrentMap<K,V> {
 
-    private StoredKeySet keySet;
-    private StoredEntrySet entrySet;
-    private StoredValueSet valueSet;
+    private StoredKeySet<K> keySet;
+    private StoredEntrySet<K,V> entrySet;
+    private StoredValueSet<V> valueSet;
 
     /**
      * Creates a map view of a {@link Database}.
@@ -69,8 +62,10 @@ public class StoredMap extends StoredContainer implements Map {
      * @throws RuntimeExceptionWrapper if a {@link
      * com.sleepycat.db.DatabaseException} is thrown.
      */
-    public StoredMap(Database database, EntryBinding keyBinding,
-                     EntryBinding valueBinding, boolean writeAllowed) {
+    public StoredMap(Database database,
+                     EntryBinding<K> keyBinding,
+                     EntryBinding<V> valueBinding,
+                     boolean writeAllowed) {
 
         super(new DataView(database, keyBinding, valueBinding, null,
                            writeAllowed, null));
@@ -98,8 +93,9 @@ public class StoredMap extends StoredContainer implements Map {
      * @throws RuntimeExceptionWrapper if a {@link
      * com.sleepycat.db.DatabaseException} is thrown.
      */
-    public StoredMap(Database database, EntryBinding keyBinding,
-                     EntryBinding valueBinding, 
+    public StoredMap(Database database,
+                     EntryBinding<K> keyBinding,
+                     EntryBinding<V> valueBinding,
                      PrimaryKeyAssigner keyAssigner) {
 
         super(new DataView(database, keyBinding, valueBinding, null,
@@ -127,8 +123,10 @@ public class StoredMap extends StoredContainer implements Map {
      * @throws RuntimeExceptionWrapper if a {@link
      * com.sleepycat.db.DatabaseException} is thrown.
      */
-    public StoredMap(Database database, EntryBinding keyBinding,
-                     EntityBinding valueEntityBinding, boolean writeAllowed) {
+    public StoredMap(Database database,
+                     EntryBinding<K> keyBinding,
+                     EntityBinding<V> valueEntityBinding,
+                     boolean writeAllowed) {
 
         super(new DataView(database, keyBinding, null, valueEntityBinding,
                            writeAllowed, null));
@@ -156,8 +154,9 @@ public class StoredMap extends StoredContainer implements Map {
      * @throws RuntimeExceptionWrapper if a {@link
      * com.sleepycat.db.DatabaseException} is thrown.
      */
-    public StoredMap(Database database, EntryBinding keyBinding,
-                     EntityBinding valueEntityBinding, 
+    public StoredMap(Database database,
+                     EntryBinding<K> keyBinding,
+                     EntityBinding<V> valueEntityBinding,
                      PrimaryKeyAssigner keyAssigner) {
 
         super(new DataView(database, keyBinding, null, valueEntityBinding,
@@ -189,26 +188,26 @@ public class StoredMap extends StoredContainer implements Map {
     private void initView() {
 
         /* entrySet */
-        if (isOrdered()) {
-            entrySet = new StoredSortedEntrySet(view);
+        if (areKeyRangesAllowed()) {
+            entrySet = new StoredSortedEntrySet<K,V>(view);
         } else {
-            entrySet = new StoredEntrySet(view);
+            entrySet = new StoredEntrySet<K,V>(view);
         }
 
         /* keySet */
         DataView newView = view.keySetView();
-        if (isOrdered()) {
-            keySet = new StoredSortedKeySet(newView);
+        if (areKeyRangesAllowed()) {
+            keySet = new StoredSortedKeySet<K>(newView);
         } else {
-            keySet = new StoredKeySet(newView);
+            keySet = new StoredKeySet<K>(newView);
         }
 
         /* valueSet */
         newView = view.valueSetView();
-        if (isOrdered() && newView.canDeriveKeyFromValue()) {
-            valueSet = new StoredSortedValueSet(newView);
+        if (areKeyRangesAllowed() && newView.canDeriveKeyFromValue()) {
+            valueSet = new StoredSortedValueSet<V>(newView);
         } else {
-            valueSet = new StoredValueSet(newView);
+            valueSet = new StoredValueSet<V>(newView);
         }
     }
 
@@ -217,15 +216,15 @@ public class StoredMap extends StoredContainer implements Map {
      * duplicates are allowed, this method returns the first duplicate, in the
      * order in which duplicates are configured, that maps to the specified
      * key.
-     * 
+     *
      * This method conforms to the {@link Map#get} interface.
      *
      * @throws RuntimeExceptionWrapper if a {@link
      * com.sleepycat.db.DatabaseException} is thrown.
      */
-    public Object get(Object key) {
+    public V get(Object key) {
 
-        return super.get(key);
+        return (V) getValue(key);
     }
 
     /**
@@ -253,9 +252,9 @@ public class StoredMap extends StoredContainer implements Map {
      * @throws RuntimeExceptionWrapper if a {@link
      * com.sleepycat.db.DatabaseException} is thrown.
      */
-    public Object put(Object key, Object value) {
+    public V put(K key, V value) {
 
-        return super.put(key, value);
+        return (V) putKeyValue(key, value);
     }
 
     /**
@@ -281,14 +280,14 @@ public class StoredMap extends StoredContainer implements Map {
      * @throws RuntimeExceptionWrapper if a {@link
      * com.sleepycat.db.DatabaseException} is thrown.
      */
-    public Object append(Object value) {
+    public K append(V value) {
 
         boolean doAutoCommit = beginAutoCommit();
         try {
             Object[] key = new Object[1];
             view.append(value, key, null);
             commitAutoCommit(doAutoCommit);
-            return key[0];
+            return (K) key[0];
         } catch (Exception e) {
             throw handleException(e, doAutoCommit);
         }
@@ -305,11 +304,151 @@ public class StoredMap extends StoredContainer implements Map {
      * @throws RuntimeExceptionWrapper if a {@link
      * com.sleepycat.db.DatabaseException} is thrown.
      */
-    public Object remove(Object key) {
+    public V remove(Object key) {
 
         Object[] oldVal = new Object[1];
         removeKey(key, oldVal);
-        return oldVal[0];
+        return (V) oldVal[0];
+    }
+
+    /**
+     * If the specified key is not already associated with a value, associate
+     * it with the given value.  This method conforms to the {@link
+     * ConcurrentMap#putIfAbsent} interface.
+     *
+     * @throws RuntimeExceptionWrapper if a {@link
+     * com.sleepycat.db.DatabaseException} is thrown.
+     */
+    public V putIfAbsent(K key, V value) {
+        DataCursor cursor = null;
+        boolean doAutoCommit = beginAutoCommit();
+        try {
+            cursor = new DataCursor(view, true);
+            V oldValue;
+            while (true) {
+                OperationStatus status =
+                    cursor.putNoOverwrite(key, value, false /*useCurrentKey*/);
+                if (status == OperationStatus.SUCCESS) {
+                    /* We inserted the key.  Return null.  */
+                    oldValue = null;
+                    break;
+                } else {
+                    status = cursor.getSearchKey(key, null /*value*/,
+                                                 false /*lockForWrite*/);
+                    if (status == OperationStatus.SUCCESS) {
+                        /* The key is present. Return the current value. */
+                        oldValue = (V) cursor.getCurrentValue();
+                        break;
+                    } else {
+
+                        /*
+                         * If Serializable isolation is not configured, another
+                         * thread can delete the record after our attempt to
+                         * insert it failed above.  Loop back and try again.
+                         */
+                        continue;
+                    }
+                }
+            }
+            closeCursor(cursor);
+            commitAutoCommit(doAutoCommit);
+            return oldValue;
+        } catch (Exception e) {
+            closeCursor(cursor);
+            throw handleException(e, doAutoCommit);
+        }
+    }
+
+    /**
+     * Remove entry for key only if currently mapped to given value.  This
+     * method conforms to the {@link ConcurrentMap#remove(Object,Object)}
+     * interface.
+     *
+     * @throws RuntimeExceptionWrapper if a {@link
+     * com.sleepycat.db.DatabaseException} is thrown.
+     */
+    public boolean remove(Object key, Object value) {
+        DataCursor cursor = null;
+        boolean doAutoCommit = beginAutoCommit();
+        try {
+            cursor = new DataCursor(view, true, key);
+            OperationStatus status = cursor.getFirst(true /*lockForWrite*/);
+            boolean removed;
+            if (status == OperationStatus.SUCCESS &&
+                cursor.getCurrentValue().equals(value)) {
+                cursor.delete();
+                removed = true;
+            } else {
+                removed = false;
+            }
+            closeCursor(cursor);
+            commitAutoCommit(doAutoCommit);
+            return removed;
+        } catch (Exception e) {
+            closeCursor(cursor);
+            throw handleException(e, doAutoCommit);
+        }
+    }
+
+    /**
+     * Replace entry for key only if currently mapped to some value.  This
+     * method conforms to the {@link ConcurrentMap#replace(Object,Object)}
+     * interface.
+     *
+     * @throws RuntimeExceptionWrapper if a {@link
+     * com.sleepycat.db.DatabaseException} is thrown.
+     */
+    public V replace(K key, V value) {
+        DataCursor cursor = null;
+        boolean doAutoCommit = beginAutoCommit();
+        try {
+            cursor = new DataCursor(view, true, key);
+            OperationStatus status = cursor.getFirst(true /*lockForWrite*/);
+            V oldValue;
+            if (status == OperationStatus.SUCCESS) {
+                oldValue = (V) cursor.getCurrentValue();
+                cursor.putCurrent(value);
+            } else {
+                oldValue = null;
+            }
+            closeCursor(cursor);
+            commitAutoCommit(doAutoCommit);
+            return oldValue;
+        } catch (Exception e) {
+            closeCursor(cursor);
+            throw handleException(e, doAutoCommit);
+        }
+    }
+
+    /**
+     * Replace entry for key only if currently mapped to given value.  This
+     * method conforms to the {@link
+     * ConcurrentMap#replace(Object,Object,Object)} interface.
+     *
+     * @throws RuntimeExceptionWrapper if a {@link
+     * com.sleepycat.db.DatabaseException} is thrown.
+     */
+    public boolean replace(K key, V oldValue, V newValue) {
+        DataCursor cursor = null;
+        boolean doAutoCommit = beginAutoCommit();
+        try {
+            cursor = new DataCursor(view, true, key);
+            OperationStatus status = cursor.getFirst(true /*lockForWrite*/);
+            boolean replaced;
+            if (status == OperationStatus.SUCCESS &&
+                cursor.getCurrentValue().equals(oldValue)) {
+                cursor.putCurrent(newValue);
+                replaced = true;
+            } else {
+                replaced = false;
+            }
+            closeCursor(cursor);
+            commitAutoCommit(doAutoCommit);
+            return replaced;
+        } catch (Exception e) {
+            closeCursor(cursor);
+            throw handleException(e, doAutoCommit);
+        }
     }
 
     /**
@@ -351,29 +490,30 @@ public class StoredMap extends StoredContainer implements Map {
      * @throws RuntimeExceptionWrapper if a {@link
      * com.sleepycat.db.DatabaseException} is thrown.
      */
-    public void putAll(Map map) {
+    public void putAll(Map<? extends K, ? extends V> map) {
 
         boolean doAutoCommit = beginAutoCommit();
-        Iterator entries = null;
+        Iterator i = null;
         try {
-            entries = map.entrySet().iterator();
-            while (entries.hasNext()) {
-                Map.Entry entry = (Map.Entry) entries.next();
-                put(entry.getKey(), entry.getValue());
+            Collection coll = map.entrySet();
+            i = storedOrExternalIterator(coll);
+            while (i.hasNext()) {
+                Map.Entry entry = (Map.Entry) i.next();
+                putKeyValue(entry.getKey(), entry.getValue());
             }
-            StoredIterator.close(entries);
+            StoredIterator.close(i);
             commitAutoCommit(doAutoCommit);
         } catch (Exception e) {
-            StoredIterator.close(entries);
+            StoredIterator.close(i);
             throw handleException(e, doAutoCommit);
         }
     }
 
     /**
      * Returns a set view of the keys contained in this map.  A {@link
-     * java.util.SortedSet} is returned if the map is ordered.  The returned
-     * collection will be read-only if the map is read-only.  This method
-     * conforms to the {@link Map#keySet()} interface.
+     * java.util.SortedSet} is returned if the map supports key ranges.  The
+     * returned collection will be read-only if the map is read-only.  This
+     * method conforms to the {@link Map#keySet()} interface.
      *
      * <p>Note that the return value is a StoredCollection and must be treated
      * as such; for example, its iterators must be explicitly closed.</p>
@@ -384,19 +524,19 @@ public class StoredMap extends StoredContainer implements Map {
      * @throws RuntimeExceptionWrapper if a {@link
      * com.sleepycat.db.DatabaseException} is thrown.
      *
-     * @see #isOrdered
+     * @see #areKeyRangesAllowed
      * @see #isWriteAllowed
      */
-    public Set keySet() {
+    public Set<K> keySet() {
 
         return keySet;
     }
 
     /**
      * Returns a set view of the mappings contained in this map.  A {@link
-     * java.util.SortedSet} is returned if the map is ordered.  The returned
-     * collection will be read-only if the map is read-only.  This method
-     * conforms to the {@link Map#entrySet()} interface.
+     * java.util.SortedSet} is returned if the map supports key ranges.  The
+     * returned collection will be read-only if the map is read-only.  This
+     * method conforms to the {@link Map#entrySet()} interface.
      *
      * <p>Note that the return value is a StoredCollection and must be treated
      * as such; for example, its iterators must be explicitly closed.</p>
@@ -407,17 +547,17 @@ public class StoredMap extends StoredContainer implements Map {
      * @throws RuntimeExceptionWrapper if a {@link
      * com.sleepycat.db.DatabaseException} is thrown.
      *
-     * @see #isOrdered
+     * @see #areKeyRangesAllowed
      * @see #isWriteAllowed
      */
-    public Set entrySet() {
+    public Set<Map.Entry<K,V>> entrySet() {
 
         return entrySet;
     }
 
     /**
      * Returns a collection view of the values contained in this map.  A {@link
-     * java.util.SortedSet} is returned if the map is ordered and the
+     * java.util.SortedSet} is returned if the map supports key ranges and the
      * value/entity binding can be used to derive the map's key from its
      * value/entity object.  The returned collection will be read-only if the
      * map is read-only.  This method conforms to the {@link Map#values()}
@@ -432,10 +572,10 @@ public class StoredMap extends StoredContainer implements Map {
      * @throws RuntimeExceptionWrapper if a {@link
      * com.sleepycat.db.DatabaseException} is thrown.
      *
-     * @see #isOrdered
+     * @see #areKeyRangesAllowed
      * @see #isWriteAllowed
      */
-    public Collection values() {
+    public Collection<V> values() {
 
         return valueSet;
     }
@@ -447,9 +587,6 @@ public class StoredMap extends StoredContainer implements Map {
      * supported by the standard Map interface.  This method does not exist in
      * the standard {@link Map} interface.
      *
-     * <p>Note that the return value is a StoredCollection and must be treated
-     * as such; for example, its iterators must be explicitly closed.</p>
-     *
      * <p>If no mapping for the given key is present, an empty collection is
      * returned.  If duplicates are not allowed, at most a single value will be
      * in the collection returned.  If duplicates are allowed, the returned
@@ -461,11 +598,11 @@ public class StoredMap extends StoredContainer implements Map {
      * @throws RuntimeExceptionWrapper if a {@link
      * com.sleepycat.db.DatabaseException} is thrown.
      */
-    public Collection duplicates(Object key) {
+    public Collection<V> duplicates(K key) {
 
         try {
             DataView newView = view.valueSetView(key);
-            return new StoredValueSet(newView, true);
+            return new StoredValueSet(newView);
         } catch (KeyRangeException e) {
             return Collections.EMPTY_SET;
         } catch (Exception e) {
@@ -473,6 +610,41 @@ public class StoredMap extends StoredContainer implements Map {
         }
     }
 
+    /**
+     * Returns a new map from primary key to value for the subset of records
+     * having a given secondary key (duplicates).  This method does not exist
+     * in the standard {@link Map} interface.
+     *
+     * <p>If no mapping for the given key is present, an empty collection is
+     * returned.  If duplicates are not allowed, at most a single value will be
+     * in the collection returned.  If duplicates are allowed, the returned
+     * collection's add() method may be used to add values for the given
+     * key.</p>
+     *
+     * @param secondaryKey is the secondary key for which duplicates values
+     * will be represented by the returned map.
+     *
+     * @param primaryKeyBinding is the binding used for keys in the returned
+     * map.
+     *
+     * @throws RuntimeExceptionWrapper if a {@link
+     * com.sleepycat.db.DatabaseException} is thrown.
+     */
+    public <PK> Map<PK,V> duplicatesMap(K secondaryKey,
+                                        EntryBinding primaryKeyBinding) {
+        try {
+            DataView newView =
+                view.duplicatesView(secondaryKey, primaryKeyBinding);
+            if (isOrdered()) {
+                return new StoredSortedMap(newView);
+            } else {
+                return new StoredMap(newView);
+            }
+        } catch (Exception e) {
+            throw StoredContainer.convertException(e);
+        }
+    }
+
     /**
      * Compares the specified object with this map for equality.  A value
      * comparison is performed by this method and the stored values are
@@ -499,6 +671,11 @@ public class StoredMap extends StoredContainer implements Map {
 	return super.hashCode();
     }
 
+    // Inherit javadoc
+    public int size() {
+        return values().size();
+    }
+
     /**
      * Converts the map to a string representation for debugging.  WARNING: All
      * mappings will be converted to strings and returned and therefore the
@@ -514,4 +691,3 @@ public class StoredMap extends StoredContainer implements Map {
         return entrySet().toString();
     }
 }
-
