@@ -1,18 +1,13 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1999-2005
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1999-2009 Oracle.  All rights reserved.
  *
- * $Id: qam_conv.c,v 12.1 2005/06/16 20:23:32 bostic Exp $
+ * $Id$
  */
 
 #include "db_config.h"
 
-#ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-#endif
-
 #include "db_int.h"
 #include "dbinc/db_page.h"
 #include "dbinc/db_swap.h"
@@ -23,16 +18,18 @@
  * __qam_mswap --
  *	Swap the bytes on the queue metadata page.
  *
- * PUBLIC: int __qam_mswap __P((PAGE *));
+ * PUBLIC: int __qam_mswap __P((ENV *, PAGE *));
  */
 int
-__qam_mswap(pg)
+__qam_mswap(env, pg)
+	ENV *env;
 	PAGE *pg;
 {
 	u_int8_t *p;
 
-	 __db_metaswap(pg);
+	COMPQUIET(env, NULL);
 
+	 __db_metaswap(pg);
 	 p = (u_int8_t *)pg + sizeof(DBMETA);
 
 	SWAP32(p);		/* first_recno */
@@ -53,11 +50,11 @@ __qam_mswap(pg)
  *	stored on disk.
  *  We only need to fix up a few fields in the header
  *
- * PUBLIC: int __qam_pgin_out __P((DB_ENV *, db_pgno_t, void *, DBT *));
+ * PUBLIC: int __qam_pgin_out __P((ENV *, db_pgno_t, void *, DBT *));
  */
 int
-__qam_pgin_out(dbenv, pg, pp, cookie)
-	DB_ENV *dbenv;
+__qam_pgin_out(env, pg, pp, cookie)
+	ENV *env;
 	db_pgno_t pg;
 	void *pp;
 	DBT *cookie;
@@ -66,14 +63,13 @@ __qam_pgin_out(dbenv, pg, pp, cookie)
 	QPAGE *h;
 
 	COMPQUIET(pg, 0);
-	COMPQUIET(dbenv, NULL);
 	pginfo = (DB_PGINFO *)cookie->data;
 	if (!F_ISSET(pginfo, DB_AM_SWAP))
 		return (0);
 
 	h = pp;
 	if (h->type == P_QAMMETA)
-	    return (__qam_mswap(pp));
+	    return (__qam_mswap(env, pp));
 
 	M_32_SWAP(h->lsn.file);
 	M_32_SWAP(h->lsn.offset);
