@@ -1,24 +1,17 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1999-2005
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1999-2009 Oracle.  All rights reserved.
  *
- * $Id: tcl_env.c,v 12.15 2005/11/02 20:21:37 bostic Exp $
+ * $Id$
  */
 
 #include "db_config.h"
 
-#ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-
-#include <stdlib.h>
-#include <string.h>
+#include "db_int.h"
+#ifdef HAVE_SYSTEM_INCLUDE_FILES
 #include <tcl.h>
 #endif
-
-#include "db_int.h"
-#include "dbinc/db_shash.h"
 #include "dbinc/lock.h"
 #include "dbinc/txn.h"
 #include "dbinc/tcl_db.h"
@@ -55,6 +48,8 @@ env_Cmd(clientData, interp, objc, objv)
 		"attributes",
 		"errfile",
 		"errpfx",
+		"event",
+		"failchk",
 		"id_reset",
 		"lock_detect",
 		"lock_id",
@@ -66,10 +61,12 @@ env_Cmd(clientData, interp, objc, objv)
 		"lock_vec",
 		"log_archive",
 		"log_compare",
+		"log_config",
 		"log_cursor",
 		"log_file",
 		"log_flush",
 		"log_get",
+		"log_get_config",
 		"log_put",
 		"log_stat",
 		"lsn_reset",
@@ -77,10 +74,26 @@ env_Cmd(clientData, interp, objc, objv)
 		"mpool_stat",
 		"mpool_sync",
 		"mpool_trickle",
+		"mutex",
+		"mutex_free",
+		"mutex_get_align",
+		"mutex_get_incr",
+		"mutex_get_max",
+		"mutex_get_tas_spins",
+		"mutex_lock",
+		"mutex_set_tas_spins",
+		"mutex_stat",
+		"mutex_unlock",
 		"rep_config",
 		"rep_elect",
 		"rep_flush",
+		"rep_get_clockskew",
 		"rep_get_config",
+		"rep_get_limit",
+		"rep_get_nsites",
+		"rep_get_request",
+		"rep_get_timeout",
+		"rep_lease",
 		"rep_limit",
 		"rep_process_message",
 		"rep_request",
@@ -88,6 +101,9 @@ env_Cmd(clientData, interp, objc, objv)
 		"rep_stat",
 		"rep_sync",
 		"rep_transport",
+		"repmgr",
+		"repmgr_site_list",
+		"repmgr_stat",
 		"rpcid",
 		"set_flags",
 		"test",
@@ -97,10 +113,12 @@ env_Cmd(clientData, interp, objc, objv)
 		"txn_timeout",
 		"verbose",
 #endif
+		"cdsgroup",
 		"close",
 		"dbremove",
 		"dbrename",
 		"get_cachesize",
+		"get_cache_max",
 		"get_data_dirs",
 		"get_encrypt_flags",
 		"get_errpfx",
@@ -119,7 +137,6 @@ env_Cmd(clientData, interp, objc, objv)
 		"get_mp_max_write",
 		"get_mp_mmapsize",
 		"get_open_flags",
-		"get_rep_limit",
 		"get_shm_key",
 		"get_tas_spins",
 		"get_timeout",
@@ -127,6 +144,8 @@ env_Cmd(clientData, interp, objc, objv)
 		"get_tx_max",
 		"get_tx_timestamp",
 		"get_verbose",
+		"resize_cache",
+		"set_data_dir",
 		"txn",
 		"txn_checkpoint",
 		NULL
@@ -136,6 +155,8 @@ env_Cmd(clientData, interp, objc, objv)
 		ENVATTR,
 		ENVERRFILE,
 		ENVERRPFX,
+		ENVEVENT,
+		ENVFAILCHK,
 		ENVIDRESET,
 		ENVLKDETECT,
 		ENVLKID,
@@ -147,10 +168,12 @@ env_Cmd(clientData, interp, objc, objv)
 		ENVLKVEC,
 		ENVLOGARCH,
 		ENVLOGCMP,
+		ENVLOGCONFIG,
 		ENVLOGCURSOR,
 		ENVLOGFILE,
 		ENVLOGFLUSH,
 		ENVLOGGET,
+		ENVLOGGETCONFIG,
 		ENVLOGPUT,
 		ENVLOGSTAT,
 		ENVLSNRESET,
@@ -158,10 +181,26 @@ env_Cmd(clientData, interp, objc, objv)
 		ENVMPSTAT,
 		ENVMPSYNC,
 		ENVTRICKLE,
+		ENVMUTEX,
+		ENVMUTFREE,
+		ENVMUTGETALIGN,
+		ENVMUTGETINCR,
+		ENVMUTGETMAX,
+		ENVMUTGETTASSPINS,
+		ENVMUTLOCK,
+		ENVMUTSETTASSPINS,
+		ENVMUTSTAT,
+		ENVMUTUNLOCK,
 		ENVREPCONFIG,
 		ENVREPELECT,
 		ENVREPFLUSH,
+		ENVREPGETCLOCKSKEW,
 		ENVREPGETCONFIG,
+		ENVREPGETLIMIT,
+		ENVREPGETNSITES,
+		ENVREPGETREQUEST,
+		ENVREPGETTIMEOUT,
+		ENVREPLEASE,
 		ENVREPLIMIT,
 		ENVREPPROCMESS,
 		ENVREPREQUEST,
@@ -169,6 +208,9 @@ env_Cmd(clientData, interp, objc, objv)
 		ENVREPSTAT,
 		ENVREPSYNC,
 		ENVREPTRANSPORT,
+		ENVREPMGR,
+		ENVREPMGRSITELIST,
+		ENVREPMGRSTAT,
 		ENVRPCID,
 		ENVSETFLAGS,
 		ENVTEST,
@@ -178,10 +220,12 @@ env_Cmd(clientData, interp, objc, objv)
 		ENVTXNTIMEOUT,
 		ENVVERB,
 #endif
+		ENVCDSGROUP,
 		ENVCLOSE,
 		ENVDBREMOVE,
 		ENVDBRENAME,
 		ENVGETCACHESIZE,
+		ENVGETCACHEMAX,
 		ENVGETDATADIRS,
 		ENVGETENCRYPTFLAGS,
 		ENVGETERRPFX,
@@ -200,7 +244,6 @@ env_Cmd(clientData, interp, objc, objv)
 		ENVGETMPMAXWRITE,
 		ENVGETMPMMAPSIZE,
 		ENVGETOPENFLAG,
-		ENVGETREPLIMIT,
 		ENVGETSHMKEY,
 		ENVGETTASSPINS,
 		ENVGETTIMEOUT,
@@ -208,27 +251,27 @@ env_Cmd(clientData, interp, objc, objv)
 		ENVGETTXMAX,
 		ENVGETTXTIMESTAMP,
 		ENVGETVERBOSE,
+		ENVRESIZECACHE,
+		ENVSETDATADIR,
 		ENVTXN,
 		ENVTXNCKP
 	};
 	DBTCL_INFO *envip;
 	DB_ENV *dbenv;
-	Tcl_Obj *myobjv[3], *res;
-	char newname[MSG_SIZE];
-	int cmdindex, i, intvalue1, intvalue2, ncache, result, ret;
-	u_int32_t bytes, gbytes, value;
+	Tcl_Obj **listobjv, *myobjv[3], *res;
+	db_timeout_t timeout;
 	size_t size;
-	long shm_key;
 	time_t timeval;
+	u_int32_t bytes, gbytes, value;
+	long shm_key;
+	int cmdindex, i, intvalue, listobjc, ncache, result, ret;
 	const char *strval, **dirs;
+	char *strarg, newname[MSG_SIZE];
 #ifdef CONFIG_TEST
 	DBTCL_INFO *logcip;
 	DB_LOGC *logc;
-	Tcl_Obj **repobjv;
 	u_int32_t lockid;
 	long newval, otherval;
-	int repobjc;
-	char *strarg;
 #endif
 
 	Tcl_ResetResult(interp);
@@ -260,6 +303,29 @@ env_Cmd(clientData, interp, objc, objv)
 	res = NULL;
 	switch ((enum envcmds)cmdindex) {
 #ifdef CONFIG_TEST
+	case ENVEVENT:
+		/*
+		 * Two args for this.  Error if different.
+		 */
+		if (objc != 3) {
+			Tcl_WrongNumArgs(interp, 2, objv, NULL);
+			return (TCL_ERROR);
+		}
+		result = tcl_EventNotify(interp, dbenv, objv[2], envip);
+		break;
+	case ENVFAILCHK:
+		/*
+		 * No args for this.  Error if there are some.
+		 */
+		if (objc > 2) {
+			Tcl_WrongNumArgs(interp, 2, objv, NULL);
+			return (TCL_ERROR);
+		}
+		_debug_check();
+		ret = dbenv->failchk(dbenv, 0);
+		result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
+		    "failchk");
+		break;
 	case ENVIDRESET:
 		result = tcl_EnvIdReset(interp, objc, objv, dbenv);
 		break;
@@ -313,7 +379,7 @@ env_Cmd(clientData, interp, objc, objv)
 		result = Tcl_GetLongFromObj(interp, objv[3], &otherval);
 		if (result != TCL_OK)
 			return (result);
-		ret = __lock_id_set(dbenv,
+		ret = __lock_id_set(dbenv->env,
 		    (u_int32_t)newval, (u_int32_t)otherval);
 		result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
 		    "lock id_free");
@@ -330,6 +396,16 @@ env_Cmd(clientData, interp, objc, objv)
 	case ENVLOGCMP:
 		result = tcl_LogCompare(interp, objc, objv);
 		break;
+	case ENVLOGCONFIG:
+		/*
+		 * Two args for this.  Error if different.
+		 */
+		if (objc != 3) {
+			Tcl_WrongNumArgs(interp, 2, objv, NULL);
+			return (TCL_ERROR);
+		}
+		result = tcl_LogConfig(interp, dbenv, objv[2]);
+		break;
 	case ENVLOGCURSOR:
 		snprintf(newname, sizeof(newname),
 		    "%s.logc%d", envip->i_name, envip->i_envlogcid);
@@ -369,6 +445,16 @@ env_Cmd(clientData, interp, objc, objv)
 	case ENVLOGGET:
 		result = tcl_LogGet(interp, objc, objv, dbenv);
 		break;
+	case ENVLOGGETCONFIG:
+		/*
+		 * Two args for this.  Error if different.
+		 */
+		if (objc != 3) {
+			Tcl_WrongNumArgs(interp, 2, objv, NULL);
+			return (TCL_ERROR);
+		}
+		result = tcl_LogGetConfig(interp, dbenv, objv[2]);
+		break;
 	case ENVLOGPUT:
 		result = tcl_LogPut(interp, objc, objv, dbenv);
 		break;
@@ -387,6 +473,40 @@ env_Cmd(clientData, interp, objc, objv)
 	case ENVMP:
 		result = tcl_Mp(interp, objc, objv, dbenv, envip);
 		break;
+	case ENVMUTEX:
+		result = tcl_Mutex(interp, objc, objv, dbenv);
+		break;
+	case ENVMUTFREE:
+		result = tcl_MutFree(interp, objc, objv, dbenv);
+		break;
+	case ENVMUTGETALIGN:
+		result = tcl_MutGet(interp, dbenv, DBTCL_MUT_ALIGN);
+		break;
+	case ENVMUTGETINCR:
+		result = tcl_MutGet(interp, dbenv, DBTCL_MUT_INCR);
+		break;
+	case ENVMUTGETMAX:
+		result = tcl_MutGet(interp, dbenv, DBTCL_MUT_MAX);
+		break;
+	case ENVMUTGETTASSPINS:
+		result = tcl_MutGet(interp, dbenv, DBTCL_MUT_TAS);
+		break;
+	case ENVMUTLOCK:
+		result = tcl_MutLock(interp, objc, objv, dbenv);
+		break;
+	case ENVMUTSETTASSPINS:
+		if (objc != 3) {
+			Tcl_WrongNumArgs(interp, 2, objv, NULL);
+			return (TCL_ERROR);
+		}
+		result = tcl_MutSet(interp, objv[2], dbenv, DBTCL_MUT_TAS);
+		break;
+	case ENVMUTSTAT:
+		result = tcl_MutStat(interp, objc, objv, dbenv);
+		break;
+	case ENVMUTUNLOCK:
+		result = tcl_MutUnlock(interp, objc, objv, dbenv);
+		break;
 	case ENVREPCONFIG:
 		/*
 		 * Two args for this.  Error if different.
@@ -403,6 +523,9 @@ env_Cmd(clientData, interp, objc, objv)
 	case ENVREPFLUSH:
 		result = tcl_RepFlush(interp, objc, objv, dbenv);
 		break;
+	case ENVREPGETCLOCKSKEW:
+		result = tcl_RepGetTwo(interp, dbenv, DBTCL_GETCLOCK);
+		break;
 	case ENVREPGETCONFIG:
 		/*
 		 * Two args for this.  Error if different.
@@ -413,6 +536,43 @@ env_Cmd(clientData, interp, objc, objv)
 		}
 		result = tcl_RepGetConfig(interp, dbenv, objv[2]);
 		break;
+	case ENVREPGETLIMIT:
+		result = tcl_RepGetTwo(interp, dbenv, DBTCL_GETLIMIT);
+		break;
+	case ENVREPGETNSITES:
+		if (objc != 2) {
+			Tcl_WrongNumArgs(interp, 1, objv, NULL);
+			return (TCL_ERROR);
+		}
+		ret = dbenv->rep_get_nsites(dbenv, &value);
+		if ((result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
+		    "env rep_get_nsites")) == TCL_OK)
+			res = Tcl_NewLongObj((long)value);
+		break;
+	case ENVREPGETREQUEST:
+		result = tcl_RepGetTwo(interp, dbenv, DBTCL_GETREQ);
+		break;
+	case ENVREPGETTIMEOUT:
+		/*
+		 * Two args for this.  Error if different.
+		 */
+		if (objc != 3) {
+			Tcl_WrongNumArgs(interp, 2, objv, NULL);
+			return (TCL_ERROR);
+		}
+		result = tcl_RepGetTimeout(interp, dbenv, objv[2]);
+		break;
+	case ENVREPLEASE:
+		if (objc != 3) {
+			Tcl_WrongNumArgs(interp, 2, objv, NULL);
+			return (TCL_ERROR);
+		}
+		result = Tcl_ListObjGetElements(interp, objv[2],
+		    &listobjc, &listobjv);
+		if (result == TCL_OK)
+			result = tcl_RepLease(interp,
+			    listobjc, listobjv, dbenv);
+		break;
 	case ENVREPLIMIT:
 		result = tcl_RepLimit(interp, objc, objv, dbenv);
 		break;
@@ -437,10 +597,19 @@ env_Cmd(clientData, interp, objc, objv)
 			return (TCL_ERROR);
 		}
 		result = Tcl_ListObjGetElements(interp, objv[2],
-		    &repobjc, &repobjv);
+		    &listobjc, &listobjv);
 		if (result == TCL_OK)
 			result = tcl_RepTransport(interp,
-			    repobjc, repobjv, dbenv, envip);
+			    listobjc, listobjv, dbenv, envip);
+		break;
+	case ENVREPMGR:
+		result = tcl_RepMgr(interp, objc, objv, dbenv);
+		break;
+	case ENVREPMGRSITELIST:
+		result = tcl_RepMgrSiteList(interp, objc, objv, dbenv);
+		break;
+	case ENVREPMGRSTAT:
+		result = tcl_RepMgrStat(interp, objc, objv, dbenv);
 		break;
 	case ENVRPCID:
 		/*
@@ -452,7 +621,7 @@ env_Cmd(clientData, interp, objc, objv)
 		}
 		/*
 		 * !!! Retrieve the client ID from the dbp handle directly.
-		 * This is for testing purposes only.  It is dbp-private data.
+		 * This is for testing purposes only.  It is BDB-private data.
 		 */
 		res = Tcl_NewLongObj((long)dbenv->cl_id);
 		break;
@@ -467,7 +636,7 @@ env_Cmd(clientData, interp, objc, objv)
 		result = Tcl_GetLongFromObj(interp, objv[3], &otherval);
 		if (result != TCL_OK)
 			return (result);
-		ret = __txn_id_set(dbenv,
+		ret = __txn_id_set(dbenv->env,
 		    (u_int32_t)newval, (u_int32_t)otherval);
 		result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
 		    "txn setid");
@@ -531,6 +700,9 @@ env_Cmd(clientData, interp, objc, objv)
 		result = tcl_EnvVerbose(interp, dbenv, objv[2], objv[3]);
 		break;
 #endif
+	case ENVCDSGROUP:
+		result = tcl_CDSGroup(interp, objc, objv, dbenv, envip);
+		break;
 	case ENVCLOSE:
 		/*
 		 * No args for this.  Error if there are some.
@@ -574,6 +746,19 @@ env_Cmd(clientData, interp, objc, objv)
 			res = Tcl_NewListObj(3, myobjv);
 		}
 		break;
+	case ENVGETCACHEMAX:
+		if (objc != 2) {
+			Tcl_WrongNumArgs(interp, 1, objv, NULL);
+			return (TCL_ERROR);
+		}
+		ret = dbenv->get_cache_max(dbenv, &gbytes, &bytes);
+		if ((result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
+		    "env get_cache_max")) == TCL_OK) {
+			myobjv[0] = Tcl_NewLongObj((long)gbytes);
+			myobjv[1] = Tcl_NewLongObj((long)bytes);
+			res = Tcl_NewListObj(2, myobjv);
+		}
+		break;
 	case ENVGETDATADIRS:
 		if (objc != 2) {
 			Tcl_WrongNumArgs(interp, 1, objv, NULL);
@@ -637,10 +822,10 @@ env_Cmd(clientData, interp, objc, objv)
 			Tcl_WrongNumArgs(interp, 1, objv, NULL);
 			return (TCL_ERROR);
 		}
-		ret = dbenv->get_lg_filemode(dbenv, &intvalue1);
+		ret = dbenv->get_lg_filemode(dbenv, &intvalue);
 		if ((result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
 		    "env get_lg_filemode")) == TCL_OK)
-			res = Tcl_NewLongObj((long)intvalue1);
+			res = Tcl_NewLongObj((long)intvalue);
 		break;
 	case ENVGETLGMAX:
 		if (objc != 2) {
@@ -700,21 +885,21 @@ env_Cmd(clientData, interp, objc, objv)
 			Tcl_WrongNumArgs(interp, 1, objv, NULL);
 			return (TCL_ERROR);
 		}
-		ret = dbenv->get_mp_max_openfd(dbenv, &intvalue1);
+		ret = dbenv->get_mp_max_openfd(dbenv, &intvalue);
 		if ((result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
 		    "env get_mp_max_openfd")) == TCL_OK)
-			res = Tcl_NewIntObj(intvalue1);
+			res = Tcl_NewIntObj(intvalue);
 		break;
 	case ENVGETMPMAXWRITE:
 		if (objc != 2) {
 			Tcl_WrongNumArgs(interp, 1, objv, NULL);
 			return (TCL_ERROR);
 		}
-		ret = dbenv->get_mp_max_write(dbenv, &intvalue1, &intvalue2);
+		ret = dbenv->get_mp_max_write(dbenv, &intvalue, &timeout);
 		if ((result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
 		    "env get_mp_max_write")) == TCL_OK) {
-			myobjv[0] = Tcl_NewIntObj(intvalue1);
-			myobjv[1] = Tcl_NewIntObj(intvalue2);
+			myobjv[0] = Tcl_NewIntObj(intvalue);
+			myobjv[1] = Tcl_NewIntObj((int)timeout);
 			res = Tcl_NewListObj(2, myobjv);
 		}
 		break;
@@ -731,19 +916,6 @@ env_Cmd(clientData, interp, objc, objv)
 	case ENVGETOPENFLAG:
 		result = env_GetOpenFlag(interp, objc, objv, dbenv);
 		break;
-	case ENVGETREPLIMIT:
-		if (objc != 2) {
-			Tcl_WrongNumArgs(interp, 1, objv, NULL);
-			return (TCL_ERROR);
-		}
-		ret = dbenv->get_rep_limit(dbenv, &gbytes, &bytes);
-		if ((result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
-		    "env get_rep_limit")) == TCL_OK) {
-			myobjv[0] = Tcl_NewLongObj((long)gbytes);
-			myobjv[1] = Tcl_NewLongObj((long)bytes);
-			res = Tcl_NewListObj(2, myobjv);
-		}
-		break;
 	case ENVGETSHMKEY:
 		if (objc != 2) {
 			Tcl_WrongNumArgs(interp, 1, objv, NULL);
@@ -800,6 +972,38 @@ env_Cmd(clientData, interp, objc, objv)
 	case ENVGETVERBOSE:
 		result = env_GetVerbose(interp, objc, objv, dbenv);
 		break;
+	case ENVRESIZECACHE:
+		if ((result = Tcl_ListObjGetElements(
+		    interp, objv[2], &listobjc, &listobjv)) != TCL_OK)
+			break;
+		if (objc != 3 || listobjc != 2) {
+			Tcl_WrongNumArgs(interp, 2, objv,
+			    "?-resize_cache {gbytes bytes}?");
+			result = TCL_ERROR;
+			break;
+		}
+		result = _GetUInt32(interp, listobjv[0], &gbytes);
+		if (result != TCL_OK)
+			break;
+		result = _GetUInt32(interp, listobjv[1], &bytes);
+		if (result != TCL_OK)
+			break;
+		ret = dbenv->set_cachesize(dbenv, gbytes, bytes, 0);
+		result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
+		    "resize_cache");
+		break;
+	case ENVSETDATADIR:
+		/*
+		 * One args for this.  Error if different.
+		 */
+		if (objc != 3) {
+			Tcl_WrongNumArgs(interp, 2, objv, "pfx");
+			return (TCL_ERROR);
+		}
+		strarg = Tcl_GetStringFromObj(objv[2], NULL);
+		ret = dbenv->set_data_dir(dbenv, strarg);
+		return (_ReturnSetup(interp, ret, DB_RETOK_STD(ret),
+		    "env set data dir"));
 	case ENVTXN:
 		result = tcl_Txn(interp, objc, objv, dbenv, envip);
 		break;
@@ -861,7 +1065,6 @@ tcl_EnvRemove(interp, objc, objv, dbenv, envip)
 		ENVREM_USE_ENVIRON,
 		ENVREM_USE_ENVIRON_ROOT
 	};
-	DB_ENV *e;
 	u_int32_t cflag, enc_flag, flag, forceflag, sflag;
 	int i, optindex, result, ret;
 	char *datadir, *home, *logdir, *passwd, *server, *tmpdir;
@@ -988,14 +1191,15 @@ tcl_EnvRemove(interp, objc, objv, dbenv, envip)
 	 * one of the user.  Don't bother with the info stuff.
 	 */
 	if (dbenv == NULL) {
-		if ((ret = db_env_create(&e, cflag)) != 0) {
+		if ((ret = db_env_create(&dbenv, cflag)) != 0) {
 			result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
 			    "db_env_create");
 			goto error;
 		}
 		if (server != NULL) {
 			_debug_check();
-			ret = e->set_rpc_server(e, NULL, server, 0, 0, 0);
+			ret = dbenv->set_rpc_server(
+			    dbenv, NULL, server, 0, 0, 0);
 			result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
 			    "set_rpc_server");
 			if (result != TCL_OK)
@@ -1003,7 +1207,7 @@ tcl_EnvRemove(interp, objc, objv, dbenv, envip)
 		}
 		if (datadir != NULL) {
 			_debug_check();
-			ret = e->set_data_dir(e, datadir);
+			ret = dbenv->set_data_dir(dbenv, datadir);
 			result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
 			    "set_data_dir");
 			if (result != TCL_OK)
@@ -1011,7 +1215,7 @@ tcl_EnvRemove(interp, objc, objv, dbenv, envip)
 		}
 		if (logdir != NULL) {
 			_debug_check();
-			ret = e->set_lg_dir(e, logdir);
+			ret = dbenv->set_lg_dir(dbenv, logdir);
 			result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
 			    "set_log_dir");
 			if (result != TCL_OK)
@@ -1019,24 +1223,27 @@ tcl_EnvRemove(interp, objc, objv, dbenv, envip)
 		}
 		if (tmpdir != NULL) {
 			_debug_check();
-			ret = e->set_tmp_dir(e, tmpdir);
+			ret = dbenv->set_tmp_dir(dbenv, tmpdir);
 			result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
 			    "set_tmp_dir");
 			if (result != TCL_OK)
 				goto error;
 		}
 		if (passwd != NULL) {
-			ret = e->set_encrypt(e, passwd, enc_flag);
+			ret = dbenv->set_encrypt(dbenv, passwd, enc_flag);
 			result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
 			    "set_encrypt");
 		}
-		if (sflag != 0 && (ret = e->set_flags(e, sflag, 1)) != 0) {
+		if (sflag != 0 &&
+		    (ret = dbenv->set_flags(dbenv, sflag, 1)) != 0) {
 			_debug_check();
 			result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
 			    "set_flags");
 			if (result != TCL_OK)
 				goto error;
 		}
+		dbenv->set_errpfx(dbenv, "EnvRemove");
+		dbenv->set_errcall(dbenv, _ErrorFunc);
 	} else {
 		/*
 		 * We have to clean up any info associated with this env,
@@ -1045,7 +1252,6 @@ tcl_EnvRemove(interp, objc, objv, dbenv, envip)
 		 */
 		_EnvInfoDelete(interp, envip);
 		envip = NULL;
-		e = dbenv;
 	}
 
 	flag |= forceflag;
@@ -1054,9 +1260,8 @@ tcl_EnvRemove(interp, objc, objv, dbenv, envip)
 	 * the environment.
 	 */
 	_debug_check();
-	ret = e->remove(e, home, flag);
-	result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
-	    "env remove");
+	ret = dbenv->remove(dbenv, home, flag);
+	result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret), "env remove");
 error:
 	return (result);
 }
@@ -1240,17 +1445,37 @@ tcl_EnvVerbose(interp, dbenv, which, onoff)
 {
 	static const char *verbwhich[] = {
 		"deadlock",
+		"fileops",
+		"fileops_all",
 		"recovery",
 		"register",
 		"rep",
+		"rep_elect",
+		"rep_lease",
+		"rep_misc",
+		"rep_msgs",
+		"rep_sync",
+		"rep_test",
+		"repmgr_connfail",
+		"repmgr_misc",
 		"wait",
 		NULL
 	};
 	enum verbwhich {
 		ENVVERB_DEADLOCK,
+		ENVVERB_FILEOPS,
+		ENVVERB_FILEOPS_ALL,
 		ENVVERB_RECOVERY,
 		ENVVERB_REGISTER,
 		ENVVERB_REPLICATION,
+		ENVVERB_REP_ELECT,
+		ENVVERB_REP_LEASE,
+		ENVVERB_REP_MISC,
+		ENVVERB_REP_MSGS,
+		ENVVERB_REP_SYNC,
+		ENVVERB_REP_TEST,
+		ENVVERB_REPMGR_CONNFAIL,
+		ENVVERB_REPMGR_MISC,
 		ENVVERB_WAITSFOR
 	};
 	static const char *verbonoff[] = {
@@ -1273,6 +1498,12 @@ tcl_EnvVerbose(interp, dbenv, which, onoff)
 	case ENVVERB_DEADLOCK:
 		wh = DB_VERB_DEADLOCK;
 		break;
+	case ENVVERB_FILEOPS:
+		wh = DB_VERB_FILEOPS;
+		break;
+	case ENVVERB_FILEOPS_ALL:
+		wh = DB_VERB_FILEOPS_ALL;
+		break;
 	case ENVVERB_RECOVERY:
 		wh = DB_VERB_RECOVERY;
 		break;
@@ -1282,6 +1513,30 @@ tcl_EnvVerbose(interp, dbenv, which, onoff)
 	case ENVVERB_REPLICATION:
 		wh = DB_VERB_REPLICATION;
 		break;
+	case ENVVERB_REP_ELECT:
+		wh = DB_VERB_REP_ELECT;
+		break;
+	case ENVVERB_REP_LEASE:
+		wh = DB_VERB_REP_LEASE;
+		break;
+	case ENVVERB_REP_MISC:
+		wh = DB_VERB_REP_MISC;
+		break;
+	case ENVVERB_REP_MSGS:
+		wh = DB_VERB_REP_MSGS;
+		break;
+	case ENVVERB_REP_SYNC:
+		wh = DB_VERB_REP_SYNC;
+		break;
+	case ENVVERB_REP_TEST:
+		wh = DB_VERB_REP_TEST;
+		break;
+	case ENVVERB_REPMGR_CONNFAIL:
+		wh = DB_VERB_REPMGR_CONNFAIL;
+		break;
+	case ENVVERB_REPMGR_MISC:
+		wh = DB_VERB_REPMGR_MISC;
+		break;
 	case ENVVERB_WAITSFOR:
 		wh = DB_VERB_WAITSFOR;
 		break;
@@ -1321,9 +1576,11 @@ tcl_EnvAttr(interp, objc, objv, dbenv)
 	Tcl_Obj *CONST objv[];		/* The argument objects */
 	DB_ENV *dbenv;			/* Env pointer */
 {
-	int result;
+	ENV *env;
 	Tcl_Obj *myobj, *retlist;
+	int result;
 
+	env = dbenv->env;
 	result = TCL_OK;
 
 	if (objc > 2) {
@@ -1333,42 +1590,42 @@ tcl_EnvAttr(interp, objc, objv, dbenv)
 	retlist = Tcl_NewListObj(0, NULL);
 	/*
 	 * XXX
-	 * We peek at the dbenv to determine what subsystems
-	 * we have available in this env.
+	 * We peek at the ENV to determine what subsystems we have available
+	 * in this environment.
 	 */
 	myobj = NewStringObj("-home", strlen("-home"));
 	if ((result = Tcl_ListObjAppendElement(interp,
 	    retlist, myobj)) != TCL_OK)
 		goto err;
-	myobj = NewStringObj(dbenv->db_home, strlen(dbenv->db_home));
+	myobj = NewStringObj(env->db_home, strlen(env->db_home));
 	if ((result = Tcl_ListObjAppendElement(interp,
 	    retlist, myobj)) != TCL_OK)
 		goto err;
-	if (CDB_LOCKING(dbenv)) {
+	if (CDB_LOCKING(env)) {
 		myobj = NewStringObj("-cdb", strlen("-cdb"));
 		if ((result = Tcl_ListObjAppendElement(interp,
 		    retlist, myobj)) != TCL_OK)
 			goto err;
 	}
-	if (CRYPTO_ON(dbenv)) {
+	if (CRYPTO_ON(env)) {
 		myobj = NewStringObj("-crypto", strlen("-crypto"));
 		if ((result = Tcl_ListObjAppendElement(interp,
 		    retlist, myobj)) != TCL_OK)
 			goto err;
 	}
-	if (LOCKING_ON(dbenv)) {
+	if (LOCKING_ON(env)) {
 		myobj = NewStringObj("-lock", strlen("-lock"));
 		if ((result = Tcl_ListObjAppendElement(interp,
 		    retlist, myobj)) != TCL_OK)
 			goto err;
 	}
-	if (LOGGING_ON(dbenv)) {
+	if (LOGGING_ON(env)) {
 		myobj = NewStringObj("-log", strlen("-log"));
 		if ((result = Tcl_ListObjAppendElement(interp,
 		    retlist, myobj)) != TCL_OK)
 			goto err;
 	}
-	if (MPOOL_ON(dbenv)) {
+	if (MPOOL_ON(env)) {
 		myobj = NewStringObj("-mpool", strlen("-mpool"));
 		if ((result = Tcl_ListObjAppendElement(interp,
 		    retlist, myobj)) != TCL_OK)
@@ -1380,13 +1637,13 @@ tcl_EnvAttr(interp, objc, objv, dbenv)
 		    retlist, myobj)) != TCL_OK)
 			goto err;
 	}
-	if (REP_ON(dbenv)) {
+	if (REP_ON(env)) {
 		myobj = NewStringObj("-rep", strlen("-rep"));
 		if ((result = Tcl_ListObjAppendElement(interp,
 		    retlist, myobj)) != TCL_OK)
 			goto err;
 	}
-	if (TXN_ON(dbenv)) {
+	if (TXN_ON(env)) {
 		myobj = NewStringObj("-txn", strlen("-txn"));
 		if ((result = Tcl_ListObjAppendElement(interp,
 		    retlist, myobj)) != TCL_OK)
@@ -1397,6 +1654,42 @@ err:
 	return (result);
 }
 
+/*
+ * tcl_EventNotify --
+ *	Call DB_ENV->set_event_notify().
+ *
+ * PUBLIC: int tcl_EventNotify  __P((Tcl_Interp *, DB_ENV *, Tcl_Obj *,
+ * PUBLIC:    DBTCL_INFO *));
+ *
+ *	Note that this normally can/should be achieved as an argument to
+ * berkdb env, but we need to test changing the event function on
+ * the fly.
+ */
+int
+tcl_EventNotify(interp, dbenv, eobj, ip)
+	Tcl_Interp *interp;		/* Interpreter */
+	DB_ENV *dbenv;
+	Tcl_Obj *eobj;		/* The event proc */
+	DBTCL_INFO *ip;
+{
+	int ret;
+
+	/*
+	 * We don't need to crack the event procedure out now.
+	 */
+	/*
+	 * If we're replacing an existing event proc, decrement it now.
+	 */
+	if (ip->i_event != NULL) {
+		Tcl_DecrRefCount(ip->i_event);
+	}
+	ip->i_event = eobj;
+	Tcl_IncrRefCount(ip->i_event);
+	_debug_check();
+	ret = dbenv->set_event_notify(dbenv, _EventFunc);
+	return (_ReturnSetup(interp, ret, DB_RETOK_STD(ret), "env event"));
+}
+
 /*
  * PUBLIC: int tcl_EnvSetFlags __P((Tcl_Interp *, DB_ENV *, Tcl_Obj *,
  * PUBLIC:    Tcl_Obj *));
@@ -1414,10 +1707,7 @@ tcl_EnvSetFlags(interp, dbenv, which, onoff)
 	static const char *sfwhich[] = {
 		"-auto_commit",
 		"-direct_db",
-		"-direct_log",
-		"-dsync_log",
-		"-log_inmemory",
-		"-log_remove",
+		"-multiversion",
 		"-nolock",
 		"-nommap",
 		"-nopanic",
@@ -1430,10 +1720,7 @@ tcl_EnvSetFlags(interp, dbenv, which, onoff)
 	enum sfwhich {
 		ENVSF_AUTOCOMMIT,
 		ENVSF_DIRECTDB,
-		ENVSF_DIRECTLOG,
-		ENVSF_DSYNCLOG,
-		ENVSF_LOG_INMEMORY,
-		ENVSF_LOG_REMOVE,
+		ENVSF_MULTIVERSION,
 		ENVSF_NOLOCK,
 		ENVSF_NOMMAP,
 		ENVSF_NOPANIC,
@@ -1465,17 +1752,8 @@ tcl_EnvSetFlags(interp, dbenv, which, onoff)
 	case ENVSF_DIRECTDB:
 		wh = DB_DIRECT_DB;
 		break;
-	case ENVSF_DIRECTLOG:
-		wh = DB_DIRECT_LOG;
-		break;
-	case ENVSF_DSYNCLOG:
-		wh = DB_DSYNC_LOG;
-		break;
-	case ENVSF_LOG_INMEMORY:
-		wh = DB_LOG_INMEMORY;
-		break;
-	case ENVSF_LOG_REMOVE:
-		wh = DB_LOG_AUTOREMOVE;
+	case ENVSF_MULTIVERSION:
+		wh = DB_MULTIVERSION;
 		break;
 	case ENVSF_NOLOCK:
 		wh = DB_NOLOCKING;
@@ -1516,13 +1794,24 @@ tcl_EnvSetFlags(interp, dbenv, which, onoff)
 	}
 	ret = dbenv->set_flags(dbenv, wh, on);
 	return (_ReturnSetup(interp, ret, DB_RETOK_STD(ret),
-	    "env set verbose"));
+	    "env set flags"));
 }
 
 /*
- * PUBLIC: int tcl_EnvTest __P((Tcl_Interp *, int, Tcl_Obj * CONST*, DB_ENV *));
- *
  * tcl_EnvTest --
+ *	The "$env test ..." command is a sort of catch-all for any sort of
+ * desired test hook manipulation.  The "abort", "check" and "copy" subcommands
+ * all set one or another certain location in the DB_ENV handle to a specific
+ * value.  (In the case of "check", the value is an integer passed in with the
+ * command itself.  For the other two, the "value" is a predefined enum
+ * constant, specified by name.)
+ *	The "$env test force ..." subcommand invokes other, more arbitrary
+ * manipulations.
+ *	Although these functions may not all seem closely related, putting them
+ * all under the name "test" has the aesthetic appeal of keeping the rest of the
+ * API clean.
+ *
+ * PUBLIC: int tcl_EnvTest __P((Tcl_Interp *, int, Tcl_Obj * CONST*, DB_ENV *));
  */
 int
 tcl_EnvTest(interp, objc, objv, dbenv)
@@ -1535,12 +1824,14 @@ tcl_EnvTest(interp, objc, objv, dbenv)
 		"abort",
 		"check",
 		"copy",
+		"force",
 		NULL
 	};
 	enum envtestcmd {
 		ENVTEST_ABORT,
 		ENVTEST_CHECK,
-		ENVTEST_COPY
+		ENVTEST_COPY,
+		ENVTEST_FORCE
 	};
 	static const char *envtestat[] = {
 		"electinit",
@@ -1569,13 +1860,23 @@ tcl_EnvTest(interp, objc, objv, dbenv)
 		ENVTEST_POSTSYNC,
 		ENVTEST_SUBDB_LOCKS
 	};
+	static const char *envtestforce[] = {
+		"noarchive_timeout",
+		NULL
+	};
+	enum envtestforce {
+		ENVTEST_NOARCHIVE_TIMEOUT
+	};
+	ENV *env;
 	int *loc, optindex, result, testval;
 
+	env = dbenv->env;
 	result = TCL_OK;
 	loc = NULL;
 
 	if (objc != 4) {
-		Tcl_WrongNumArgs(interp, 2, objv, "abort|copy location");
+		Tcl_WrongNumArgs(interp,
+		    2, objv, "abort|check|copy|force <args>");
 		return (TCL_ERROR);
 	}
 
@@ -1589,18 +1890,31 @@ tcl_EnvTest(interp, objc, objv, dbenv)
 	}
 	switch ((enum envtestcmd)optindex) {
 	case ENVTEST_ABORT:
-		loc = &dbenv->test_abort;
+		loc = &env->test_abort;
 		break;
 	case ENVTEST_CHECK:
-		loc = &dbenv->test_check;
+		loc = &env->test_check;
 		if (Tcl_GetIntFromObj(interp, objv[3], &testval) != TCL_OK) {
 			result = IS_HELP(objv[3]);
 			return (result);
 		}
 		goto done;
 	case ENVTEST_COPY:
-		loc = &dbenv->test_copy;
+		loc = &env->test_copy;
 		break;
+	case ENVTEST_FORCE:
+		if (Tcl_GetIndexFromObj(interp, objv[3], envtestforce, "arg",
+			TCL_EXACT, &optindex) != TCL_OK) {
+			result = IS_HELP(objv[3]);
+			return (result);
+		}
+		/*
+		 * In the future we might add more, and then we'd use a switch
+		 * statement.
+		 */
+		DB_ASSERT(env,
+		    (enum envtestforce)optindex == ENVTEST_NOARCHIVE_TIMEOUT);
+		return (tcl_RepNoarchiveTimeout(interp, dbenv));
 	default:
 		Tcl_SetResult(interp, "Illegal store location", TCL_STATIC);
 		return (TCL_ERROR);
@@ -1616,11 +1930,11 @@ tcl_EnvTest(interp, objc, objv, dbenv)
 	}
 	switch ((enum envtestat)optindex) {
 	case ENVTEST_ELECTINIT:
-		DB_ASSERT(loc == &dbenv->test_abort);
+		DB_ASSERT(env, loc == &env->test_abort);
 		testval = DB_TEST_ELECTINIT;
 		break;
 	case ENVTEST_ELECTVOTE1:
-		DB_ASSERT(loc == &dbenv->test_abort);
+		DB_ASSERT(env, loc == &env->test_abort);
 		testval = DB_TEST_ELECTVOTE1;
 		break;
 	case ENVTEST_NONE:
@@ -1648,7 +1962,7 @@ tcl_EnvTest(interp, objc, objv, dbenv)
 		testval = DB_TEST_POSTSYNC;
 		break;
 	case ENVTEST_SUBDB_LOCKS:
-		DB_ASSERT(loc == &dbenv->test_abort);
+		DB_ASSERT(env, loc == &env->test_abort);
 		testval = DB_TEST_SUBDB_LOCKS;
 		break;
 	default:
@@ -1675,12 +1989,14 @@ env_DbRemove(interp, objc, objv, dbenv)
 {
 	static const char *envdbrem[] = {
 		"-auto_commit",
+		"-notdurable",
 		"-txn",
 		"--",
 		NULL
 	};
 	enum envdbrem {
 		TCL_EDBREM_COMMIT,
+		TCL_EDBREM_NOTDURABLE,
 		TCL_EDBREM_TXN,
 		TCL_EDBREM_ENDARG
 	};
@@ -1741,6 +2057,9 @@ env_DbRemove(interp, objc, objv, dbenv)
 		case TCL_EDBREM_ENDARG:
 			endarg = 1;
 			break;
+		case TCL_EDBREM_NOTDURABLE:
+			flag |= DB_TXN_NOT_DURABLE;
+			break;
 		}
 		/*
 		 * If, at any time, parsing the args we get an error,
@@ -1770,7 +2089,7 @@ env_DbRemove(interp, objc, objv, dbenv)
 			subdbtmp =
 			    Tcl_GetByteArrayFromObj(objv[i++], &subdblen);
 			if ((ret = __os_malloc(
-			    dbenv, (size_t)subdblen + 1, &subdb)) != 0) {
+			    dbenv->env, (size_t)subdblen + 1, &subdb)) != 0) {
 				Tcl_SetResult(interp,
 				    db_strerror(ret), TCL_STATIC);
 				return (0);
@@ -1788,7 +2107,7 @@ env_DbRemove(interp, objc, objv, dbenv)
 	    "env dbremove");
 error:
 	if (subdb)
-		__os_free(dbenv, subdb);
+		__os_free(dbenv->env, subdb);
 	return (result);
 }
 
@@ -1901,7 +2220,7 @@ env_DbRename(interp, objc, objv, dbenv)
 			subdbtmp =
 			    Tcl_GetByteArrayFromObj(objv[i++], &subdblen);
 			if ((ret = __os_malloc(
-			    dbenv, (size_t)subdblen + 1, &subdb)) != 0) {
+			    dbenv->env, (size_t)subdblen + 1, &subdb)) != 0) {
 				Tcl_SetResult(interp,
 				    db_strerror(ret), TCL_STATIC);
 				return (0);
@@ -1911,7 +2230,7 @@ env_DbRename(interp, objc, objv, dbenv)
 		}
 		subdbtmp = Tcl_GetByteArrayFromObj(objv[i++], &newlen);
 		if ((ret = __os_malloc(
-		    dbenv, (size_t)newlen + 1, &newname)) != 0) {
+		    dbenv->env, (size_t)newlen + 1, &newname)) != 0) {
 			Tcl_SetResult(interp,
 			    db_strerror(ret), TCL_STATIC);
 			return (0);
@@ -1929,9 +2248,9 @@ env_DbRename(interp, objc, objv, dbenv)
 	    "env dbrename");
 error:
 	if (subdb)
-		__os_free(dbenv, subdb);
+		__os_free(dbenv->env, subdb);
 	if (newname)
-		__os_free(dbenv, newname);
+		__os_free(dbenv->env, newname);
 	return (result);
 }
 
@@ -1958,10 +2277,7 @@ env_GetFlags(interp, objc, objv, dbenv)
 		{ DB_AUTO_COMMIT, "-auto_commit" },
 		{ DB_CDB_ALLDB, "-cdb_alldb" },
 		{ DB_DIRECT_DB, "-direct_db" },
-		{ DB_DIRECT_LOG, "-direct_log" },
-		{ DB_DSYNC_LOG, "-dsync_log" },
-		{ DB_LOG_AUTOREMOVE, "-log_remove" },
-		{ DB_LOG_INMEMORY, "-log_inmemory" },
+		{ DB_MULTIVERSION, "-multiversion" },
 		{ DB_NOLOCKING, "-nolock" },
 		{ DB_NOMMAP, "-nommap" },
 		{ DB_NOPANIC, "-nopanic" },
@@ -2020,16 +2336,19 @@ env_GetOpenFlag(interp, objc, objv, dbenv)
 		char *arg;
 	} open_flags[] = {
 		{ DB_CREATE, "-create" },
+		{ DB_FAILCHK, "-failchk" },
 		{ DB_INIT_CDB, "-cdb" },
 		{ DB_INIT_LOCK, "-lock" },
 		{ DB_INIT_LOG, "-log" },
 		{ DB_INIT_MPOOL, "-mpool" },
+		{ DB_INIT_REP, "-rep" },
 		{ DB_INIT_TXN, "-txn" },
 		{ DB_LOCKDOWN, "-lockdown" },
 		{ DB_PRIVATE, "-private" },
 		{ DB_RECOVER, "-recover" },
 		{ DB_RECOVER_FATAL, "-recover_fatal" },
 		{ DB_REGISTER, "-register" },
+		{ DB_FAILCHK, "-failchk" },
 		{ DB_SYSTEM_MEM, "-system_mem" },
 		{ DB_THREAD, "-thread" },
 		{ DB_USE_ENVIRON, "-use_environ" },
@@ -2179,8 +2498,9 @@ env_GetTimeout(interp, objc, objv, dbenv)
 		u_int32_t flag;
 		char *arg;
 	} timeout_flags[] = {
-		{ DB_SET_TXN_TIMEOUT, "txn" },
 		{ DB_SET_LOCK_TIMEOUT, "lock" },
+		{ DB_SET_REG_TIMEOUT, "reg" },
+		{ DB_SET_TXN_TIMEOUT, "txn" },
 		{ 0, NULL }
 	};
 	Tcl_Obj *res;
@@ -2232,9 +2552,19 @@ env_GetVerbose(interp, objc, objv, dbenv)
 		char *arg;
 	} verbose_flags[] = {
 		{ DB_VERB_DEADLOCK, "deadlock" },
+		{ DB_VERB_FILEOPS, "fileops" },
+		{ DB_VERB_FILEOPS_ALL, "fileops_all" },
 		{ DB_VERB_RECOVERY, "recovery" },
 		{ DB_VERB_REGISTER, "register" },
 		{ DB_VERB_REPLICATION, "rep" },
+		{ DB_VERB_REP_ELECT, "rep_elect" },
+		{ DB_VERB_REP_LEASE, "rep_lease" },
+		{ DB_VERB_REP_MISC, "rep_misc" },
+		{ DB_VERB_REP_MSGS, "rep_msgs" },
+		{ DB_VERB_REP_SYNC, "rep_sync" },
+		{ DB_VERB_REP_TEST, "rep_test" },
+		{ DB_VERB_REPMGR_CONNFAIL, "repmgr_connfail" },
+		{ DB_VERB_REPMGR_MISC, "repmgr_misc" },
 		{ DB_VERB_WAITSFOR, "wait" },
 		{ 0, NULL }
 	};
@@ -2328,8 +2658,8 @@ tcl_EnvSetErrpfx(interp, dbenv, ip, pfx)
 	 * If the user already set one, free it.
 	 */
 	if (ip->i_errpfx != NULL)
-		__os_free(dbenv, ip->i_errpfx);
-	if ((ret = __os_strdup(dbenv, pfx, &ip->i_errpfx)) != 0) {
+		__os_free(dbenv->env, ip->i_errpfx);
+	if ((ret = __os_strdup(dbenv->env, pfx, &ip->i_errpfx)) != 0) {
 		result = _ReturnSetup(interp, ret,
 		    DB_RETOK_STD(ret), "__os_strdup");
 		ip->i_errpfx = NULL;
