@@ -1,9 +1,8 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 2001-2005
-#	Sleepycat Software.  All rights reserved.
+# Copyright (c) 2001-2009 Oracle.  All rights reserved.
 #
-# $Id: bigfile002.tcl,v 12.2 2005/06/16 20:23:51 bostic Exp $
+# $Id$
 #
 # TEST	bigfile002
 # TEST	This one should be faster and not require so much disk space,
@@ -11,7 +10,11 @@
 # TEST	with 1K pages.  Dirty page 6000000.  Sync.
 proc bigfile002 { args } {
 	source ./include.tcl
-
+	global is_fat32
+	if { $is_fat32 } { 
+		puts "Skipping bigfile002 for FAT32 file system."
+		return
+	}
 	puts "Bigfile002: Creating large, sparse file through mpool..."
 	flush stdout
 
@@ -31,7 +34,7 @@ proc bigfile002 { args } {
 	error_check_good pg_set [$pg is_setto A] 1
 
 	# Put page back.
-	error_check_good pg_put [$pg put -dirty] 0
+	error_check_good pg_put [$pg put] 0
 
 	# Fsync.
 	error_check_good fsync [$file fsync] 0
