@@ -1,9 +1,8 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 1999-2003
-#       Sleepycat Software.  All rights reserved.
+# Copyright (c) 1999-2009 Oracle.  All rights reserved.
 #
-# $Id: env011.tcl,v 1.3 2003/01/08 05:49:57 bostic Exp $
+# $Id$
 #
 # TEST	env011
 # TEST	Run with region overwrite flag.
@@ -19,9 +18,9 @@ proc env011 { } {
 	set db [eval \
 	    {berkdb_open -auto_commit -env $e -btree -create -mode 0644} ]
 	error_check_good dbopen [is_valid_db $db] TRUE
-	set ret [eval {$db put} -auto_commit "aaa" "data"]
+	set ret [eval {$db put} "aaa" "data"]
 	error_check_good put $ret 0
-	set ret [eval {$db put} -auto_commit "bbb" "data"]
+	set ret [eval {$db put} "bbb" "data"]
 	error_check_good put $ret 0
 	error_check_good db_close [$db close] 0
 	error_check_good envclose [$e close] 0
