@@ -1,22 +1,25 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 1996-2005
-#	Sleepycat Software.  All rights reserved.
+# Copyright (c) 1996-2009 Oracle.  All rights reserved.
 #
-# $Id: memp001.tcl,v 12.1 2005/06/16 20:23:56 bostic Exp $
+# $Id$
 #
 
 # TEST	memp001
 # TEST	Randomly updates pages.
 proc memp001 { } {
-
+	source ./include.tcl
 	memp001_body 1 ""
 	memp001_body 3 ""
 	memp001_body 1 -private
 	memp001_body 3 -private
+	if { $is_qnx_test } {
+		puts "Skipping remainder of memp001 for\
+		    environments in system memory on QNX"
+		return
+	}
 	memp001_body 1 "-system_mem -shm_key 1"
 	memp001_body 3 "-system_mem -shm_key 1"
-
 }
 
 proc memp001_body { ncache flags } {
@@ -94,7 +97,7 @@ proc memp001_body { ncache flags } {
 		set p(10) [get_range $mpool 40]
 		set p(7) [replace $mpool $p(7)]
 		set p(8) [replace $mpool $p(8)]
-		set p(9) [replace $mpool $p(9)]
+		set p(9) [replace $mpool $p(9) ]
 		set p(10) [replace $mpool $p(10)]
 		#
 		# We now need to put all the pages we have here or
@@ -139,7 +142,7 @@ proc file_create { fname nblocks blocksize } {
 
 proc get_range { mpool max } {
 	set pno [berkdb random_int 0 $max]
-	set p [$mpool get $pno]
+	set p [eval $mpool get $pno]
 	error_check_good page [is_valid_page $p $mpool] TRUE
 	set got [$p pgnum]
 	if { $got != $pno } {
@@ -151,16 +154,16 @@ proc get_range { mpool max } {
 	return $p
 }
 
-proc replace { mpool p } {
+proc replace { mpool p { args "" } } {
 	set pgno [$p pgnum]
 
 	set ret [$p init "Page is unpinned by [pid]"]
 	error_check_good page_init $ret 0
 
-	set ret [$p put -dirty]
+	set ret [$p put]
 	error_check_good page_put $ret 0
 
-	set p2 [$mpool get $pgno]
+	set p2 [eval $mpool get $args $pgno]
 	error_check_good page [is_valid_page $p2 $mpool] TRUE
 
 	return $p2
