@@ -1,5 +1,9 @@
+# See the file LICENSE for redistribution information.
+#
+# Copyright (c) 1996-2009 Oracle.  All rights reserved.
+#
 # Code to load up the tests in to the Queue database
-# $Id: parallel.tcl,v 12.3 2005/10/27 14:09:28 carol Exp $
+# $Id$
 proc load_queue { file  {dbdir RUNQUEUE} nitems } {
 	global serial_tests
 	global num_serial
@@ -148,7 +152,7 @@ proc run_parallel { nprocs {list run_all} {nitems ALL} } {
 			close $f
 		} res]
 	}
-	watch_procs $pidlist 300 600000
+	watch_procs $pidlist 300 1000000
 
 	set failed 0
 	for { set i 0 } { $i <= $nprocs } { incr i } {
@@ -288,9 +292,18 @@ proc mkparalleldirs { nprocs basename queuedir } {
 		catch {file mkdir $destdir}
 		puts "Created $destdir"
 		if { $is_windows_test == 1 } {
-			catch {file mkdir $destdir/Debug}
+			catch {file mkdir $destdir/$buildpath}
 			catch {eval file copy \
-			    [eval glob {$dir/Debug/*.dll}] $destdir/Debug}
+			    [eval glob {$dir/$buildpath/*.dll}] $destdir/$buildpath}
+			catch {eval file copy \
+			    [eval glob {$dir/$buildpath/db_{checkpoint,deadlock}$EXE} \
+			    {$dir/$buildpath/db_{dump,load,printlog,recover,stat,upgrade}$EXE} \
+			    {$dir/$buildpath/db_{archive,verify,hotbackup}$EXE}] \
+			    {$dir/$buildpath/dbkill$EXE} \
+			    $destdir/$buildpath}
+			catch {eval file copy \
+			    [eval glob -nocomplain {$dir/$buildpath/db_{reptest,repsite}$EXE}] \
+			    $destdir/$buildpath}
 		}
 		catch {eval file copy \
 		    [eval glob {$dir/{.libs,include.tcl}}] $destdir}
@@ -298,8 +311,10 @@ proc mkparalleldirs { nprocs basename queuedir } {
 		catch {eval file copy \
 		    [eval glob {$dir/db_{checkpoint,deadlock}$EXE} \
 		    {$dir/db_{dump,load,printlog,recover,stat,upgrade}$EXE} \
-		    {$dir/db_{archive,verify}$EXE}] \
+		    {$dir/db_{archive,verify,hotbackup}$EXE}] \
 		    $destdir}
+		catch {eval file copy \
+		    [eval glob -nocomplain {$dir/db_{reptest,repsite}$EXE}] $destdir}
 
 		# Create modified copies of include.tcl in parallel
 		# directories so paths still work.
@@ -312,7 +327,6 @@ proc mkparalleldirs { nprocs basename queuedir } {
 		regsub {src_root } $d {src_root ../} d
 		set tdir "TESTDIR.$i"
 		regsub -all {TESTDIR} $d $tdir d
-		regsub {KILL \.} $d {KILL ..} d
 		set outfile [open $destdir/include.tcl w]
 		puts $outfile $d
 		close $outfile
@@ -328,6 +342,7 @@ proc mkparalleldirs { nprocs basename queuedir } {
 
 proc run_ptest { nprocs test args } {
 	global parms
+	global valid_methods
 	set basename ./PARALLEL_TESTDIR
 	set queuedir NULL
 	source ./include.tcl
@@ -335,8 +350,7 @@ proc run_ptest { nprocs test args } {
 	mkparalleldirs $nprocs $basename $queuedir
 
 	if { [info exists parms($test)] } {
-		foreach method \
-		    "hash queue queueext recno rbtree frecno rrecno btree" {
+		foreach method $valid_methods {
 			if { [eval exec_ptest $nprocs $basename \
 			    $test $method $args] != 0 } {
 				break
