@@ -1,30 +1,37 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 1996-2005
-#	Sleepycat Software.  All rights reserved.
+# Copyright (c) 1996-2009 Oracle.  All rights reserved.
 #
-# $Id: recd005.tcl,v 12.1 2005/06/16 20:23:58 bostic Exp $
+# $Id$
 #
 # TEST	recd005
 # TEST	Verify reuse of file ids works on catastrophic recovery.
 # TEST
 # TEST	Make sure that we can do catastrophic recovery even if we open
 # TEST	files using the same log file id.
-proc recd005 { method args} {
+proc recd005 { method args } {
 	source ./include.tcl
 	global rand_init
 
+	set envargs ""
+	set zero_idx [lsearch -exact $args "-zero_log"]
+	if { $zero_idx != -1 } {
+		set args [lreplace $args $zero_idx $zero_idx]
+		set envargs "-zero_log"
+	}
+
 	set args [convert_args $method $args]
 	set omethod [convert_method $method]
 
-	puts "Recd005: $method catastrophic recovery"
+	puts "Recd005: $method catastrophic recovery ($envargs)"
 
 	berkdb srand $rand_init
 
 	set testfile1 recd005.1.db
 	set testfile2 recd005.2.db
-	set eflags \
-	    "-create -txn -lock_max 2000 -lock_max_objects 2000 -home $testdir"
+	set max_locks 2000
+	set eflags "-create -txn -lock_max_locks $max_locks \
+	    -lock_max_objects $max_locks -home $testdir $envargs"
 
 	set tnum 0
 	foreach sizes "{1000 10} {10 1000}" {
@@ -60,14 +67,16 @@ proc recd005 { method args} {
 
 			set dbenv [eval $env_cmd]
 			puts "\tRecd005.$tnum.b: Populating databases"
-			do_one_file \
-			    $testdir $method $dbenv $env_cmd $testfile1 $s1 $op1
-			do_one_file \
-			    $testdir $method $dbenv $env_cmd $testfile2 $s2 $op2
+			eval {do_one_file  $testdir \
+			    $method $dbenv $env_cmd $testfile1 $s1 $op1 } $args
+			eval {do_one_file  $testdir \
+			    $method $dbenv $env_cmd $testfile2 $s2 $op2 } $args
 
 			puts "\tRecd005.$tnum.c: Verifying initial population"
-			check_file $testdir $env_cmd $testfile1 $op1
-			check_file $testdir $env_cmd $testfile2 $op2
+			eval {check_file \
+			    $testdir $env_cmd $testfile1 $op1 } $args
+			eval {check_file \
+			    $testdir $env_cmd $testfile2 $op2 } $args
 
 			# Now, close the environment (so that recovery will work
 			# on NT which won't allow delete of an open file).
@@ -123,8 +132,10 @@ proc recd005 { method args} {
 			puts "complete"
 
 			set dbenv [eval $env_cmd]
-			check_file $testdir $env_cmd $testfile1 $op1
-			check_file $testdir $env_cmd $testfile2 $op2
+			eval {check_file \
+			    $testdir $env_cmd $testfile1 $op1 } $args
+			eval {check_file \
+			    $testdir $env_cmd $testfile2 $op2 } $args
 			reset_env $dbenv
 
 			puts "\tRecd005.$tnum.f:\
@@ -139,7 +150,7 @@ proc recd005 { method args} {
 	}
 }
 
-proc do_one_file { dir method env env_cmd filename num op } {
+proc do_one_file { dir method env env_cmd filename num op args} {
 	source ./include.tcl
 
 	set init_file $dir/$filename.t1
@@ -150,11 +161,11 @@ proc do_one_file { dir method env env_cmd filename num op } {
 	file copy -force $dir/$filename $dir/$filename.init
 	copy_extent_file $dir $filename init
 	set oflags "-auto_commit -unknown -env $env"
-	set db [eval {berkdb_open} $oflags $filename]
+	set db [eval {berkdb_open} $oflags $args $filename]
 
 	# Dump out file contents for initial case
-	open_and_dump_file $filename $env $init_file nop \
-	    dump_file_direction "-first" "-next"
+	eval open_and_dump_file $filename $env $init_file nop \
+	    dump_file_direction "-first" "-next" $args
 
 	set txn [$env txn]
 	error_check_bad txn_begin $txn NULL
@@ -168,8 +179,8 @@ proc do_one_file { dir method env env_cmd filename num op } {
 	error_check_good sync:$db [$db sync] 0
 	file copy -force $dir/$filename $dir/$filename.afterop
 	copy_extent_file $dir $filename afterop
-	open_and_dump_file $testdir/$filename.afterop NULL \
-	    $afterop_file nop dump_file_direction "-first" "-next"
+	eval open_and_dump_file $testdir/$filename.afterop NULL \
+	    $afterop_file nop dump_file_direction "-first" "-next" $args
 	error_check_good txn_$op:$txn [$txn $op] 0
 
 	if { $op == "commit" } {
@@ -180,8 +191,8 @@ proc do_one_file { dir method env env_cmd filename num op } {
 
 	# Dump out file and save a copy.
 	error_check_good sync:$db [$db sync] 0
-	open_and_dump_file $testdir/$filename NULL $final_file nop \
-	    dump_file_direction "-first" "-next"
+	eval open_and_dump_file $testdir/$filename NULL $final_file nop \
+	    dump_file_direction "-first" "-next" $args
 	file copy -force $dir/$filename $dir/$filename.final
 	copy_extent_file $dir $filename final
 
@@ -205,15 +216,15 @@ proc do_one_file { dir method env env_cmd filename num op } {
 	error_check_good close:$db [$db close] 0
 }
 
-proc check_file { dir env_cmd filename op } {
+proc check_file { dir env_cmd filename op args} {
 	source ./include.tcl
 
 	set init_file $dir/$filename.t1
 	set afterop_file $dir/$filename.t2
 	set final_file $dir/$filename.t3
 
-	open_and_dump_file $testdir/$filename NULL $final_file nop \
-	    dump_file_direction "-first" "-next"
+	eval open_and_dump_file $testdir/$filename NULL $final_file nop \
+	    dump_file_direction "-first" "-next" $args
 	if { $op == "abort" } {
 		filesort $init_file $init_file.sort
 		filesort $final_file $final_file.sort
