@@ -1,9 +1,8 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 2001-2005
-#	Sleepycat Software.  All rights reserved.
+# Copyright (c) 2001-2009 Oracle.  All rights reserved.
 #
-# $Id: rep013.tcl,v 12.5 2005/10/18 19:04:17 carol Exp $
+# $Id$
 #
 # TEST	rep013
 # TEST	Replication and swapping master/clients with open dbs.
@@ -17,17 +16,41 @@
 proc rep013 { method { niter 10 } { tnum "013" } args } {
 
 	source ./include.tcl
-	if { $is_windows9x_test == 1 } { 
+	global databases_in_memory 
+	global repfiles_in_memory
+
+	if { $is_windows9x_test == 1 } {
 		puts "Skipping replication test on Win 9x platform."
 		return
-	} 
+	}
+
+	# Run for all access methods.
+	if { $checking_valid_methods } {
+		return "ALL"
+	}
+
 	set args [convert_args $method $args]
 	set logsets [create_logsets 3]
 
+	# Set up named in-memory database testing. 
+	set msg "using on-disk databases"
+	if { $databases_in_memory } {
+		set msg "using named in-memory databases"
+		if { [is_queueext $method] } { 
+			puts -nonewline "Skipping rep$tnum for method "
+			puts "$method with named in-memory databases"
+			return
+		}
+	}
+
+	set msg2 "and on-disk replication files"
+	if { $repfiles_in_memory } {
+		set msg2 "and in-memory replication files"
+	}
+
 	# Run the body of the test with and without recovery.
-	set recopts { "" "-recover" }
 	set anyopts { "" "anywhere" }
-	foreach r $recopts {
+	foreach r $test_recopts {
 		foreach l $logsets {
 			foreach a $anyopts {
 				set logindex [lsearch -exact $l "in-memory"]
@@ -37,7 +60,7 @@ proc rep013 { method { niter 10 } { tnum "013" } args } {
 					continue
 				}
 				puts "Rep$tnum ($r $a): Replication and \
-				    ($method) master/client swapping."
+				    ($method) master/client swapping $msg $msg2."
 				puts "Rep$tnum: Master logs are [lindex $l 0]"
 				puts "Rep$tnum: Client 0 logs are [lindex $l 1]"
 				puts "Rep$tnum: Client 1 logs are [lindex $l 2]"
@@ -50,6 +73,20 @@ proc rep013 { method { niter 10 } { tnum "013" } args } {
 proc rep013_sub { method niter tnum logset recargs anyopt largs } {
 	global testdir
 	global anywhere
+	global databases_in_memory
+	global repfiles_in_memory
+	global rep_verbose
+	global verbose_type
+
+	set verbargs ""
+	if { $rep_verbose == 1 } {
+		set verbargs " -verbose {$verbose_type on} "
+	}
+
+	set repmemargs ""
+	if { $repfiles_in_memory } {
+		set repmemargs "-rep_inmem_files "
+	}
 
 	env_cleanup $testdir
 	set orig_tdir $testdir
@@ -87,45 +124,32 @@ proc rep013_sub { method niter tnum logset recargs anyopt largs } {
 	# Open a master.
 	repladd 1
 	set ma_envcmd "berkdb_env_noerr -create $m_txnargs \
-	    $m_logargs -lock_max 2500 -errpfx ENV1 \
+	    $m_logargs -errpfx ENV1 $verbargs $repmemargs \
 	    -cachesize {0 4194304 3} \
 	    -home $masterdir -rep_transport \[list 1 replsend\]"
-#	set ma_envcmd "berkdb_env_noerr -create $m_txnargs \
-#	    $m_logargs -lock_max 2500 \
-#	    -cachesize {0 4194304 3} \
-#	    -errpfx ENV1 -verbose {rep on} -errfile /dev/stderr \
-#	    -home $masterdir -rep_transport \[list 1 replsend\]"
 	set env1 [eval $ma_envcmd $recargs -rep_master]
-	error_check_good master_env [is_valid_env $env1] TRUE
 
 	# Open two clients
 	repladd 2
 	set cl_envcmd "berkdb_env_noerr -create $c_txnargs \
-	    $c_logargs -lock_max 2500 -errpfx ENV2 \
+	    $c_logargs -errpfx ENV2 $verbargs $repmemargs \
 	    -cachesize {0 2097152 2} \
 	    -home $clientdir -rep_transport \[list 2 replsend\]"
-#	set cl_envcmd "berkdb_env_noerr -create $c_txnargs \
-#	    $c_logargs -lock_max 2500 \
-#	    -cachesize {0 2097152 2} \
-#	    -errpfx ENV2 -verbose {rep on} -errfile /dev/stderr \
-#	    -home $clientdir -rep_transport \[list 2 replsend\]"
 	set env2 [eval $cl_envcmd $recargs -rep_client]
-	error_check_good client_env [is_valid_env $env2] TRUE
 
 	repladd 3
 	set cl2_envcmd "berkdb_env_noerr -create $c2_txnargs \
-	    $c2_logargs -lock_max 2500 -errpfx ENV3 \
+	    $c2_logargs -errpfx ENV3 $verbargs $repmemargs \
 	    -cachesize {0 1048576 1} \
 	    -home $clientdir2 -rep_transport \[list 3 replsend\]"
-#	set cl2_envcmd "berkdb_env_noerr -create $c2_txnargs \
-#	    $c2_logargs -lock_max 2500 \
-#	    -cachesize {0 1048576 1} \
-#	    -errpfx ENV3 -verbose {rep on} -errfile /dev/stderr \
-#	    -home $clientdir2 -rep_transport \[list 3 replsend\]"
 	set cl2env [eval $cl2_envcmd $recargs -rep_client]
-	error_check_good client2_env [is_valid_env $cl2env] TRUE
 
-	set testfile "test$tnum.db"
+	# Set database name for in-memory or on-disk.
+	if { $databases_in_memory } {
+		set testfile { "" "test.db" }
+	} else { 
+		set testfile "test.db"
+	} 
 
 	set omethod [convert_method $method]
 
@@ -172,7 +196,7 @@ proc rep013_sub { method niter tnum logset recargs anyopt largs } {
 
 	# Run a modified test001 in the master (and update clients).
 	puts "\tRep$tnum.a: Running test001 in replicated env."
-	eval rep_test $method $masterenv $masterdb $niter 0 0 0 0 $largs
+	eval rep_test $method $masterenv $masterdb $niter 0 0 0 $largs
 	set envlist "{$env1 1} {$env2 2} {$cl2env 3}"
 	process_msgs $envlist
 
@@ -212,7 +236,7 @@ proc rep013_sub { method niter tnum logset recargs anyopt largs } {
 		set nstart [expr $nstart + $niter]
 		puts "\tRep$tnum.c.$i: Run test in master and client2 only"
 		eval rep_test \
-		    $method $masterenv $masterdb $niter $nstart $nstart 0 0 $largs
+		    $method $masterenv $masterdb $niter $nstart $nstart 0 $largs
 		set envlist "{$masterenv $mid} {$cl2env 3}"
 		process_msgs $envlist
 
@@ -244,20 +268,23 @@ proc rep013_sub { method niter tnum logset recargs anyopt largs } {
 	}
 	puts "\tRep$tnum.e: Check message handling of client."
 	set req3 [stat_field $cl2env rep_stat "Client service requests"]
-	set miss3 [stat_field $cl2env rep_stat "Client service req misses"]
 	set rereq1 [stat_field $env1 rep_stat "Client rerequests"]
 	set rereq2 [stat_field $env2 rep_stat "Client rerequests"]
 	if { $anyopt == "anywhere" } {
 		error_check_bad req $req3 0
-		error_check_bad miss $miss3 0
 		error_check_bad rereq1 $rereq1 0
 		error_check_bad rereq2 $rereq2 0
 	} else {
 		error_check_good req $req3 0
-		error_check_good miss $miss3 0
 		error_check_good rereq1 $rereq1 0
 		error_check_good rereq2 $rereq2 0
 	}
+
+	# Check that databases are in-memory or on-disk as expected.
+	check_db_location $env1 
+	check_db_location $env2 
+	check_db_location $cl2env
+	
 	puts "\tRep$tnum.f: Closing"
 	error_check_good masterdb [$masterdb close] 0
 	error_check_good clientdb [$clientdb close] 0
