@@ -1,3 +1,9 @@
+/*-
+ * See the file LICENSE for redistribution information.
+ *
+ * Copyright (c) 2002-2009 Oracle.  All rights reserved.
+ */
+
 #include <sys/types.h>
 
 #include <ctype.h>
@@ -14,7 +20,7 @@ chk(f)
 
 	if (freopen(f, "r", stdin) == NULL) {
 		fprintf(stderr, "%s: %s\n", f, strerror(errno));
-		exit (1);
+		exit(EXIT_FAILURE);
 	}
 	for (l = 1, r = 0; (ch = getchar()) != EOF;) {
 		if (ch != ',')
