@@ -1,17 +1,19 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 2002-2005
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 2002-2009 Oracle.  All rights reserved.
  *
- * $Id: TransactionTest.java,v 12.3 2005/02/11 23:17:00 mark Exp $
+ * $Id$
  */
 
 package com.sleepycat.collections.test;
 
+import java.io.File;
+import java.io.FileNotFoundException;
 import java.util.Iterator;
 import java.util.List;
 import java.util.SortedSet;
+import java.util.concurrent.atomic.AtomicInteger;
 
 import junit.framework.Test;
 import junit.framework.TestCase;
@@ -26,14 +28,21 @@ import com.sleepycat.collections.StoredSortedMap;
 import com.sleepycat.collections.TransactionRunner;
 import com.sleepycat.collections.TransactionWorker;
 import com.sleepycat.compat.DbCompat;
+import com.sleepycat.db.Cursor;
 import com.sleepycat.db.CursorConfig;
 import com.sleepycat.db.Database;
 import com.sleepycat.db.DatabaseConfig;
 import com.sleepycat.db.DatabaseEntry;
+import com.sleepycat.db.DatabaseException;
 import com.sleepycat.db.Environment;
+import com.sleepycat.db.EnvironmentConfig;
+import com.sleepycat.db.DeadlockException;
+import com.sleepycat.db.OperationStatus;
 import com.sleepycat.db.Transaction;
 import com.sleepycat.db.TransactionConfig;
 import com.sleepycat.util.RuntimeExceptionWrapper;
+import com.sleepycat.util.test.SharedTestUtils;
+import com.sleepycat.util.test.TestEnv;
 
 /**
  * @author Mark Hayes
@@ -48,9 +57,7 @@ public class TransactionTest extends TestCase {
      * Runs a command line collection test.
      * @see #usage
      */
-    public static void main(String[] args)
-        throws Exception {
-
+    public static void main(String[] args) {
         if (args.length == 1 &&
             (args[0].equals("-h") || args[0].equals("-help"))) {
             usage();
@@ -74,9 +81,7 @@ public class TransactionTest extends TestCase {
         System.exit(2);
     }
 
-    public static Test suite()
-        throws Exception {
-
+    public static Test suite() {
         TestSuite suite = new TestSuite(TransactionTest.class);
         return suite;
     }
@@ -92,10 +97,11 @@ public class TransactionTest extends TestCase {
         super(name);
     }
 
+    @Override
     public void setUp()
         throws Exception {
 
-        DbTestUtil.printTestName(DbTestUtil.qualifiedTestName(this));
+        SharedTestUtils.printTestName(SharedTestUtils.qualifiedTestName(this));
         env = TestEnv.TXN.open("TransactionTests");
         currentTxn = CurrentTransaction.getInstance(env);
         store = testStore.open(env, dbName(0));
@@ -103,6 +109,7 @@ public class TransactionTest extends TestCase {
                                   testStore.getValueBinding(), true);
     }
 
+    @Override
     public void tearDown() {
 
         try {
@@ -200,9 +207,8 @@ public class TransactionTest extends TestCase {
         DatabaseConfig dbConfig = new DatabaseConfig();
         DbCompat.setTypeBtree(dbConfig);
         dbConfig.setAllowCreate(true);
-        Database db = DbCompat.openDatabase(env, null, 
-                                            dbName(1), null,
-                                            dbConfig);
+        Database db = DbCompat.testOpenDatabase
+            (env, null, dbName(1), null, dbConfig);
         map = new StoredSortedMap(db, testStore.getKeyBinding(),
                                       testStore.getValueBinding(), true);
         assertTrue(!map.isTransactional());
@@ -214,8 +220,8 @@ public class TransactionTest extends TestCase {
         //
         dbConfig.setTransactional(true);
         currentTxn.beginTransaction(null);
-        db = DbCompat.openDatabase(env, currentTxn.getTransaction(), 
-                                   dbName(2), null, dbConfig);
+        db = DbCompat.testOpenDatabase
+            (env, currentTxn.getTransaction(), dbName(2), null, dbConfig);
         currentTxn.commitTransaction();
         map = new StoredSortedMap(db, testStore.getKeyBinding(),
                                       testStore.getValueBinding(), true);
@@ -459,7 +465,7 @@ public class TransactionTest extends TestCase {
 
         if (DbCompat.RECNO_METHOD) {
             // create a list just so we can call configuredList()
-            Database listStore = TestStore.RECNO_RENUM.open(env, null);
+            Database listStore = TestStore.RECNO_RENUM.open(env, "foo");
             List list = new StoredList(listStore, TestStore.VALUE_BINDING,
                                        true);
             assertTrue(isReadCommitted
@@ -546,7 +552,7 @@ public class TransactionTest extends TestCase {
 
         if (DbCompat.RECNO_METHOD) {
             // create a list just so we can call configuredList()
-            Database listStore = TestStore.RECNO_RENUM.open(env, null);
+            Database listStore = TestStore.RECNO_RENUM.open(env, "foo");
             List list = new StoredList(listStore, TestStore.VALUE_BINDING,
                                        true);
             assertTrue(isReadUncommitted
@@ -581,6 +587,79 @@ public class TransactionTest extends TestCase {
         assertNull(currentTxn.getTransaction());
     }
 
+    /**
+     * Tests that the CurrentTransaction static WeakHashMap does indeed allow
+     * GC to reclaim tine environment when it is closed.  At one point this was
+     * not working because the value object in the map has a reference to the
+     * environment.  This was fixed by wrapping the Environment in a
+     * WeakReference.  [#15444]
+     *
+     * This test only succeeds intermittently, probably due to its reliance
+     * on the GC call.
+     */
+    public void testCurrentTransactionGC()
+        throws Exception {
+
+        /*
+         * This test can have indeterminate results because it depends on
+         * a finalize count, so it's not part of the default run.
+         */
+        if (!SharedTestUtils.runLongTests()) {
+            return;
+        }
+
+        final StringBuffer finalizedFlag = new StringBuffer();
+
+        class MyEnv extends Environment {
+
+            /**
+             * @throws FileNotFoundException from DB core.
+             */
+            MyEnv(File home, EnvironmentConfig config)
+                throws DatabaseException, FileNotFoundException {
+
+                super(home, config);
+            }
+
+            @Override
+            protected void finalize() {
+                finalizedFlag.append('.');
+            }
+        }
+
+        MyEnv myEnv = new MyEnv(env.getHome(), env.getConfig());
+        CurrentTransaction myCurrTxn = CurrentTransaction.getInstance(myEnv);
+
+        store.close();
+        store = null;
+        map = null;
+
+        env.close();
+        env = null;
+
+        myEnv.close();
+        myEnv = null;
+
+        myCurrTxn = null;
+        currentTxn = null;
+
+        for (int i = 0; i < 10; i += 1) {
+            byte[] x = null;
+            try {
+                 x = new byte[Integer.MAX_VALUE - 1];
+            } catch (OutOfMemoryError expected) {
+            }
+            assertNull(x);
+            System.gc();
+        }
+
+        for (int i = 0; i < 10; i += 1) {
+            System.gc();
+        }
+
+        assertTrue(finalizedFlag.length() > 0);
+    }
+
     private synchronized void doReadUncommitted(StoredSortedMap dirtyMap)
         throws Exception {
 
@@ -621,20 +700,19 @@ public class TransactionTest extends TestCase {
 
     private static class ReadUncommittedThreadOne extends Thread {
 
-        private Environment env;
-        private CurrentTransaction currentTxn;
-        private TransactionTest parent;
-        private StoredSortedMap map;
+        private final CurrentTransaction currentTxn;
+        private final TransactionTest parent;
+        private final StoredSortedMap map;
 
         private ReadUncommittedThreadOne(Environment env,
                                          TransactionTest parent) {
 
-            this.env = env;
             this.currentTxn = CurrentTransaction.getInstance(env);
             this.parent = parent;
             this.map = parent.map;
         }
 
+        @Override
         public synchronized void run() {
 
             try {
@@ -662,10 +740,10 @@ public class TransactionTest extends TestCase {
 
     private static class ReadUncommittedThreadTwo extends Thread {
 
-        private Environment env;
-        private CurrentTransaction currentTxn;
-        private TransactionTest parent;
-        private StoredSortedMap map;
+        private final Environment env;
+        private final CurrentTransaction currentTxn;
+        private final TransactionTest parent;
+        private final StoredSortedMap map;
 
         private ReadUncommittedThreadTwo(Environment env,
                                          TransactionTest parent) {
@@ -676,6 +754,7 @@ public class TransactionTest extends TestCase {
             this.map = parent.map;
         }
 
+        @Override
         public synchronized void run() {
 
             try {
@@ -742,4 +821,18 @@ public class TransactionTest extends TestCase {
             finally { StoredIterator.close(i); }
         }
     }
+
+    /**
+     * Tests transaction retries performed by TransationRunner.
+     *
+     * This test is too sensitive to how lock conflict detection works on JE to
+     * make it work properly on DB core.
+     */
+
+    /**
+     * Tests transaction retries performed by TransationRunner.
+     *
+     * This test is too sensitive to how lock conflict detection works on JE to
+     * make it work properly on DB core.
+     */
 }
