@@ -1,9 +1,8 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 1999-2005
-#	Sleepycat Software.  All rights reserved.
+# Copyright (c) 1999-2009 Oracle.  All rights reserved.
 #
-# $Id: sdb003.tcl,v 12.1 2005/06/16 20:24:04 bostic Exp $
+# $Id$
 #
 # TEST	sdb003
 # TEST	Tests many subdbs
@@ -28,6 +27,7 @@ proc sdb003 { method {nentries 1000} args } {
 	puts "Subdb003: $method ($args) many subdb tests"
 
 	set txnenv 0
+	set rpcenv 0
 	set eindex [lsearch -exact $args "-env"]
 	#
 	# If we are using an env, then testfile should just be the db name.
@@ -39,6 +39,7 @@ proc sdb003 { method {nentries 1000} args } {
 		set testfile subdb003.db
 		incr eindex
 		set env [lindex $args $eindex]
+		set rpcenv [is_rpcenv $env]
 		set txnenv [is_txnenv $env]
 		if { $txnenv == 1 } {
 			append args " -auto_commit "
@@ -70,9 +71,12 @@ proc sdb003 { method {nentries 1000} args } {
 	set ndataent 10
 	set fdid [open $dict]
 	while { [gets $fdid str] != -1 && $fcount < $nentries } {
-		if { $str == "" } {
+		# Unlike the standard API, RPC doesn't support empty
+		# database names. [#15600]
+		if { $str == "" && $rpcenv == 1 } {
 			continue
 		}
+
 		set subdb $str
 		set db [eval {berkdb_open -create -mode 0644} \
 		    $args {$omethod $testfile $subdb}]
