@@ -1,9 +1,8 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 1999-2005
-#	Sleepycat Software.  All rights reserved.
+# Copyright (c) 1999-2009 Oracle.  All rights reserved.
 #
-# $Id: sdbtest002.tcl,v 12.1 2005/06/16 20:24:04 bostic Exp $
+# $Id$
 #
 # TEST	sdbtest002
 # TEST	Tests multiple access methods in one subdb access by multiple
@@ -91,11 +90,6 @@ proc sdbtest002 { {nentries 10000} } {
 		error_check_good db_open [is_valid_db $db] TRUE
 		dump_file $db $txn $t1 $checkfunc
 		error_check_good db_close [$db close] 0
-		#
-		# This is just so that t2 is there and empty
-		# since we are only appending below.
-		#
-		exec > $t2
 
 		# Now compare the keys to see if they match the dictionary (or ints)
 		if { [is_record_based $method] == 1 } {
