@@ -1,12 +1,19 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 1996-2005
-#	Sleepycat Software.  All rights reserved.
+# Copyright (c) 1996-2009 Oracle.  All rights reserved.
 #
-# $Id: test.tcl,v 12.20 2005/11/08 15:10:43 carol Exp $
+# $Id$
 
 source ./include.tcl
 
+# Add the default Windows build sub-directory to the path, so that
+# the binaries can be found without copies.
+if {[string match Win* $tcl_platform(os)]} {
+	global env
+	global buildpath
+	set env(PATH) "$env(PATH)\;$buildpath"
+}
+
 # Load DB's TCL API.
 load $tcllib
 
@@ -44,7 +51,7 @@ set datastr "abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyz"
 
 # Random number seed.
 global rand_init
-set rand_init 12082003
+set rand_init 11302005
 
 # Default record length for fixed record length access method(s)
 set fixed_len 20
@@ -58,6 +65,18 @@ set ohandles {}
 global is_envmethod
 set is_envmethod 0
 
+#
+# Set when we're running a child process in a rep test.
+#
+global is_repchild
+set is_repchild 0
+
+# Set when we want to use replication test messaging that cannot
+# share an env -- for example, because the replication processes
+# are not all from the same BDB version.
+global noenv_messaging
+set noenv_messaging 0
+
 # For testing locker id wrap around.
 global lock_curid
 global lock_maxid
@@ -75,21 +94,37 @@ if { [info exists one_test] != 1 } {
 	set one_test "ALL"
 }
 
-# This is where the test numbering and parameters now live.
-source $test_path/testparams.tcl
+# If you call a test with the proc find_valid_methods, it will
+# return the list of methods for which it will run, instead of
+# actually running.
+global checking_valid_methods
+set checking_valid_methods 0
+global valid_methods
+set valid_methods { btree rbtree queue queueext recno frecno rrecno hash }
+
+# The variable test_recopts controls whether we open envs in
+# replication tests with the -recover flag.   The default is
+# to test with and without the flag, but to run a meaningful
+# subset of rep tests more quickly, rep_subset will randomly
+# pick one or the other.
+global test_recopts
+set test_recopts { "-recover" "" }
+
+# Set up any OS-specific values.
+source $test_path/testutils.tcl
 
-# Set up any OS-specific values
 global tcl_platform
-set is_freebsd_test [is_substr $tcl_platform(os) "FreeBSD"]
-set is_hp_test [is_substr $tcl_platform(os) "HP-UX"]
-set is_linux_test [is_substr $tcl_platform(os) "Linux"]
-set is_qnx_test [is_substr $tcl_platform(os) "QNX"]
-set is_sunos_test [is_substr $tcl_platform(os) "SunOS"]
-set is_windows_test [is_substr $tcl_platform(os) "Win"]
-set is_windows9x_test [is_substr $tcl_platform(osVersion) "Windows 95"]
+set is_freebsd_test [string match FreeBSD $tcl_platform(os)]
+set is_hp_test [string match HP-UX $tcl_platform(os)]
+set is_linux_test [string match Linux $tcl_platform(os)]
+set is_qnx_test [string match QNX $tcl_platform(os)]
+set is_sunos_test [string match SunOS $tcl_platform(os)]
+set is_windows_test [string match Win* $tcl_platform(os)]
+set is_windows9x_test [string match "Windows 95" $tcl_platform(osVersion)]
 set is_je_test 0
 set upgrade_be [big_endian]
-
+global is_fat32
+set is_fat32 [string match FAT32 [lindex [file system check] 1]]
 global EXE BAT
 if { $is_windows_test == 1 } {
 	set EXE ".exe"
@@ -99,11 +134,19 @@ if { $is_windows_test == 1 } {
 	set BAT ""
 }
 
+if { $is_windows_test == 1 } {
+	set util_path "./$buildpath"
+}
+
+# This is where the test numbering and parameters now live.
+source $test_path/testparams.tcl
+source $test_path/db_reptest.tcl
+
 # Try to open an encrypted database.  If it fails, this release
 # doesn't support encryption, and encryption tests should be skipped.
 set has_crypto 1
-set stat [catch {set db \
-    [eval {berkdb open -create -btree -encryptaes test_passwd} ] } result ]
+set stat [catch {set db [eval {berkdb_open_noerr \
+    -create -btree -encryptaes test_passwd} ] } result ]
 if { $stat != 0 } {
 	# Make sure it's the right error for a non-crypto release.
 	error_check_good non_crypto_release \
@@ -115,6 +158,14 @@ if { $stat != 0 } {
 	error_check_good close_encrypted_db [$db close] 0
 }
 
+# Get the default page size of this system
+global default_pagesize
+set db [berkdb_open_noerr -create -btree]
+error_check_good "db open" [is_valid_db $db] TRUE
+set stat [catch {set default_pagesize [$db get_pagesize]} result]
+error_check_good "db get_pagesize" $stat 0
+error_check_good "db close" [$db close] 0
+
 # From here on out, test.tcl contains the procs that are used to
 # run all or part of the test suite.
 
@@ -122,6 +173,7 @@ proc run_std { { testname ALL } args } {
 	global test_names
 	global one_test
 	global has_crypto
+	global valid_methods
 	source ./include.tcl
 
 	set one_test $testname
@@ -180,6 +232,7 @@ proc run_std { { testname ALL } args } {
 	set test_list {
 	{"environment"		"env"}
 	{"archive"		"archive"}
+	{"backup"		"backup"}
 	{"file operations"	"fop"}
 	{"locking"		"lock"}
 	{"logging"		"log"}
@@ -193,7 +246,17 @@ proc run_std { { testname ALL } args } {
 	{"NDBM interface"	"ndbm"}
 	{"Hsearch interface"	"hsearch"}
 	{"secondary index"	"sindex"}
-	{"replication"		"rep"}
+	{"partition"		"partition"}
+	{"compression"		"compressed"}
+	{"replication manager" 	"repmgr"}
+	}
+
+	# If this is run_std only, run each rep test for a single
+	# access method.  If run_all, run for all access methods.
+	if { $std_only == 1 } {
+		lappend test_list {"replication"	"rep_subset"}
+	} else {
+		lappend test_list {"replication"	"rep_complete"}
 	}
 
 	# If release supports encryption, run security tests.
@@ -267,8 +330,7 @@ proc run_std { { testname ALL } args } {
 		# XXX
 		# Broken up into separate tclsh instantiations so we don't
 		# require so much memory.
-		foreach method \
-		    "btree hash queue queueext recno rbtree frecno rrecno" {
+		foreach method $valid_methods {
 			puts "Running $method tests"
 			foreach test $test_names(test) {
 				if { $run == 0 } {
@@ -337,17 +399,20 @@ proc check_output { file } {
 		^100.*|
 		^eval\s.*|
 		^exec\s.*|
+		^fileops:\s.*|
 		^jointest.*$|
 		^r\sarchive\s*|
+		^r\sbackup\s*|
 		^r\sdbm\s*|
 		^r\shsearch\s*|
 		^r\sndbm\s*|
 		^r\srpc\s*|
 		^run_recd:\s.*|
-		^run_reptest:\s.*|
+		^run_reptest\s.*|
 		^run_rpcmethod:\s.*|
 		^run_secenv:\s.*|
 		^All\sprocesses\shave\sexited.$|
+		^Backuptest\s.*|
 		^Beginning\scycle\s\d$|
 		^Byteorder:.*|
 		^Child\sruns\scomplete\.\s\sParent\smodifies\sdata\.$|
@@ -355,24 +420,33 @@ proc check_output { file } {
 		^Ending\srecord.*|
 		^Environment\s.*?specified;\s\sskipping\.$|
 		^Executing\srecord\s.*|
+		^Freeing\smutex\s.*|
 		^Join\stest:\.*|
 		^Method:\s.*|
+                ^Putting\s.*databases.*|
 		^Repl:\stest\d\d\d:.*|
 		^Repl:\ssdb\d\d\d:.*|
+		^Running\stest\ssdb.*|
+		^Running\stest\stest.*|
+                ^run_inmem_db\s.*rep.*|
+                ^run_inmem_log\s.*rep.*|
+                ^run_mixedmode_log\s.*rep.*|
 		^Script\swatcher\sprocess\s.*|
 		^Secondary\sindex\sjoin\s.*|
-		^Sleepycat\sSoftware:\sBerkeley\sDB\s.*|
+		^Berkeley\sDB\s.*|
 		^Test\ssuite\srun\s.*|
+                ^Test\s.*rep.*|
 		^Unlinking\slog:\serror\smessage\sOK$|
 		^Verifying\s.*|
 		^\t*\.\.\.dbc->get.*$|
 		^\t*\.\.\.dbc->put.*$|
-		^\t*\.\.\.key\s\d*$|
+		^\t*\.\.\.key\s\d.*$|
 		^\t*\.\.\.Skipping\sdbc.*|
 		^\t*and\s\d*\sduplicate\sduplicates\.$|
 		^\t*About\sto\srun\srecovery\s.*complete$|
 		^\t*Add\sa\sthird\sversion\s.*|
 		^\t*Archive[:\.].*|
+		^\t*Backuptest.*|
 		^\t*Bigfile[0-9][0-9][0-9].*|
 		^\t*Building\s.*|
 		^\t*closing\ssecondaries\.$|
@@ -401,6 +475,7 @@ proc check_output { file } {
 		^\t*op_recover_rec:\sRunning\srecovery.*|
 		^\t*[r|R]ecd[0-9][0-9][0-9].*|
 		^\t*[r|R]ep[0-9][0-9][0-9].*|
+		^\t*[r|R]epmgr[0-9][0-9][0-9].*|
 		^\t*[r|R]ep_test.*|
 		^\t*[r|R]pc[0-9][0-9][0-9].*|
 		^\t*[r|R]src[0-9][0-9][0-9].*|
@@ -413,6 +488,7 @@ proc check_output { file } {
 		^\t*[s|S]ec[0-9][0-9][0-9].*|
 		^\t*[s|S]i[0-9][0-9][0-9].*|
 		^\t*[s|S]ijoin.*|
+		^\t*Salvage\stests\sof.*|
 		^\t*sdb[0-9][0-9][0-9].*|
 		^\t*Skipping\s.*|
 		^\t*Subdb[0-9][0-9][0-9].*|
@@ -443,6 +519,8 @@ proc r { args } {
 	global has_crypto
 	global rand_init
 	global one_test
+	global test_recopts
+	global checking_valid_methods
 
 	source ./include.tcl
 
@@ -473,6 +551,8 @@ proc r { args } {
 			lock -
 			log -
 			memp -
+			multi_repmgr -
+			mutex -
 			rsrc -
 			sdbtest -
 			txn {
@@ -489,15 +569,16 @@ proc r { args } {
 				}
 			}
 			archive -
+			backup -
 			dbm -
 			hsearch -
 			ndbm -
 			shelltest {
 				if { $one_test == "ALL" } {
-					if { $display } { puts "r $sub" }
+					if { $display } { puts "eval $sub" }
 					if { $run } {
 						check_handles
-						$sub
+						eval $sub
 					}
 				}
 			}
@@ -510,6 +591,16 @@ proc r { args } {
 					eval run_test $test $display $run
 				}
 			}
+			compressed {
+				set tindex [lsearch $test_names(test) $starttest]
+				if { $tindex == -1 } {
+					set tindex 0
+				}
+				set clist [lrange $test_names(test) $tindex end]
+				foreach test $clist {
+					eval run_compressed btree $test $display $run
+				}
+			}
 			join {
 				eval r $saveflags join1
 				eval r $saveflags join2
@@ -519,10 +610,10 @@ proc r { args } {
 				eval r $saveflags join6
 			}
 			join1 {
-				if { $display } { puts jointest }
+				if { $display } { puts "eval jointest" }
 				if { $run } {
 					check_handles
-					jointest
+					eval jointest
 				}
 			}
 			joinbench {
@@ -532,47 +623,130 @@ proc r { args } {
 				puts "[timestamp]"
 			}
 			join2 {
-				if { $display } { puts "jointest 512" }
+				if { $display } { puts "eval jointest 512" }
 				if { $run } {
 					check_handles
-					jointest 512
+					eval jointest 512
 				}
 			}
 			join3 {
 				if { $display } {
-					puts "jointest 8192 0 -join_item"
+					puts "eval jointest 8192 0 -join_item"
 				}
 				if { $run } {
 					check_handles
-					jointest 8192 0 -join_item
+					eval jointest 8192 0 -join_item
 				}
 			}
 			join4 {
-				if { $display } { puts "jointest 8192 2" }
+				if { $display } { puts "eval jointest 8192 2" }
 				if { $run } {
 					check_handles
-					jointest 8192 2
+					eval jointest 8192 2
 				}
 			}
 			join5 {
-				if { $display } { puts "jointest 8192 3" }
+				if { $display } { puts "eval jointest 8192 3" }
 				if { $run } {
 					check_handles
-					jointest 8192 3
+					eval jointest 8192 3
 				}
 			}
 			join6 {
-				if { $display } { puts "jointest 512 3" }
+				if { $display } { puts "eval jointest 512 3" }
 				if { $run } {
 					check_handles
-					jointest 512 3
+					eval jointest 512 3
+				}
+			}
+			partition {
+				foreach method { btree hash } {
+					foreach test "$test_names(recd)\
+					    $test_names(test)" {
+						run_range_partition\
+						    $test $method $display $run 
+						run_partition_callback\
+						    $test $method $display $run
+					}
 				}
 			}
 			recd {
 				check_handles
-				run_recds $run $display [lrange $args 1 end]
+				eval {run_recds all $run $display} [lrange $args 1 end]
+			}
+			repmgr {
+				set tindex [lsearch $test_names(repmgr) $starttest]
+				if { $tindex == -1 } {
+					set tindex 0
+				}
+				set rlist [lrange $test_names(repmgr) $tindex end]
+				foreach test $rlist {
+					run_test $test $display $run
+				}
 			}
 			rep {
+				r rep_subset $starttest
+			}
+			# To run a subset of the complete rep tests, use
+			# rep_subset, which randomly picks an access type to
+			# use, and randomly picks whether to open envs with
+			# the -recover flag.
+			rep_subset {
+				if  { [is_partition_callback $args] == 1 } {
+					set nodump 1
+				} else {
+					set nodump 0
+				}
+				berkdb srand $rand_init
+				set tindex [lsearch $test_names(rep) $starttest]
+				if { $tindex == -1 } {
+					set tindex 0
+				}
+				set rlist [lrange $test_names(rep) $tindex end]
+				foreach test $rlist {
+					set random_recopt \
+					    [berkdb random_int 0 1]
+					if { $random_recopt == 1 } {
+						set test_recopts "-recover"
+					} else {
+						set test_recopts {""}
+					}
+
+					set method_list \
+					    [find_valid_methods $test]
+					set list_length \
+					    [expr [llength $method_list] - 1]
+					set method_index \
+					    [berkdb random_int 0 $list_length]
+					set rand_method \
+					    [lindex $method_list $method_index]
+
+					if { $display } {
+						puts "eval $test $rand_method; \
+						    verify_dir \
+						    $testdir \"\" 1 0 $nodump; \
+						    salvage_dir $testdir"
+					}
+					if { $run } {
+				 		check_handles
+						eval $test $rand_method
+						verify_dir $testdir "" 1 0 $nodump
+						salvage_dir $testdir
+					}
+				}
+				if { $one_test == "ALL" } {
+					if { $display } {
+						#puts "basic_db_reptest"
+						#puts "basic_db_reptest 1"
+					}
+					if { $run } {
+						#basic_db_reptest
+						#basic_db_reptest 1
+					}
+				}
+				set test_recopts { "-recover" "" }
+			}
+			rep_complete {
 				set tindex [lsearch $test_names(rep) $starttest]
 				if { $tindex == -1 } {
 					set tindex 0
@@ -581,8 +755,18 @@ proc r { args } {
 				foreach test $rlist {
 					run_test $test $display $run
 				}
+				if { $one_test == "ALL" } {
+					if { $display } {
+						#puts "basic_db_reptest"
+						#puts "basic_db_reptest 1"
+					}
+					if { $run } {
+						#basic_db_reptest
+						#basic_db_reptest 1
+					}
+				}
 			}
-			repmethod {		
+			repmethod {
 				# We seed the random number generator here
 				# instead of in run_repmethod so that we
 				# aren't always reusing the first few
@@ -624,6 +808,7 @@ proc r { args } {
 							check_handles
 							eval run_rpcmethod -btree
 							verify_dir $testdir "" 1
+							salvage_dir $testdir
 						} else {
 							run_test run_rpcmethod $display $run
 						}
@@ -635,8 +820,8 @@ proc r { args } {
 
 					}
 					set rpc_svc $old_rpc_src
-					}
 				}
+			}
 			sec {
 				# Skip secure mode tests if release
 				# does not support encryption.
@@ -649,12 +834,12 @@ proc r { args } {
 				if { $run } {
 					run_subsystem $sub 0 1
 				}
-			}	
+			}
 			secmethod {
 				# Skip secure mode tests if release
 				# does not support encryption.
 				if { $has_crypto == 0 } {
-					return 
+					return
 				}
 				foreach test $test_names(test) {
 					eval run_test run_secmethod \
@@ -700,7 +885,7 @@ proc r { args } {
 			rrecno {
 				foreach test $test_names(test) {
 					eval run_method [lindex $args 0] $test \
-					    $display $run [lrange $args 1 end]
+					    $display $run stdout [lrange $args 1 end]
 				}
 			}
 
@@ -713,7 +898,6 @@ proc r { args } {
 		flush stderr
 	} res] != 0} {
 		global errorInfo;
-
 		set fnl [string first "\n" $errorInfo]
 		set theError [string range $errorInfo 0 [expr $fnl - 1]]
 		if {[string first FAIL $errorInfo] == -1} {
@@ -748,14 +932,24 @@ proc run_subsystem { sub { display 0 } { run 1} } {
 
 proc run_test { test {display 0} {run 1} args } {
 	source ./include.tcl
-	foreach method "hash queue queueext recno rbtree frecno rrecno btree" {
+	global valid_methods
+
+	foreach method $valid_methods {
 		if { $display } {
-			puts "eval $test -$method $args; verify_dir $testdir \"\" 1"
+			puts "eval $test -$method $args; \
+			    verify_dir $testdir \"\" 1; \
+			    salvage_dir $testdir"
+		}
+		if  { [is_partition_callback $args] == 1 } {
+			set nodump 1
+		} else {
+			set nodump 0
 		}
 		if { $run } {
 	 		check_handles
-			eval $test -$method $args
-			verify_dir $testdir "" 1
+			eval {$test -$method} $args
+			verify_dir $testdir "" 1 0 $nodump
+			salvage_dir $testdir
 		}
 	}
 }
@@ -769,24 +963,32 @@ proc run_method { method test {display 0} {run 1} \
 	global parms
 	source ./include.tcl
 
+	if  { [is_partition_callback $args] == 1 } {
+		set nodump  1
+	} else {
+		set nodump  0
+	}
+
 	if {[catch {
 		if { $display } {
-			puts -nonewline $outfile "eval $test $method"
-			puts -nonewline $outfile " $parms($test) $args"
-			puts $outfile " ; verify_dir $testdir \"\" 1"
+			puts -nonewline $outfile "eval \{ $test \} $method"
+			puts -nonewline $outfile " $parms($test) { $args }"
+			puts -nonewline $outfile " ; verify_dir $testdir \"\" 1 0 $nodump"
+			puts $outfile " ; salvage_dir $testdir"
 		}
 		if { $run } {
 			check_handles $outfile
 			puts $outfile "[timestamp]"
-			eval $test $method $parms($test) $args
+			eval {$test} $method $parms($test) $args
 			if { $__debug_print != 0 } {
 				puts $outfile ""
 			}
-			# verify all databases the test leaves behind
-			verify_dir $testdir "" 1
+			# Verify all databases the test leaves behind
+			verify_dir $testdir "" 1 0 $nodump
 			if { $__debug_on != 0 } {
 				debug $__debug_test
 			}
+			salvage_dir $testdir
 		}
 		flush stdout
 		flush stderr
@@ -814,7 +1016,7 @@ proc run_rpcmethod { method {largs ""} } {
 	global rpc_svc
 	source ./include.tcl
 
-	puts "run_rpcmethod: $method $largs"
+	puts "run_rpcmethod: $method $largs using $rpc_svc"
 
 	set save_largs $largs
 	set dpid [rpc_server_start]
@@ -862,6 +1064,8 @@ proc run_rpcmethod { method {largs ""} } {
 				append largs " -env $env "
 
 				puts "[timestamp]"
+				puts "Running test $test with RPC service $rpc_svc"
+				puts "eval $test $method $parms($test) $largs"
 				eval $test $method $parms($test) $largs
 				if { $__debug_print != 0 } {
 					puts ""
@@ -987,6 +1191,81 @@ proc run_rpcnoserver { method {largs ""} } {
 
 }
 
+# Run a testNNN or recdNNN test with range partitioning.
+proc run_range_partition { test method {display 0} {run 1}\
+    {outfile stdout} args } { 
+
+	# The only allowed access method for range partitioning is btree.
+	if { [is_btree $method] == 0 } {
+		if { $display == 0 } {
+			puts "Skipping range partition\
+			    tests for method $method"
+		}
+		return
+	}
+
+	# If we've passed in explicit partitioning args, use them; 
+	# otherwise set them.  This particular selection hits some 
+	# interesting cases where we set the key to "key".
+	set largs $args
+	if { [is_partitioned $args] == 0 } {
+		lappend largs  -partition {ab cd key key1 zzz}
+	}
+
+	if { [string first recd $test] == 0 } {
+		eval {run_recd $method $test $run $display} $largs
+	} elseif { [string first test $test] == 0 } {
+		eval {run_method $method $test $display $run $outfile} $largs
+	} else {
+		puts "Skipping test $test with range partitioning."
+	}
+}
+
+# Run a testNNN or recdNNN test with partition callbacks.
+proc run_partition_callback { test method {display 0} {run 1}\
+    {outfile stdout} args } {
+
+	# The only allowed access methods are btree and hash.
+	if { [is_btree $method] == 0 && [is_hash $method] == 0 } {
+		if { $display == 0 } {
+			puts "Skipping partition callback tests\
+			    for method $method"
+		}
+		return
+	}
+
+	# If we've passed in explicit partitioning args, use them; 
+	# otherwise set them. 
+	set largs $args
+	if { [is_partition_callback $args] == 0 } {
+		lappend largs  -partition_callback 5 part 
+	}
+
+	if { [string first recd $test] == 0 } {
+		eval {run_recd $method $test $run $display} $largs
+	} elseif { [string first test $test] == 0 } {
+		eval {run_method $method $test $display $run $outfile} $largs
+	} else {
+		puts "Skipping test $test with partition callbacks."
+	}
+}
+
+#
+# Run method tests for btree only using compression.
+#
+proc run_compressed { method test {display 0} {run 1} \
+    { outfile stdout } args } {
+
+	if { [is_btree $method] == 0 } {
+		puts "Skipping compression test for method $method."
+		return
+	}
+
+	set largs $args
+	append largs " -compress "
+	eval run_method $method $test $display $run $outfile $largs
+}
+
 #
 # Run method tests in secure mode.
 #
@@ -1032,7 +1311,7 @@ proc run_secenv { method test {largs ""} } {
 	set stat [catch {
 		check_handles
 		set env [eval {berkdb_env -create -mode 0644 -home $testdir \
-		    -encryptaes $passwd -cachesize {0 2097152 1}}]
+		    -encryptaes $passwd -pagesize 512 -cachesize {0 4194304 1}}]
 		error_check_good env_open [is_valid_env $env] TRUE
 		append largs " -env $env "
 
@@ -1087,13 +1366,9 @@ proc run_secenv { method test {largs ""} } {
 proc run_reptest { method test {droppct 0} {nclients 1} {do_del 0} \
     {do_sec 0} {do_oob 0} {largs "" } } {
 	source ./include.tcl
-	if { $is_windows9x_test == 1 } { 
+	if { $is_windows9x_test == 1 } {
 		puts "Skipping replication test on Win 9x platform."
 		return
-	} 
-	if { $test == "test115" } {
-		puts "Skipping $test under repmethod (SR #13394)."
-		return
 	}
 
 	global __debug_on
@@ -1104,7 +1379,7 @@ proc run_reptest { method test {droppct 0} {nclients 1} {do_del 0} \
 	global passwd
 	global has_crypto
 
-	puts "run_reptest: \
+	puts "run_reptest \
 	    $method $test $droppct $nclients $do_del $do_sec $do_oob $largs"
 
 	env_cleanup $testdir
@@ -1189,10 +1464,10 @@ proc run_reptest { method test {droppct 0} {nclients 1} {do_del 0} \
 proc run_repmethod { method test {numcl 0} {display 0} {run 1} \
     {outfile stdout} {largs ""} } {
 	source ./include.tcl
-	if { $is_windows9x_test == 1 } { 
+	if { $is_windows9x_test == 1 } {
 		puts "Skipping replication test on Win 9x platform."
 		return
-	} 
+	}
 
 	global __debug_on
 	global __debug_print
@@ -1222,6 +1497,14 @@ proc run_repmethod { method test {numcl 0} {display 0} {run 1} \
 	}
 	set drindex [berkdb random_int 0 $drop_len]
 	set droppct [lindex $drop_list $drindex]
+
+	# Do not drop messages on Windows.  Since we can't set 
+	# re-request times with less than millisecond precision, 
+	# dropping messages will cause test failures. 
+	if { $is_windows_test == 1 } {
+		set droppct 0
+	}
+
  	set do_sec [berkdb random_int 0 1]
 	set do_oob [berkdb random_int 0 1]
 	set do_del [berkdb random_int 0 1]
@@ -1253,8 +1536,13 @@ proc run_envmethod { method test {display 0} {run 1} {outfile stdout} \
 	set save_largs $largs
 	set envargs ""
 
+	# Enlarge the logging region by default - sdb004 needs this because
+	# it uses very long subdb names, and the names are stored in the 
+	# env region.
+	set logargs " -log_regionmax 2057152 "
+
 	# Enlarge the cache by default - some compaction tests need it.
-	set cacheargs "-cachesize {0 2097152 1}"
+	set cacheargs "-cachesize {0 4194304 1} -pagesize 512"
 	env_cleanup $testdir
 
 	if { $display == 1 } {
@@ -1274,6 +1562,12 @@ proc run_envmethod { method test {display 0} {run 1} {outfile stdout} \
 		}
 	}
 
+	set sindex [lsearch -exact $largs "-log_max"]
+	if { $sindex >= 0 } {
+		append envargs " -log_max 100000 "
+		set largs [lreplace $largs $sindex $sindex]
+	}
+
 	# Test for -thread option and pass to berkdb_env open.  Leave in
 	# $largs because -thread can also be passed to an individual
 	# test as an arg.  Double the number of lockers because a threaded
@@ -1291,7 +1585,7 @@ proc run_envmethod { method test {display 0} {run 1} {outfile stdout} \
 	}
 
 	# We raise the number of locks and objects - there are a few
-	# compaction tests that require a large number. 
+	# compaction tests that require a large number.
 	set lockargs " -lock_max_locks 40000 -lock_max_objects 20000 "
 
 	if { $run == 1 } {
@@ -1299,7 +1593,7 @@ proc run_envmethod { method test {display 0} {run 1} {outfile stdout} \
 		set stat [catch {
 			check_handles
 			set env [eval {berkdb_env -create -txn -mode 0644 \
-			    -home $testdir} $cacheargs $lockargs $envargs]
+			    -home $testdir} $logargs $cacheargs $lockargs $envargs]
 			error_check_good env_open [is_valid_env $env] TRUE
 			append largs " -env $env "
 
@@ -1321,8 +1615,8 @@ proc run_envmethod { method test {display 0} {run 1} {outfile stdout} \
 			flush stderr
 			set largs $save_largs
 			error_check_good envclose [$env close] 0
-			error_check_good envremove [berkdb envremove \
-			    -home $testdir] 0
+#			error_check_good envremove [berkdb envremove \
+#			    -home $testdir] 0
 		} res]
 		if { $stat != 0} {
 			global errorInfo;
@@ -1340,6 +1634,27 @@ proc run_envmethod { method test {display 0} {run 1} {outfile stdout} \
 	}
 }
 
+proc run_compact { method } {
+	source ./include.tcl
+	for {set tnum 111} {$tnum <= 115} {incr tnum} {
+		run_envmethod $method test$tnum 0 1 stdout -log_max
+
+		puts "\tTest$tnum: Test Recovery"
+		set env1 [eval  berkdb env -create -txn \
+		    -recover_fatal -home $testdir]
+		error_check_good env_close [$env1 close] 0
+		error_check_good verify_dir \
+		    [verify_dir $testdir "" 0 0 1 ] 0
+		puts "\tTest$tnum: Remove db and test Recovery"
+		exec sh -c "rm -f $testdir/*.db"
+		set env1 [eval  berkdb env -create -txn \
+		    -recover_fatal -home $testdir]
+		error_check_good env_close [$env1 close] 0
+		error_check_good verify_dir \
+		    [verify_dir $testdir "" 0 0 1 ] 0
+	}
+}
+
 proc run_recd { method test {run 1} {display 0} args } {
 	global __debug_on
 	global __debug_print
@@ -1359,7 +1674,7 @@ proc run_recd { method test {run 1} {display 0} args } {
 	}
 	if {[catch {
 		if { $display } {
-			puts "eval $test $method $parms($test) $args"
+			puts "eval { $test } $method $parms($test) { $args }"
 		}
 		if { $run } {
 			check_handles
@@ -1378,7 +1693,7 @@ proc run_recd { method test {run 1} {display 0} args } {
 			    set upgrade_dir $upgrade_dir; \
 			    set upgrade_method $upgrade_method; \
 			    set upgrade_name $upgrade_name; \
-			    eval $test $method $parms($test) $args" \
+			    eval { $test } $method $parms($test) {$args}" \
 			    >&@ stdout
 			} res]
 
@@ -1411,18 +1726,31 @@ proc run_recd { method test {run 1} {display 0} args } {
 	}
 }
 
-proc run_recds { {run 1} {display 0} args } {
+proc recds {method args} {
+	eval {run_recds $method 1 0} $args
+}
+
+proc run_recds {{run_methods "all"} {run 1} {display 0} args } {
 	source ./include.tcl
 	global log_log_record_types
 	global test_names
 	global gen_upgrade_log
 	global encrypt
+	global valid_methods
 
 	set log_log_record_types 1
+	set run_zero 0
+	if { $run_methods == "all" } {
+		set run_methods  $valid_methods
+		set run_zero 1
+	}
 	logtrack_init
 
-	foreach method \
-	    "btree rbtree hash queue queueext recno frecno rrecno" {
+	# Define a small set of tests to run with log file zeroing.
+	set zero_log_tests \
+	    {recd001 recd002 recd003 recd004 recd005 recd006 recd007}
+
+	foreach method $run_methods {
 		check_handles
 #set test_names(recd) "recd005 recd017"
 		foreach test $test_names(recd) {
@@ -1438,16 +1766,28 @@ proc run_recds { {run 1} {display 0} args } {
 				puts "Skipping $test for crypto run."
 				continue
 			}
-			if { [catch {eval \
-			    run_recd $method $test $run $display \
-			    $args} ret ] != 0 } {
+			if { [catch {eval {run_recd $method $test $run \
+			    $display} $args} ret ] != 0 } {
 				puts $ret
 			}
+	
+			# If it's one of the chosen tests, and btree, run with 
+			# log file zeroing.
+			set zlog_idx [lsearch -exact $zero_log_tests $test]
+			if { $run_zero == 1 && \
+			    $method == "btree" && $zlog_idx > -1 } {
+				if { [catch {eval {run_recd $method $test \
+				    $run $display -zero_log} $args} ret ] != 0 } { 
+					puts $ret
+				}	
+			}
+
 			if { $gen_upgrade_log == 1 } {
 				save_upgrade_files $testdir
 			}
 		}
 	}
+
 	# We can skip logtrack_summary during the crypto upgrade run -
 	# it doesn't introduce any new log types.
 	if { $run } {
@@ -1458,10 +1798,39 @@ proc run_recds { {run 1} {display 0} args } {
 	set log_log_record_types 0
 }
 
+# A small subset of tests to be used in conjunction with the 
+# automated builds.  Ideally these tests will cover a lot of ground
+# but run in only 15 minutes or so.  You can put any test in the 
+# list of tests and it will be run all the ways that run_all 
+# runs it. 
+proc run_smoke { } { 
+	source ./include.tcl
+	global valid_methods
+	
+	fileremove -f SMOKE.OUT
+
+	set smoke_tests { \ 
+	    lock001 log001 test001 test004 sdb001 sec001 rep001 txn001 }
+
+	# Run each test in all its permutations, and 
+	# concatenate the results in the file SMOKE.OUT.
+	foreach test $smoke_tests {
+		run_all $test
+		set in [open ALL.OUT r]
+		set out [open SMOKE.OUT a]
+		while { [gets $in str] != -1 } {
+			puts $out $str
+		}
+		close $in
+		close $out
+	}
+}
+
 proc run_all { { testname ALL } args } {
 	global test_names
 	global one_test
 	global has_crypto
+	global valid_methods
 	source ./include.tcl
 
 	fileremove -f ALL.OUT
@@ -1551,8 +1920,7 @@ proc run_all { { testname ALL } args } {
 		# XXX
 		# Broken up into separate tclsh instantiations so
 		# we don't require so much memory.
-		foreach method \
-		   "btree rbtree hash queue queueext recno frecno rrecno" {
+		foreach method $valid_methods {
 			puts "Running $method tests with pagesize $pgsz"
 			foreach sub {test sdb si} {
 				foreach test $test_names($sub) {
@@ -1586,7 +1954,7 @@ proc run_all { { testname ALL } args } {
 	#
 	# Run access method tests at default page size in one env.
 	#
-	foreach method "btree rbtree hash queue queueext recno frecno rrecno" {
+	foreach method $valid_methods {
 		puts "Running $method tests in a txn env"
 		foreach sub {test sdb si} {
 			foreach test $test_names($sub) {
@@ -1617,7 +1985,7 @@ proc run_all { { testname ALL } args } {
 	# Run access method tests at default page size in thread-enabled env.
 	# We're not truly running threaded tests, just testing the interface.
 	#
-	foreach method "btree rbtree hash queue queueext recno frecno rrecno" {
+	foreach method $valid_methods {
 		puts "Running $method tests in a threaded txn env"
 		foreach sub {test sdb si} {
 			foreach test $test_names($sub) {
@@ -1647,7 +2015,7 @@ proc run_all { { testname ALL } args } {
 	#
 	# Run access method tests at default page size with -alloc enabled.
 	#
-	foreach method "btree rbtree hash queue queueext recno frecno rrecno" {
+	foreach method $valid_methods {
 		puts "Running $method tests in an env with -alloc"
 		foreach sub {test sdb si} {
 			foreach test $test_names($sub) {
@@ -1679,13 +2047,230 @@ proc run_all { { testname ALL } args } {
 	#
 	set test_list [list {"testNNN under replication"	"repmethod"}]
 
-	# If we're on Windows, Linux, FreeBSD, or Solaris, run the 
+	# If we're on Windows, Linux, FreeBSD, or Solaris, run the
 	# bigfile tests.  These create files larger than 4 GB.
 	if { $is_freebsd_test == 1 || $is_linux_test == 1 || \
 	    $is_sunos_test == 1 || $is_windows_test == 1 } {
-		lappend test_list {"big files"	"bigfile"}	
+		lappend test_list {"big files"	"bigfile"}
 	}
-	
+
+	# If release supports encryption, run security tests.
+	#
+	if { $has_crypto == 1 } {
+		lappend test_list {"testNNN with security"	"secmethod"}
+	}
+	#
+	# If configured for RPC, then run rpc tests too.
+	#
+	if { [file exists ./berkeley_db_svc] ||
+	     [file exists ./berkeley_db_cxxsvc] ||
+	     [file exists ./berkeley_db_javasvc] } {
+		lappend test_list {"RPC"	"rpc"}
+	}
+
+	foreach pair $test_list {
+		set msg [lindex $pair 0]
+		set cmd [lindex $pair 1]
+		puts "Running $msg tests"
+		if [catch {exec $tclsh_path << \
+		    "global one_test; set one_test $one_test; \
+		    source $test_path/test.tcl; \
+		    r $rflags $cmd $args" >>& ALL.OUT } res] {
+			set o [open ALL.OUT a]
+			puts $o "FAIL: $cmd test: $res"
+			close $o
+		}
+	}
+
+	# If not actually running, no need to check for failure.
+	if { $run == 0 } {
+		return
+	}
+
+	set failed 0
+	set o [open ALL.OUT r]
+	while { [gets $o line] >= 0 } {
+		if { [regexp {^FAIL} $line] != 0 } {
+			set failed 1
+		}
+	}
+	close $o
+	set o [open ALL.OUT a]
+	if { $failed == 0 } {
+		puts "Regression Tests Succeeded"
+		puts $o "Regression Tests Succeeded"
+	} else {
+		puts "Regression Tests Failed; see ALL.OUT for log"
+		puts $o "Regression Tests Failed"
+	}
+
+	puts -nonewline "Test suite run completed at: "
+	puts [clock format [clock seconds] -format "%H:%M %D"]
+	puts -nonewline $o "Test suite run completed at: "
+	puts $o [clock format [clock seconds] -format "%H:%M %D"]
+	close $o
+}
+
+proc run_all_new { { testname ALL } args } {
+	global test_names
+	global one_test
+	global has_crypto
+	global valid_methods
+	source ./include.tcl
+
+	fileremove -f ALL.OUT
+
+	set one_test $testname
+	if { $one_test != "ALL" } {
+		# Source testparams again to adjust test_names.
+		source $test_path/testparams.tcl
+	}
+
+	set exflgs [eval extractflags $args]
+	set flags [lindex $exflgs 1]
+	set display 1
+	set run 1
+	set am_only 0
+	set parallel 0
+	set nparalleltests 0
+	set rflags {--}
+	foreach f $flags {
+		switch $f {
+			m {
+				set am_only 1
+			}
+			n {
+				set display 1
+				set run 0
+				set rflags [linsert $rflags 0 "-n"]
+			}
+		}
+	}
+
+	set o [open ALL.OUT a]
+	if { $run == 1 } {
+		puts -nonewline "Test suite run started at: "
+		puts [clock format [clock seconds] -format "%H:%M %D"]
+		puts [berkdb version -string]
+
+		puts -nonewline $o "Test suite run started at: "
+		puts $o [clock format [clock seconds] -format "%H:%M %D"]
+		puts $o [berkdb version -string]
+	}
+	close $o
+	#
+	# First run standard tests.  Send in a -A to let run_std know
+	# that it is part of the "run_all" run, so that it doesn't
+	# print out start/end times.
+	#
+	lappend args -A
+	eval {run_std} $one_test $args
+
+	set test_pagesizes [get_test_pagesizes]
+	set args [lindex $exflgs 0]
+	set save_args $args
+
+	#
+	# Run access method tests at default page size in one env.
+	#
+	foreach method $valid_methods {
+		puts "Running $method tests in a txn env"
+		foreach sub {test sdb si} {
+			foreach test $test_names($sub) {
+				if { $run == 0 } {
+					set o [open ALL.OUT a]
+					run_envmethod -$method $test $display \
+					    $run $o $args
+					close $o
+				}
+				if { $run } {
+					if [catch {exec $tclsh_path << \
+					    "global one_test; \
+					    set one_test $one_test; \
+					    source $test_path/test.tcl; \
+					    run_envmethod -$method $test \
+				  	    $display $run stdout $args" \
+					    >>& ALL.OUT } res] {
+						set o [open ALL.OUT a]
+						puts $o "FAIL: run_envmethod \
+						    $method $test: $res"
+						close $o
+					}
+				}
+			}
+		}
+	}
+	#
+	# Run access method tests at default page size in thread-enabled env.
+	# We're not truly running threaded tests, just testing the interface.
+	#
+	foreach method $valid_methods {
+		puts "Running $method tests in a threaded txn env"
+		set thread_tests "test001"	
+		foreach test $thread_tests {
+			if { $run == 0 } {
+				set o [open ALL.OUT a]
+				eval {run_envmethod -$method $test \
+				    $display $run $o -thread}
+				close $o
+			}
+			if { $run } {
+				if [catch {exec $tclsh_path << \
+				    "global one_test; \
+				    set one_test $one_test; \
+				    source $test_path/test.tcl; \
+				    eval {run_envmethod -$method $test \
+			  	    $display $run stdout -thread}" \
+				    >>& ALL.OUT } res] {
+					set o [open ALL.OUT a]
+					puts $o "FAIL: run_envmethod \
+					    $method $test -thread: $res"
+					close $o
+				}
+			}
+		}
+	}
+	#
+	# Run access method tests at default page size with -alloc enabled.
+	#
+	foreach method $valid_methods {
+		puts "Running $method tests in an env with -alloc"
+		set alloc_tests "test001"
+		foreach test $alloc_tests {
+			if { $run == 0 } {
+				set o [open ALL.OUT a]
+				eval {run_envmethod -$method $test \
+				    $display $run $o -alloc}
+				close $o
+			}
+			if { $run } {
+				if [catch {exec $tclsh_path << \
+				    "global one_test; \
+				    set one_test $one_test; \
+				    source $test_path/test.tcl; \
+				    eval {run_envmethod -$method $test \
+			  	    $display $run stdout -alloc}" \
+				    >>& ALL.OUT } res] {
+					set o [open ALL.OUT a]
+					puts $o "FAIL: run_envmethod \
+					    $method $test -alloc: $res"
+					close $o
+				}
+			}
+		}
+	}
+
+	# Run standard access method tests under replication.
+	#
+	set test_list [list {"testNNN under replication"	"repmethod"}]
+
+	# If we're on Windows, Linux, FreeBSD, or Solaris, run the
+	# bigfile tests.  These create files larger than 4 GB.
+	if { $is_freebsd_test == 1 || $is_linux_test == 1 || \
+	    $is_sunos_test == 1 || $is_windows_test == 1 } {
+		lappend test_list {"big files"	"bigfile"}
+	}
+
 	# If release supports encryption, run security tests.
 	#
 	if { $has_crypto == 1 } {
@@ -1766,7 +2351,7 @@ proc run_envmethod1 { method {display 0} {run 1} { outfile stdout } args } {
 		env_cleanup $testdir
 		error_check_good envremove [berkdb envremove -home $testdir] 0
 		set env [eval {berkdb_env -create -cachesize {0 10000000 0}} \
-		    {-mode 0644 -home $testdir}]
+ 		    {-pagesize 512 -mode 0644 -home $testdir}]
 		error_check_good env_open [is_valid_env $env] TRUE
 		append largs " -env $env "
 	}
@@ -1874,14 +2459,18 @@ proc sindex { {display 0} {run 1} {outfile stdout} {verbose 0} args } {
 						puts -nonewline $outfile \
 						    "eval $test {\[list\
 						    $pm $sm $sm\]} $n ;"
-						puts $outfile " verify_dir \
-						    $testdir \"\" 1"
+						puts -nonewline $outfile \
+						    " verify_dir \
+						    $testdir \"\" 1; "
+						puts $outfile " salvage_dir \
+						    $testdir "
 					}
 					if { $run } {
 			 			check_handles $outfile
 						eval $test \
 						    {[list $pm $sm $sm]} $n
 						verify_dir $testdir "" 1
+						salvage_dir $testdir
 					}
 				}
 			}
@@ -2000,3 +2589,45 @@ proc get_test_pagesizes { } {
 	}
 	error_check_good NOTREACHED 0 1
 }
+
+proc run_timed_once { timedtest args } {
+	set start [timestamp -r]
+	set ret [catch {
+		eval $timedtest $args
+		flush stdout
+		flush stderr
+	} res]
+	set stop [timestamp -r]
+	if { $ret != 0 } {
+		global errorInfo
+
+		set fnl [string first "\n" $errorInfo]
+		set theError [string range $errorInfo 0 [expr $fnl - 1]]
+		if {[string first FAIL $errorInfo] == -1} {
+			error "FAIL:[timestamp]\
+			    run_timed: $timedtest: $theError"
+		} else {
+			error $theError;
+		}
+	}
+	return [expr $stop - $start]
+}
+
+proc run_timed { niter timedtest args } {
+	if { $niter < 1 } {
+		error "run_timed: Invalid number of iterations $niter"
+	}
+	set sum 0
+	set e {}
+	for { set i 1 } { $i <= $niter } { incr i } {
+		set elapsed [eval run_timed_once $timedtest $args]
+		lappend e $elapsed
+		set sum [expr $sum + $elapsed]
+		puts "Test $timedtest run $i completed: $elapsed seconds"
+	}
+	if { $niter > 1 } {
+		set avg [expr $sum / $niter]
+		puts "Average $timedtest time: $avg"
+		puts "Raw $timedtest data: $e"
+	}
+}
