@@ -1,14 +1,13 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 1996-2005
-#	Sleepycat Software.  All rights reserved.
+# Copyright (c) 1996-2009 Oracle.  All rights reserved.
 #
-# $Id: test013.tcl,v 12.1 2005/06/16 20:24:05 bostic Exp $
+# $Id$
 #
 # TEST	test013
 # TEST	Partial put test
 # TEST		Overwrite entire records using partial puts.
-# TEST		Make surethat NOOVERWRITE flag works.
+# TEST		Make sure that NOOVERWRITE flag works.
 # TEST
 # TEST	1. Insert 10000 keys and retrieve them (equal key/data pairs).
 # TEST	2. Attempt to overwrite keys with NO_OVERWRITE set (expect error).
@@ -200,8 +199,8 @@ proc test013 { method {nentries 10000} args } {
 
 	puts "\tTest013.e: close, open, and dump file"
 	# Now, reopen the file and run the last test again.
-	open_and_dump_file $testfile $env $t1 $checkfunc \
-	    dump_file_direction "-first" "-next"
+	eval open_and_dump_file $testfile $env $t1 $checkfunc \
+	    dump_file_direction "-first" "-next" $args
 
 	if { [is_record_based $method] == 0 } {
 		filesort $t1 $t3
@@ -213,8 +212,8 @@ proc test013 { method {nentries 10000} args } {
 	# Now, reopen the file and run the last test again in the
 	# reverse direction.
 	puts "\tTest013.f: close, open, and dump file in reverse direction"
-	open_and_dump_file $testfile $env $t1 $checkfunc \
-	    dump_file_direction "-last" "-prev"
+	eval open_and_dump_file $testfile $env $t1 $checkfunc \
+	    dump_file_direction "-last" "-prev" $args
 
 	if { [is_record_based $method] == 0 } {
 		filesort $t1 $t3
