@@ -1,9 +1,8 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 1996-2003
-#	Sleepycat Software.  All rights reserved.
+# Copyright (c) 1996-2009 Oracle.  All rights reserved.
 #
-# $Id: test020.tcl,v 11.18 2003/01/08 05:53:45 bostic Exp $
+# $Id$
 #
 # TEST	test020
 # TEST	In-Memory database tests.
@@ -17,6 +16,11 @@ proc test020 { method {nentries 10000} args } {
 		puts "Test020 skipping for method $method"
 		return
 	}
+
+	if { [is_partitioned $args] == 1 } {
+		puts "Test020 skipping for partitioned $method"
+		return
+	}
 	# Create the database and open the dictionary
 	set t1 $testdir/t1
 	set t2 $testdir/t2
