@@ -1,9 +1,8 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 1996-2005
-#	Sleepycat Software.  All rights reserved.
+# Copyright (c) 1996-2009 Oracle.  All rights reserved.
 #
-# $Id: test023.tcl,v 12.1 2005/06/16 20:24:06 bostic Exp $
+# $Id$
 #
 # TEST	test023
 # TEST	Duplicate test
@@ -20,6 +19,13 @@ proc test023 { method args } {
 
 	set args [convert_args $method $args]
 	set omethod [convert_method $method]
+
+	# Btree with compression does not support unsorted duplicates.
+	if { [is_compressed $args] == 1 } {
+		puts "Test023 skipping for btree with compression."
+		return
+	}
+
 	puts "Test023: $method delete duplicates/check cursor operations"
 	if { [is_record_based $method] == 1 || \
 	    [is_rbtree $method] == 1 } {
