@@ -1,9 +1,8 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 2003-2005
-#	Sleepycat Software.  All rights reserved.
+# Copyright (c) 2003-2009 Oracle.  All rights reserved.
 #
-# $Id: test106.tcl,v 12.2 2005/07/28 18:45:45 carol Exp $
+# $Id$
 #
 # TEST	test106
 # TEST
@@ -19,6 +18,8 @@ proc test106 { method {nitems 100} {niter 200} {tnum "106"} args } {
 
 	set args [convert_args $method $args]
 	set omethod [convert_method $method]
+	set pageargs ""
+	split_pageargs $args pageargs
 
 	if { [is_btree $method] != 1 } {
 		puts "\tTest$tnum: Skipping for method $method."
@@ -39,7 +40,8 @@ proc test106 { method {nitems 100} {niter 200} {tnum "106"} args } {
 	#
 	set eindex [lsearch -exact $args "-env"]
 	if { $eindex == -1 } {
-		set env [berkdb_env -create -home $testdir -txn]
+		set env \
+		    [eval {berkdb_env -create -home $testdir -txn} $pageargs]
 	} else {
 		incr eindex
 		set env [lindex $args $eindex]
@@ -51,7 +53,6 @@ proc test106 { method {nitems 100} {niter 200} {tnum "106"} args } {
 		set testdir [get_home $env]
 	}
 
-
 	cleanup $testdir $env
 
 	# The bulk of the work of this test is done in t106script.tcl.
@@ -105,7 +106,7 @@ proc test106 { method {nitems 100} {niter 200} {tnum "106"} args } {
 		tclkill $dpid
 	}
 
-	# If this test created the env, close it. 
+	# If this test created the env, close it.
 	if { $eindex == -1 } {
 		error_check_good env_close [$env close] 0
 	}
